{
        Program Name : BehavMan
        Package Name : LB
        Purpose      : Behavior manager
        Institution  : Conroy & Conroy Co.
        Date Written : 18-July-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBehavior_Manager is a class that is used to manage behaviors for an
        object.
}

unit Behavman ;

interface

{$I EDEFINES.INC}

uses
{$IFDEF Delphi}
     { Borland... }
     Classes, { TList }
{$ENDIF}
     { C&C... }
     Behavior, { PBehavior }
     Collect, { TCollection }
     Common, { TObject }
     Events ; { PEvent }

type TSBehavior_Manager = object( TBase_Object )
                              private { Instance data... }
                                  _Collection : TCollection ;

                              public { Constructors and destructors... }
                                  constructor Init ;

                                  destructor Done ; virtual ;

                              public { API... }
                                  procedure Add( _Behavior : PSBehavior ) ;
                                      virtual ;

                                  { Checks all behaviors to see if event is
                                    allowed. }
                                  function Allow_Normal_Processing( Event : PSEvent ) : boolean ;
                                      virtual ;

                                  { Ensures the each behavior has a chance to
                                    execute on the event. }
                                  procedure Execute( Event : PSEvent ) ; virtual ;

                                  procedure Remove( _Behavior : PSBehavior ) ; virtual ;
                          end ;
     PSBehavior_Manager = ^TSBehavior_Manager ;

{$IFDEF Delphi}
type TBehavior_Manager = class( TBase_Delphi_Object )
                             private { Instance data... }
                                 _Collection : TList ;

                             public { Constructors and destructors... }
                                 constructor Create ;

                                 destructor Destroy ; override ;

                             public { API... }
                                 procedure Add( _Behavior : TBehavior ) ;
                                     virtual ;

                                 { Checks all behaviors to see if event is
                                   allowed. }
                                 function Allow_Normal_Processing( Event : TEvent ) : boolean ;
                                     virtual ;

                                 { Ensures the each behavior has a chance to
                                   execute on the event. }
                                 procedure Execute( Event : TEvent ) ; virtual ;

                                 procedure Remove( _Behavior : TBehavior ) ; virtual ;
                         end ;
{$ENDIF}

implementation

{ TSBehavior_Manager methods... }

{ Constructors and destructors... }

constructor TSBehavior_Manager.Init ;

begin
    _Collection.Init( 0, 4 ) ;
end ;


destructor TSBehavior_Manager.Done ;

begin
    _Collection.Done ;
end ;



{ API... }

procedure TSBehavior_Manager.Add( _Behavior : PSBehavior ) ;

begin
    if( _Behavior <> nil ) then
    begin
        _Collection.Insert( _Behavior ) ;
        _Behavior^.Attach ;
    end ;
end ;


function TSBehavior_Manager.Allow_Normal_Processing( Event : PSEvent ) : boolean ;

var B : PSBehavior ;
    Loop : integer ;

begin
    Allow_Normal_Processing := True ;
    if( Event = nil ) then
    begin
        exit ;
    end ;
    for Loop := 0 to _Collection.Count - 1 do
    begin
        B := PSBehavior( _Collection.At( Loop ) ) ;
        if( not B^.Allow_Normal_Processing( Event ) ) then
        begin
            Allow_Normal_Processing := False ;
        end ;
    end ;
end ;


procedure TSBehavior_Manager.Execute( Event : PSEvent ) ;

var B : PSBehavior ;
    Loop : integer ;

begin
    if( Event = nil ) then
    begin
        exit ;
    end ;
    for Loop := 0 to _Collection.Count - 1 do
    begin
        B := PSBehavior( _Collection.At( Loop ) ) ;
        B^.Execute( Event ) ;
    end ;
end ;


procedure TSBehavior_Manager.Remove( _Behavior : PSBehavior ) ;

var Dummy : integer ;

begin
    Dummy := _Collection.IndexOf( _Behavior ) ;
    if( Dummy <> -1 ) then
    begin
        _Collection.Delete( _Behavior ) ;
        _Behavior^.Detach ;
    end ;
end ;



{$IFDEF Delphi}
{ TBehavior_Manager methods... }

{ Constructors and destructors... }

constructor TBehavior_Manager.Create ;

begin
    inherited Create ;

    _Collection := TList.Create ;
end ;


destructor TBehavior_Manager.Destroy ;

begin
    _Collection.Free ;
    _Collection := nil ;

    inherited Destroy ;
end ;


{ API... }

procedure TBehavior_Manager.Add( _Behavior : TBehavior ) ;

begin
    if( _Behavior <> nil ) then
    begin
        _Collection.Add( _Behavior ) ;
        _Behavior.Attach ;
    end ;
end ;


function TBehavior_Manager.Allow_Normal_Processing( Event : TEvent ) : boolean ;

var B : TBehavior ;
    Loop : integer ;

begin
    Allow_Normal_Processing := True ;
    if( Event = nil ) then
    begin
        exit ;
    end ;
    for Loop := 0 to _Collection.Count - 1 do
    begin
        B := TBehavior( _Collection[ Loop ] ) ;
        if( not B.Allow_Normal_Processing( Event ) ) then
        begin
            Allow_Normal_Processing := False ;
        end ;
    end ;
end ;


procedure TBehavior_Manager.Execute( Event : TEvent ) ;

var B : TBehavior ;
    Loop : integer ;

begin
    if( Event = nil ) then
    begin
        exit ;
    end ;
    for Loop := 0 to _Collection.Count - 1 do
    begin
        B := TBehavior( _Collection[ Loop ] ) ;
        B.Execute( Event ) ;
    end ;
end ;


procedure TBehavior_Manager.Remove( _Behavior : TBehavior ) ;

var Dummy : integer ;

begin
    Dummy := _Collection.IndexOf( _Behavior ) ;
    if( Dummy <> -1 ) then
    begin
        _Collection.Remove( _Behavior ) ;
        _Behavior.Detach ;
    end ;
end ;
{$ENDIF}


end.

