{
        Program Name : Common
        Package Name : Subroutine Library - Layer 0
        Purpose      : Common base objects (TBase_Object and TBase_Delphi_Object)
        Institution  : Conroy & Conroy Co.
        Date Written : 19-Mar-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit Common ;

interface

{$I EDEFINES.INC}

uses _Common, { TObject }
     _DebugIn, { TDebug_Interface }
     _Propert, { TProperties }
     _Streams ; { TSStream }

{$IFDEF DELPHI}
type TBase_Delphi_Object = class( TDelphi_Object )
                               private { Instance data... }
                                    _Reference_Count : integer ;
                                    _Properties : TDynamic_Properties ;
{$IFDEF Debug}
                                    Sanity : string[ 19 ] ;
{$ENDIF}

                               public { Constructors and destructors... }
                                    constructor Create ;
                                    destructor Destroy ; override ;

                               public { API... }
                                    procedure Attach ; override ;
                                    procedure Detach ; override ;
                                    function Reference_Count : integer ;
                                    function Is_Class( const Name : string ) : boolean ;
                                        override ;

                                    function Clone : TDelphi_Object ; virtual ;

                                    function Properties : TDynamic_Properties ;
                                        virtual ;
                                    function Debugger : TDebug_Interface ;
                                        virtual ;
                                    procedure Test ; virtual ;

                                    function Deserialize( Stream : TStream ) : boolean ;
                                        virtual ;
                                    function Serialize( Stream : TStream ) : boolean ;
                                        virtual ;
                                    function Load( Stream : TStream ) : boolean ;
                                        virtual ;
                                    function Persist( Stream : TStream ) : boolean ;
                                        virtual ;
                           end ;
{$ENDIF}

type PBase_Object = ^TBase_Object ;
     TBase_Object = object( TObject )
                        private { Instance data... }
                             _Memory_Type : TMemory_Type ;
                             _Reference_Count : longint ;
                             _Properties : PSDynamic_Properties ;
{$IFDEF Debug}
                             Sanity : string[ 12 ] ;
{$ENDIF}

                        public { Constructors and destructors... }
                             constructor Init ;
                             destructor Done ; virtual ;
                             procedure Free ;

                        public { API... }
                             procedure Attach ; virtual ;
                             procedure Detach ; virtual ;
                             function Classname : string ; virtual ;
                             function Is_Class( const Name : string ) : boolean ;
                                 virtual ;
                             procedure Set_Memory_Type( MT : TMemory_Type ) ;
                             function Get_Memory_Type : TMemory_Type ;
                             function Reference_Count : integer ;

                             function Clone : PBase_Object ; virtual ;
                             function Properties : PSDynamic_Properties ;
                                 virtual ;
                             function Debugger : PSDebug_Interface ; virtual ;
                             procedure Test ; virtual ;

                             function Deserialize( Stream : PSStream ) : boolean ;
                                 virtual ;
                             function Serialize( Stream : PSStream ) : boolean ;
                                 virtual ;
                             function Load( Stream : PSStream ) : boolean ;
                                 virtual ;
                             function Persist( Stream : PSStream ) : boolean ;
                                 virtual ;
                    end ; { TBase_Object }

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Subroutine Library... }
     _ObjectF ; { Object_Factory }

{$IFDEF DELPHI}
{ TBase_Delphi_Object methods... }

{ Constructors and destructors... }

constructor TBase_Delphi_Object.Create ;

begin
    inherited Create ;

  {$IFDEF Debug}
    Sanity := 'TBase_Delphi_Object' ;
  {$ENDIF}
end ;


destructor TBase_Delphi_Object.Destroy ;

begin
  {$IFDEF Debug}
    Sanity := '                   ' ;
  {$ENDIF}
    if( _Properties <> nil ) then
    begin
        _Properties.Detach ;
    end ;

    inherited Destroy ;
end ;


{ API... }

procedure TBase_Delphi_Object.Attach ;

begin
    inc( _Reference_Count ) ;
end ;


procedure TBase_Delphi_Object.Detach ;

begin
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
        Free ;
    end ;
end ;


function TBase_Delphi_Object.Reference_Count : integer ;

begin
    Reference_Count := _Reference_Count ;
end ;


function TBase_Delphi_Object.Is_Class( const Name : string ) : boolean ;

begin
    try
        Is_Class := ( Name = Classname ) ;
    except
        Is_Class := False ;
    end ;
end ;


function TBase_Delphi_Object.Clone : TDelphi_Object ;

begin
    Clone := nil ;
end ;


function TBase_Delphi_Object.Properties : TDynamic_Properties ;

begin
    if( _Properties = nil ) then
    begin
        _Properties := Object_Factory^.Dynamic_Properties ;
    end ;
    Properties := _Properties ;
end ;


procedure TBase_Delphi_Object.Test ;

var I : integer ;

begin
    I := Reference_Count ;
    Attach ;
    if( I + 1 <> Reference_Count ) then
    begin
        writeln( 'Attach failed' ) ;
    end ;
    if( I = 0 ) then { No previous attachments }
    begin
        Attach ;
        inc( I ) ;
    end ;
    Detach ;
    if( I <> Reference_Count ) then
    begin
        writeln( 'Detach failed' ) ;
    end ;
end ;


function TBase_Delphi_Object.Debugger : TDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TBase_Delphi_Object.Deserialize( Stream : TStream ) : boolean ;

begin
    Deserialize := False ;
end ;


function TBase_Delphi_Object.Persist( Stream : TStream ) : boolean ;

begin
    Persist := False ;
end ;


function TBase_Delphi_Object.Serialize( Stream : TStream ) : boolean ;

begin
    Serialize := False ;
end ;


function TBase_Delphi_Object.Load( Stream : TStream ) : boolean ;

begin
    Load := False ;
end ;
{$ENDIF}



{ TObject methods... }

constructor TBase_Object.Init ;

begin
{$IFDEF BP}
    inherited Init ;
{$ENDIF}
    _Reference_Count := 0 ;
    _Memory_Type := MT_Unknown ;
    _Properties := nil ;
{$IFDEF Debug}
    Sanity := 'TBase_Object' ;
{$ENDIF}
end ;


destructor TBase_Object.Done ;

begin
{$IFDEF Debug}
    Sanity := '            ' ;
{$ENDIF}
    if( _Properties <> nil ) then
    begin
        _Properties^.Detach ;
    end ;

    inherited Done ;
end ;


procedure TBase_Object.Free ;

begin
    if( @self <> nil ) then
    begin
        Done ;
{$IFDEF DELPHI}
        if( _Memory_Type = MT_Heap ) then
        begin
            dispose( @self ) ;
        end ;
{$ENDIF}
    end ;
end ;


procedure TBase_Object.Attach ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    inc( _Reference_Count ) ;
end ;


procedure TBase_Object.Detach ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
{$IFDEF Debug}
        if( _Reference_Count < 0 ) then
        begin
            writeln( 'Overly detached object' ) ;
        end ;
{$ENDIF}
        Free ;
    end ;
end ;


function TBase_Object.Reference_Count : integer ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Reference_Count := _Reference_Count ;
end ;


function TBase_Object.Clone : PBase_Object ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Clone := nil ;
end ;


function TBase_Object.Properties : PSDynamic_Properties ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    if( _Properties = nil ) then
    begin
        _Properties := Object_Factory^.Static_Dynamic_Properties ;
    end ;
    Properties := _Properties ;
end ;


function TBase_Object.Classname : string ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Classname := 'TBase_Object' ;
end ;


function TBase_Object.Is_Class( const Name : string ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    if( Name = Classname ) then
    begin
        Is_Class := True ;
    end else
    begin
        Is_Class := inherited Is_Class( Name ) ;
    end ;
end ;


procedure TBase_Object.Set_Memory_Type( MT : TMemory_Type ) ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    _Memory_Type := MT ;
end ;


function TBase_Object.Get_Memory_Type : TMemory_Type ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Get_Memory_Type := _Memory_Type ;
end ;


function TBase_Object.Deserialize( Stream : PSStream ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Deserialize := False ;
end ;


function TBase_Object.Persist( Stream : PSStream ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Persist := False ;
end ;


function TBase_Object.Serialize( Stream : PSStream ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Serialize := False ;
end ;


function TBase_Object.Load( Stream : PSStream ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Load := False ;
end ;


procedure TBase_Object.Test ;

var I : integer ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    I := Reference_Count ;
    Attach ;
    if( I + 1 <> Reference_Count ) then
    begin
        writeln( 'Attach failed' ) ;
    end ;
    if( I = 0 ) then { No previous attachments }
    begin
        Attach ;
        inc( I ) ;
    end ;
    Detach ;
    if( I <> Reference_Count ) then
    begin
        writeln( 'Detach failed' ) ;
    end ;
end ;


function TBase_Object.Debugger : PSDebug_Interface ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Debugger := nil ;
end ;


{$IFDEF Test}
procedure Test_Unit ;

var TBDO : TBase_Delphi_Object ;
    TBO : TBase_Object ;

begin
  {$IFDEF Delphi}
    TBDO := TBase_Delphi_Object.Create ;
    TBDO.Test ;
    TBDO.Free ;
  {$ENDIF}

    TBO.Init ;
    TBO.Test ;
    TBO.Done ;
end ;
{$ENDIF}



end.

