{$N+}
{$E+}
{
        Program Name   : Express
        Package Name   : LB
        Purpose        : Expression evaluator
        Institution    : Conroy & Conroy Co.
        Date written   : 25-Aug-2002
        Written by     : Alan Conroy
        Version        : 1.0E

        Copyright(C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *******************************************************
        *                                                     *
        *      M O D I F I C A T I O N   H I S T O R Y        *
        *                                                     *
        *      DATE    BY      REASON                         *
        *                                                     *
        *  12-Oct-2002 EAC     Fix protection violation caused*
        *                      by trailing operator.          *
        *                      Fix handling of E-notation.    *
        *   2-Nov-2002 EAC     Handle division by 0.          *
        *                      Fix compilation problems in BP.*
        *                      Change Get_Symbol parameters to*
        *                      longint.                       *
        *                      Fix error in calculation of    *
        *                      HypArcCsc.                     *
        *                      Fix problems with handling E-  *
        *                      notation.                      *
        *                      Code cleanup.                  *
        *                      Fix returning of error code.   *
        *  30-Nov-2002 EAC     Properly handle an expression  *
        *                      which is simply a number.      *
        *                      Fixed unary minus on functions.*
        *                      Initialize strings for Delphi. *
        *  4-Apr-2006  EAC     Added SHR and SHL operators.   *
        *  6-Dec-2006  EAC     Fix bug that swallowed tokens  *
        *                      when processing a function name*
        *                      that ended with "E".           *
        *                                                     *
        *******************************************************

        *******************************************************
        *                                                     *
        *      P R O G R A M   P U R P O S E                  *
        *                                                     *
        *******************************************************

          This unit contains classes and routines to evaluate expressions.

          This code was culled from the Sirius compiler expression code.
}

unit Express ;

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF DELPHI}
     SysUtils, { strdispose }
{$ELSE}
     Strings, { strdispose }
{$ENDIF}

     { C&C... }
     Common, { TBase_Object }
     Parse ; { TSParser }

{ Symbol flags... }
const SF_Right_Unary_Operator = 1 ;
const SF_Left_Unary_Operator = 2 ;
const SF_Binary_Operator = 3 ;
const SF_Function = 4 ;

type tNode_type = (Node_Literal, Node_Function, Node_Operator) ;
     pExpression_Node = ^tExpression_Node ;
     tExpression_Node = record
                            Node_Type : tNode_Type ; { Class of node }
                            Operator : longint ; { Operator or function value }
                            Precedence : integer ; { Precedence of node }
                            Back, Left, Right : pExpression_Node ;
                            Literal : pChar ;
                        end ;

type TExpression_Evaluator = object( TBase_Object )
                                 private { Instance data... }
                                     _Error : integer ;
                                     Literal_Text : string ;
                                     Parser : TSParser ;
                                     _Position : integer ; { Position of error }
                                     Stack : array[ 0..1024 ] of extended ;
                                     Whole_Length : integer ;
                                     { Length of full expression to evaluate }

                                 public { Constructors and destructors... }
                                     constructor Init ;

                                 private { Internal utility routines... }
                                     function Expression : boolean ;
                                     { Generate code for an expression. }

                                     procedure Generate_Tree( X : pExpression_Node ) ;
                                     { Evaluate tree with root X. }

                                     function Get_Subexpression : pExpression_Node ;
                                     { Get a subexpression (term) and return
                                       pointer to subexpression tree }

                                     function Number : Boolean ;
                                     procedure Optimize_Tree( X : pExpression_Node ) ;
                                     function Parse_Expression : pExpression_Node ;
                                     { Parse an expression into an expression
                                       tree, and return pointer to root }

                                     procedure Zero_Expression_Tree( X : pExpression_Node ) ;

                                 public { API... }
                                     function Calculate( Operator : integer ) : boolean ;
                                         virtual ;
                                     function ERT( X : integer ) : string ;
                                         virtual ;
                                     function Evaluate( X : string ) : boolean ;
                                     function Function_Reference( Value : integer ) : boolean ;
                                         virtual ;
                                     { Process function parameters.
                                       Value = parameter count. }

                                     function Get_Position : integer ;
                                     { Position of error, if there was one. }

                                     function Get_Result : extended ;
                                     procedure Get_Symbol( X : string ;
                                         var Flags, Address, Value, _Result : longint ) ;
                                         virtual ;
                                     function Last_Error : integer ;
                                     function Pop : extended ;
                                     procedure Push( F : extended ) ;
                                     procedure Put_Token( const X : string ) ;
                                     function Token : string ;
                             end ; { TExpression_Evaluator }

function EERT( X : integer ) : string ;
procedure Eval( Expression : string ; var _Result, Imaginary : Extended ;
    var E, P : Integer ) ;

implementation

uses Maths, { Power }
     Num1s, { Num1s }
     UStrings ; { PString }

{ Operations... }
const Op_Negate_Stack = 1 ;
const Op_Add = 2 ;
const Op_Subtract = 3 ;
const Op_Multiply = 4 ;
const Op_Divide = 5 ;
const Op_Percent = 6 ;
const Op_Exponent = 7 ;
const Op_Factorial = 8 ;

const Op_Rad = 9 ;
const Op_Deg = 10 ;
const Op_Abs = 11 ;
const Op_Cos = 12 ;
const Op_Cot = 13 ;
const Op_Csc = 14 ;
const Op_Exp = 15 ;
const Op_Int = 16 ;
const Op_Log = 17 ;
const Op_Sec = 18 ;
const Op_Sin = 19 ;
const Op_Tan = 20 ;
const Op_Fix = 21 ;
const Op_Sqr = 22 ;
const Op_Sgn = 23 ;
const Op_Fib = 24 ;

const Op_Min = 25 ;
const Op_Max = 26 ;

const Op_Arccos = 27 ;
const Op_Arccot = 28 ;
const Op_Arccsc = 29 ;
const Op_Arcsec = 30 ;
const Op_Arcsin = 31 ;
const Op_Arctan = 32 ;
const Op_Hypcos = 33 ;
const Op_Hypcot = 34 ;
const Op_Hypcsc = 35 ;
const Op_Hypsec = 36 ;
const Op_Hypsin = 37 ;
const Op_Hyptan = 38 ;
const Op_Hyparccos = 39 ;
const Op_Hyparccot = 40 ;
const Op_HypArcCsc = 41 ;
const Op_Hyparcsec = 42 ;
const Op_Hyparcsin = 43 ;
const Op_Hyparctan = 44 ;
const Op_Log10 = 45 ;
const Op_Rnd = 46 ;

const Op_Mod = 47 ;
const Op_Rem = 48 ;
const Op_D = 49 ;
const Op_Div = 50 ;

const Op_Root = 51 ;
const Op_PI = 52 ;

const Op_Shift_Left = 53 ;
const Op_Shift_Right = 54 ;
const Op_And = 55 ;
const Op_Or = 56 ;
const Op_Xor = 57 ;

{ Errors... }
const Error_Expected_Comma = 1 ;
const Error_Syntax_Error = 2 ;
const Error_Illegal_Symbol = 3 ;
const Error_Expected_Left_Parenthesis = 4 ;
const Error_Division_By_Zero = 5 ;
const Error_Illegal_Number_for_factorial = 6 ;
const Error_Floating_point_error = 7 ;
const Error_Illegal_Number = 8 ;
const Error_Illegal_Argument = 9 ;
const Error_Imaginary_Number = 10 ;
const Error_Illegal_Argument_in_LOG = 11 ;
const Error_Argument_Too_large_in_EXP = 12 ;
const Error_Missing_Right_Parenthesis = 13 ;
const Error_Integer_To_Large = 14 ;

function EERT( X : integer ) : string ;

begin
    case X of
        Error_Expected_Comma: EERT := 'Expected comma' ;
        Error_Syntax_Error: EERT := 'Syntax error' ;
        Error_Illegal_Symbol: EERT := 'Illegal symbol' ;
        Error_Expected_Left_Parenthesis: EERT := 'Expected left parenthesis' ;
        Error_Division_By_Zero: EERT := 'Division by zero' ;
        Error_Illegal_Number_for_factorial: EERT := 'Illegal number for factorial' ;
        Error_Floating_point_error: EERT := 'Floating point error' ;
        Error_Illegal_Number: EERT := 'Illegal number' ;
        Error_Illegal_Argument: EERT := 'Illegal argument' ;
        Error_Imaginary_Number: EERT := 'Imaginary number' ;
        Error_Illegal_Argument_in_LOG: EERT := 'Illegal argument in LOG' ;
        Error_Argument_Too_large_in_EXP: EERT := 'Argument too large in EXP' ;
        Error_Missing_Right_Parenthesis: EERT := 'Missing right parenthesis' ;
    end ;
end ;



{ TExpression_Evaluator methods... }

{ Constructors and destructors... }

constructor TExpression_Evaluator.Init ;

begin
    inherited Init ;

{$IFDEF DELPHI}
    Initialize( Literal_Text ) ;
{$ENDIF}
end ;



{ API... }

function TExpression_Evaluator.ERT( X : integer ) : string ;

begin
    ERT := EERT( X ) ;
end ;


procedure TExpression_Evaluator.Get_Symbol( X : string ; var Flags, Address, Value, _Result : longint ) ;

begin
    _Result := 0 ; { Assume success }
    if( X = '+' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Add ;
        Value := 16000 ; { Precedence }
        exit ;
    end ;
    if( X = '-' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Subtract ;
        Value := 16000 ; { Precedence }
        exit ;
    end ;
    if( X = '*' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Multiply ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = '/' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Divide ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = '^' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Exponent ;
        Value := 26000 ; { Precedence }
        exit ;
    end ;
    if( X = 'SHL' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Shift_Left ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'SHR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Shift_Right ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'AND' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_And ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'OR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Or ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'XOR' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Xor ;
        Value := 24000 ; { Precedence }
        exit ;
    end ;
    if( X = 'MIN' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Min ;
        Value := 10000 ; { Precedence }
        exit ;
    end ;
    if( X = 'MAX' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Max ;
        Value := 10000 ; { Precedence }
        exit ;
    end ;
    if( X = '%' ) then
    begin
        Flags := SF_Left_Unary_Operator ;
        Address := Op_Percent ;
        Value := 28000 ;
        exit ;
    end ;
    if( X = '!' ) then
    begin
        Flags := SF_Left_Unary_Operator ;
        Address := Op_Factorial ;
        Value := 28000 ;
        exit ;
    end ;
    if( X = 'MOD' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Mod ; { Modulus }
        Value := 24000 ;
        exit ;
    end ;
    if( X = 'REM' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Rem ; { Remainder }
        Value := 24000 ;
        exit ;
    end ;
    if( X = 'D' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_D ;
        Value := 10000 ;
        exit ;
    end ;
    if( X = 'DIV' ) then
    begin
        Flags := SF_Binary_Operator ;
        Address := Op_Div ;
        Value := 24000 ;
        exit ;
    end ;

    if( X = 'ROOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Root ;
        Value := 2 ;
        exit ;
    end ;
    if( X = 'RAD' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Rad ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'RND' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Rnd ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'DEG' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Deg ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ABS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Abs ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'COS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Cos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'COT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Cot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'CSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Csc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'EXP' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Exp ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'INT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Int ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'LOG' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Log ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'TAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Tan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'FIX' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Fix ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SQR' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sqr ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'SGN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Sgn ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'FIB' ) then
    begin
        Flags := SF_Function ;
        Address := Op_FIB ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCOS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCCSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcCsc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCSEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcSec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCSIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcSin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'ARCTAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_ArcTan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCOS' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCos ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCOT' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCot ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPCSC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypCsc ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPSEC' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypSec ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPSIN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hypsin ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'HYPTAN' ) then
    begin
        Flags := SF_Function ;
        Address := Op_HypTan ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCCOS' ) or ( X = 'ARCHYPCOS' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparccos ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCCOT' ) or ( X = 'ARCHYPCOT' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparccot ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCSEC' ) or ( X = 'ARCHYPSEC' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparcsec ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCSIN' ) or ( X = 'ARCHYPSIN' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparcsin ;
        Value := 1 ;
        exit ;
    end ;
    if( ( X = 'HYPARCTAN' ) or ( X = 'ARCHYPTAN' ) ) then
    begin
        Flags := SF_Function ;
        Address := Op_Hyparctan ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'LOG10' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Log10 ;
        Value := 1 ;
        exit ;
    end ;
    if( X = 'PI' ) then
    begin
        Flags := SF_Function ;
        Address := Op_Pi ;
        Value := 0 ;
        exit ;
    end ;

    _Result := 1 ; { Error }
end ;


procedure TExpression_Evaluator.Push( F : extended ) ;

begin
    Stack[ 0 ] := Stack[ 0 ] + 1 ;
    Stack[ trunc( Stack[ 0 ] ) ] := F ;
end ;


function Cvt_String( X : PString ) : string ;

begin
    Cvt_String := X^.Text ;
    dispose( X, Done ) ;
end ;


function TExpression_Evaluator.Function_Reference( Value : integer ) : boolean ;
{ Process function parameters.  Value = parameter count. }

var F : integer ;
    G : longint ;
    Y : string ;

begin
    Function_Reference := False ;

    if( Value <> 0 ) then { if parameters }
    begin
        Y := Cvt_String( Parser.Token ) ;
        G := 0 ; { Number of parameters that code was generated for }
        if( Pos( Y, '({[' ) = 0 ) then { No parameters specified }
        begin
            Parser.Put_Token( Temp_String( Y ) ) ; { Restore token }
            if( Value > 0 ) then
            { One or more required parameters }
            begin
                _Error := Error_Expected_Left_Parenthesis ;
                exit ;
            end ;
        end else { Parenthesis found }
        begin
            F := 1 ; { Number of parameters }
            while( F <= abs( Value ) ) do { Until out of parameters }
            begin
                { Process next parameter... }
                _Error := 0 ;
                if( not Expression ) then
                begin
                    exit ;
                end ;
                Inc( G ) ;
                Y := Cvt_String( Parser.Token ) ; { Get next token }
                if(
                    ( Y <> ',' )
                    and
                    ( F <= abs( Value ) )
                  ) then { No more parameters }
                begin
                    if( Pos( Y, '}])' ) = 0 ) then
                    begin
                        Parser.Put_Token( Temp_String( Y ) ) ;
                    end ;
                    if( Value < 0 ) then { Variable number of parameters }
                    begin
                        F := abs( Value ) + 1 ; { Get actual number }
                    end else
                    if( F < Value ) then { There should be more parameters }
                    begin
                        _Error := Error_Expected_Comma ;
                        exit ;
                    end ;
                end ;
                Inc( F ) ; { Increment parameter count }
            end ;
        end ;
        if( Value < 0 ) then { Variable number of parameters - push count of them }
        begin
            Push( G ) ;
        end ;
        if( ( Pos( Y, ']})' ) = 0 ) and ( G > 0 ) ) then { No parenthesis when needed }
        begin
            _Error := Error_Missing_right_parenthesis ;
            exit ;
        end ;
    end ; { if Last_Symbol^.Size <> 0 }
    Function_Reference := True ;
end ; { Function_Reference }


procedure TExpression_Evaluator.Zero_Expression_Tree( X : pExpression_Node ) ;
{ Zero storage stored at tree with root X }

var Current, Previous : pExpression_Node ;

begin
    { Start at root }
    Current := X ;
    while( Current <> nil ) do
    begin
        { Go all the way left }
        while( Current^.Left <> nil ) do
        begin
            Current := Current^.Left ;
        end ;
        if( Current^.Right <> nil ) then
        begin
            Current := Current^.Right ;
        end else
        begin
            Previous := Current^.Back ;
            if( Previous <> nil ) then
            begin
                if( Previous^.Left = Current ) then
                begin
                    Previous^.Left := nil ;
                end else
                begin
                    Previous^.Right := nil ;
                end ;
            end ;
            case Current^.Node_Type of
                Node_Literal : strdispose( Current^.Literal ) ;
            end ;
            dispose( Current ) ;
            Current := Previous ;
        end ;
    end ; { while( Current <> nil ) do }
end ; { Zero_Expression_Tree }


procedure TExpression_Evaluator.Optimize_Tree( X : pExpression_Node ) ;
{ Optimize an expression tree. }

var Dummy : integer ;
    L, R : string ;
    LX, RX, Temp : extended ;
{$IFDEF Delphi}
    LI, RI : int64 ;
{$ELSE}
    LI, RI : longint ;
{$ENDIF}
    Res : integer ;

begin
    { Optimize sub-trees }
    if( X^.Left <> nil ) then
    begin
        Optimize_Tree( X^.Left ) ;
    end ;
    if( X^.Right <> nil ) then
    begin
        Optimize_Tree( X^.Right ) ;
    end ;

    { See if we can optimize... }
    if( X^.Node_Type <> Node_Operator ) then
    begin
        exit ; { Can't optimize this any further }
    end ;
    LX := 0 ; { Avoid compiler warning }
    RX := 0 ; { Ditto }
    if( X^.Left <> nil ) then
    begin
        if( X^.Left^.Node_Type <> Node_Literal ) then
        begin
            exit ;
        end ;
        L := strpas( X^.Left^.Literal ) ;
        val( L, LX, Dummy ) ;
    end ;
    if( X^.Right <> nil ) then
    begin
        if( X^.Right^.Node_Type <> Node_Literal ) then
        begin
            exit ;
        end ;
        R := strpas( X^.Right^.Literal ) ;
        val( R, RX, Dummy ) ;
    end ;

    case X^.Operator of
        Op_Negate_Stack : LX := -LX ;
        Op_Add : LX := LX + RX ;
        Op_Subtract : LX := LX - RX ;
        Op_Multiply : LX := LX * RX ;
        Op_Divide :
            begin
                if( RX = 0.0 ) then
                begin
                    _Error := Error_Division_By_Zero ;
                end else
                begin
                    LX := LX / RX ;
                end ;
            end ;
        Op_Shift_Right :
            begin
                if( RX >= 1 ) then
                begin
                    Temp := 2 ;
                    while( RX >= 2 ) do
                    begin
                        RX := RX - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    LX := LX / Temp ;
                end ;
            end ;
        Op_Shift_Left :
            begin
                if( RX >= 1 ) then
                begin
                    Temp := 2 ;
                    while( RX >= 2 ) do
                    begin
                        RX := RX - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    LX := LX * Temp ;
                end ;
            end ;
        Op_And :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI and RI ;
                    LX := LI ;
                end ;
            end ;
        Op_Or :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI or RI ;
                    LX := LI ;
                end ;
            end ;
        Op_Xor :
            begin
                if(
                    ( int( LX ) > high( LI ) )
                    or
                    ( int( LX ) < low( LI ) )
                    or
                    ( int( RX ) > high( RI ) )
                    or
                    ( int( RX ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                end else
                begin
                    LI := trunc( int( LX ) ) ;
                    RI := trunc( int( RX ) ) ;
                    LI := LI xor RI ;
                    LX := LI ;
                end ;
            end ;
        Op_Percent : LX := LX / 100.0 ;
        Op_Exponent :
            begin
                LX := Power( LX, RX, Res ) ;
                _Error := Res ;
            end ;
        Op_Factorial : LX := Factorial( LX ) ;
        Op_Max :
            begin
                if( LX < RX ) then LX := RX ;
            end ;
        Op_Min :
            begin
                if( LX > RX ) then LX := RX ;
            end ;
        else exit ;
    end ; { case X.Operator of }

    L := num2( LX ) ;
    L := L + #0 ;
    X^.Literal := strnew( @L[ 1 ] ) ;
    X^.Node_Type := Node_Literal ;
    if( X^.Left <> nil ) then
    begin
        strdispose( X^.Left^.Literal ) ;
        dispose( X^.Left ) ;
        X^.Left := nil ;
    end ;
    if( X^.Right <> nil ) then
    begin
        strdispose( X^.Right^.Literal ) ;
        dispose( X^.Right ) ;
        X^.Right := nil ;
    end ;
end ; { Optimize_Tree }


function TExpression_Evaluator.Pop : extended ;

begin
    Pop := Stack[ trunc( Stack[ 0 ] ) ] ;
    Stack[ 0 ] := Stack[ 0 ] - 1 ;
end ;


function TExpression_Evaluator.Calculate( Operator : integer ) : boolean ;

var Dummy : integer ;
    A1, A2, A3, E1, E2, Temp : extended ;
{$IFDEF Delphi}
    LI, RI : int64 ;
{$ELSE}
    LI, RI : longint ;
{$ENDIF}
    Res : integer ;

begin
    Calculate := False ; { Assume no errors }
    case Operator of
        Op_Negate_Stack :
            begin
                Push( -Pop ) ;
            end ;
        Op_Add :
            begin
                Push( Pop + Pop ) ;
            end ;
        Op_Subtract :
            begin
                E2 := Pop ;
                E1 := Pop ;
                Push( E1 - E2 ) ;
            end ;
        Op_Multiply :
            begin
                Push( Pop * Pop ) ;
            end ;
        Op_Divide :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0.0 ) then
                begin
                    _Error := Error_Division_By_Zero ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 / E2 ) ;
            end ;
        Op_Shift_Left :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 >= 1 ) then
                begin
                    Temp := 2 ;
                    while( E2 >= 2 ) do
                    begin
                        E2 := E2 - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    Push( E1 * Temp ) ;
                end ;
            end ;
        Op_Shift_Right :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 >= 1 ) then
                begin
                    Temp := 2 ;
                    while( E2 >= 2 ) do
                    begin
                        E2 := E2 - 1 ;
                        Temp := Temp + Temp ;
                    end ;
                    Push( E1 / Temp ) ;
                end ;
            end ;
        Op_And :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI and RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Or :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI or RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Xor :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if(
                    ( int( E1 ) > high( LI ) )
                    or
                    ( int( E1 ) < low( LI ) )
                    or
                    ( int( E2 ) > high( RI ) )
                    or
                    ( int( E2 ) < low( RI ) )
                  ) then
                begin
                    _Error := Error_Integer_To_Large ;
                    Calculate := True ;
                    exit ;
                end ;
                LI := trunc( int( E1 ) ) ;
                RI := trunc( int( E2 ) ) ;
                LI := LI xor RI ;
                E1 := LI ;
                Push( E1 ) ;
            end ;
        Op_Percent :
            begin
                Push( Pop / 100.0 ) ;
            end ;
        Op_Exponent :
            begin
                E2 := Pop ;
                E1 := Pop ;
                E1 := Power( E1, E2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 ) ;
            end ;
        Op_Factorial :
            begin
                Push( Factorial( Pop ) ) ;
            end ;
        Op_Max :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E1 > E2 ) then
                begin
                    E1 := E2 ;
                end ;
                Push( E1 ) ;
            end ;
        Op_Min :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E1 < E2 ) then
                begin
                    E1 := E2 ;
                end ;
                Push( E1 ) ;
            end ;

        Op_Rad :
            begin
                Push( Pop / ( 180. / Pi ) ) ;
            end ;
        Op_Deg :
            begin
                Push( Pop * ( 180. / Pi ) ) ;
            end ;
        Op_Abs :
            begin
                Push( Abs( Pop ) ) ;
            end ;
        Op_Cos :
            begin
                Push( Cos( Pop ) ) ;
            end ;
        Op_Cot :
            begin
                Push( 1 / Tan( Pop ) ) ;
            end ;
        Op_Csc :
            begin
                Push( 1. / Sin( Pop ) ) ;
            end ;
        Op_Exp :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Exp( E1 ) ) ;
            end ;
        Op_Int :
            begin
                Push( Int( Pop ) ) ;
            end ;
        Op_Log :
            begin
                E1 := Pop ;
                if( E1 <= 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_Sec :
            begin
                Push( 1. / Cos( Pop ) ) ;
            end ;
        Op_Sin :
            begin
                Push( Sin( Pop ) ) ;
            end ;
        Op_Tan :
            begin
                Push( Tan( Pop ) ) ;
            end ;
        Op_Fix :
            begin
                Push( Trunc( Pop ) ) ;
            end ;
        Op_Sqr :
            begin
                E1 := Pop ;
                if( E1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Sqrt( E1 ) ) ;
            end ;
        Op_Sgn :
            begin
                Push( Sgn( Pop ) ) ;
            end ;
        Op_Fib :
            begin
                E1 := Pop ;
                if( E1 > 182 ) then { No fibonacci > 182 }
                begin
                    _Error := 7 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Int( Abs( E1 ) ) ;
                if( E1 < 2 ) then
                begin
                    Push( 1 ) ;
                end else
                begin
                    A1 := 1 ;
                    A2 := 1 ;
                    for Dummy := 2 to Trunc( E1 ) do
                    begin
                        A3 := A2 ;
                        A2 := A1 ;
                        A1 := A2 + A3
                    end ;
                    Push( A1 + A2 ) ;
                end
            end ;
        Op_Arccos :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 - A1 ;
                if( A1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( Sqrt( A1 ) / E1 ) ) ;
            end ;
        Op_ArcCot :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( 1. / E1 ) ) ;
            end ;
        Op_ArcCsc :
            begin
                E1 := Power( Pop, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 - 1 ;
                if( E1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( 1. / Sqrt( E1 ) ) ) ;
            end ;
        Op_ArcSec :
            begin
                E1 := Power( Pop, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 - 1 ;
                if( E1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Arctan( Sqrt( E1 ) ) ) ;
            end ;
        Op_Arcsin :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 - A1 ;
                if( A1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                if( A1 = 0 ) then
                begin
                    _Error := 5 ;
                    exit ;
                end ;
                Push( Arcsin( E1 ) ) ;
            end ;
        Op_ArcTan :
            begin
                Push( Arctan( Pop ) ) ;
            end ;
        Op_Hypcos :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Hypcos( E1 ) ) ;
            end ;
        Op_HypCot :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Exp( E1 ) - Exp( -E1 ) ;
                if( A1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypCot( E1 ) ) ;
            end ;
        Op_Hypcsc :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Exp( E1 ) - Exp( -E1 ) ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( 2. / E1 ) ;
            end ;
        Op_HypSec :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Exp( E1 ) + Exp( -E1 ) ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( 2. / E1 ) ;
            end ;
        Op_HypSin :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypSin( E1 ) ) ;
            end ;
        Op_HypTan :
            begin
                E1 := Pop ;
                if( ( E1 > 88 ) or ( E1 < -88 ) ) then
                begin
                    _Error := 12 ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Exp( E1 ) + Exp( -E1 ) ;
                if( A1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypTan( E1 ) ) ;
            end ;
        Op_HypArcCos :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := A1 - 1 ;
                if( A1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 + Sqrt( A1 ) ;
                if( E1 <= 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_HypArcCot :
            begin
                E1 := Pop ;
                if( E1 - 1 <= 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcCot( E1 ) ) ;
            end ;
        Op_HypArcCsc :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcCsc( E1 ) ) ;
            end ;
        Op_HypArcSec :
            begin
                E1 := Pop ;
                if( E1 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := Power( E1, 2, Res ) ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                A1 := 1 / A1 - 1 ;
                if( A1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcSec( E1 ) ) ;
            end ;
        Op_HypArcSin :
            begin
                E1 := Pop ;
                A1 := Power( E1, 2, Res ) + 1 ;
                if( Res <> 0 ) then
                begin
                    _Error := Res ;
                    Calculate := True ;
                    exit ;
                end ;
                if( A1 < 0 ) then
                begin
                    _Error := 10 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 + Sqrt( A1 ) ;
                if( E1 <= 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Ln( E1 ) ) ;
            end ;
        Op_HypArcTan:
            begin
                E1 := Pop ;
                if( abs( E1 ) >= 1 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( HypArcTan( E1 ) ) ;  
            end ;
        Op_Log10 :
            begin
                E1 := Pop ;
                if( E1 <= 0 ) then
                begin
                    _Error := 11 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Log10( E1 ) ) ;
            end ;
        Op_PI :
            begin
                Push( Pi ) ;
            end ;
        Op_Rnd :
            begin
                Push( random( trunc( Pop ) ) ) ;
            end ;
        Op_Root :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 <> int( E2 ) ) then
                begin
                    _Error := Error_Illegal_argument ;
                    Calculate := True ;
                    exit ;
                end ;
                if( E1 < 0 ) then { Negative value }
                begin
                    if( ( trunc( E2 ) and 1 ) = 0 ) then { Even number }
                    begin
                        _Error := Error_Imaginary_number ;
                        Calculate := True ;
                        exit ;
                    end ;
                end ;
                Push( Root( E1, trunc( E2 ) ) ) ;
            end ;

        Op_Mod :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( E1 - Int( E1 / E2 ) * E2 ) ;
            end ;
        Op_Rem :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                E1 := E1 / E2 ;
                Push( E1 - Trunc( E1 ) ) ;
            end ;
        Op_Div :
            begin
                E2 := Pop ;
                E1 := Pop ;
                if( E2 = 0 ) then
                begin
                    _Error := 5 ;
                    Calculate := True ;
                    exit ;
                end ;
                Push( Int( Int( E1 ) / Int( E2 ) ) ) ;
            end ;
        Op_D :
            begin
                E2 := Pop ;
                E1 := Pop ;
                A2 := 0 ;
                while( E1 > 0 ) do
                begin
                    A2 := A2 + random( trunc( E2 ) ) + 1 ;
                    E1 := E1 - 1 ;
                end ;
                Push( A2 ) ;
            end ;
    end ;
end ; { TExpression_Evaluator.Calculate }


procedure TExpression_Evaluator.Generate_Tree( X : pExpression_Node ) ; { Evaluate tree with root X. }

var Current, Previous : pExpression_Node ;
    Dummy : integer ;
    F : extended ;
    S : string ;

begin
    { Start at root }
    Current := X ;
    while( Current <> nil ) do
    begin
        { Go all the way left }
        while( Current^.Left <> nil ) do
        begin
            Current := Current^.Left ;
        end ;

        { Go right if no more left }
        if( Current^.Right <> nil ) then
        begin
            Current := Current^.Right ;
        end else
        begin
            { Process terminal node }
            Previous := Current^.Back ;
            if( Previous <> nil ) then
            begin
                if( Previous^.Left = Current ) then { On left of parent }
                begin
                    Previous^.Left := nil ;
                end else
                begin
                    Previous^.Right := nil ;
                end ;
            end ;
            case Current^.Node_Type of
                Node_Operator, Node_Function :
                    begin
                        if( Calculate( Current^.Operator ) ) then
                        begin
                            Stack[ 0 ] := 0 ;
                            exit ;
                        end ;
                    end ;
                Node_Literal : begin
                                   S := strpas( Current^.Literal ) ;
                                   val( S, F, Dummy ) ;
                                   Push( F ) ;
                                   strdispose( Current^.Literal ) ;
                               end ;
            end ;
            dispose( Current ) ;
            Current := Previous ;
        end ;
    end ; { while( Current <> nil ) do }
end ; { Generate_Tree }


function TExpression_Evaluator.Number : Boolean ;

var A : extended ;
    B : Integer ;
    Negate : boolean ;
    S, Saved : string ;

Label Loop ;

begin
    Saved := Parser.Non_Terminals.Text ;
    {Parser.Non_Terminals := '0123456789.E' ;}
    Literal_Text := Cvt_String( Parser.Token ) ;
    Negate := False ;
    while( Literal_Text = '+' ) do { Eat unary plus }
    begin
        Literal_Text := Cvt_String( Parser.Token ) ;
    end ;
    while( Literal_Text = '-' ) do { Accumulate unary minus }
    begin
        Negate := not Negate ;
        Literal_Text := Cvt_String( Parser.Token ) ;
    end ;
    if( Literal_Text = '.' ) then
    begin
        Literal_Text := '0.' ;
    end ;
    if( copy( Literal_Text, length( Literal_Text ), 1 ) = 'E' ) then
    begin
        S := Cvt_String( Parser.Token ) ;
        if( ( S = '-' ) or ( S = '+' ) ) then
        begin
            Literal_Text := Literal_Text + S + Cvt_String( Parser.Token ) ;
        end else
        begin
            Parser.Put_Token( Temp_String( S ) ) ;
        end ;
    end ;
Loop:
{$IFDEF DELPHI}
  {$HINTS OFF}
{$ENDIF}
    val( Literal_Text, A, B ) ; { See if a valid numer }
{$IFDEF DELPHI}
  {$HINTS ON}
{$ENDIF}
    if( B = 0 ) then { So far, so good }
    begin
        if( not Parser.Token_EOL ) then { End of line is end of token }
        begin
            B := length( Literal_Text ) ;
            Literal_Text := Literal_Text + Cvt_String( Parser.Token ) ; { See if some more of a numeric specification }
            if( copy( Literal_Text, length( Literal_Text ), 1 ) = 'E' ) then
            begin
                S := Cvt_String( Parser.Token ) ;
                if( ( S = '-' ) or ( S = '+' ) ) then
                begin
                    Literal_Text := Literal_Text + S + Cvt_String( Parser.Token ) ;
                end ;
            end ;
            if( B <> length( Literal_Text ) ) then { Source file didn't end }
            begin
                goto Loop ;
            end ;
        end ;
    end ;
    if( copy( Literal_Text, 1, 1 ) = 'E' ) then
    begin
        B := 2 ;
    end ;
    if( ( B > 1 ) and ( Literal_Text[ B - 1 ] = 'E' ) ) then
    begin
        dec( B ) ;
    end ;
    if( B > 0 ) then
    begin
        Parser.Put_Token( Temp_String( Copy( Literal_Text, B, Length( Literal_Text ) - B + 1 ) ) ) ;
    end ;
    if( B = 1 ) then
    begin
        Number := False ;
    end else
    begin
        if( B > 0 ) then
        begin
            Literal_Text := Copy( Literal_Text, 1, B - 1 ) ;
        end ;
        if( Negate ) then
        begin
            Literal_Text := '-' + Literal_Text ;
        end ;
        Number := True ;
    end ;
    Parser.Non_Terminals.Assign( Temp_String( Saved ) ) ;
end ; { TExpression_Evaluator.Number }


function TExpression_Evaluator.Get_Subexpression : pExpression_Node ;
{ Get a subexpression (term) and return pointer to subexpression tree }

label Loop, Operator_Loop ;

var Subexpression, Temp : pExpression_Node ;
    Operator : pExpression_Node ;
    X : string ;
    AA : array[ 0..255 ] of char ;
    Address : longint ;
    Flags : longint ;
    _Result, Value : longint ;

begin
    Subexpression := nil ;
Loop:
    X := Cvt_String( Parser.Token ) ; { Get next token }
    if( length( X ) = 0 ) then
    begin
        _Error := Error_Syntax_Error ;
        Get_Subexpression := Subexpression ;
        exit ;
    end ;
    if( pos( X[ 1 ], '({[' ) > 0 ) then { Start of parenthesis}
    begin
        Temp := Parse_Expression ;
        if( Subexpression = nil ) then { First part of subexpression }
        begin
            Subexpression := Temp ;
        end else
        begin
            Operator := Subexpression ;
            while( Operator^.Left <> nil ) do
            begin
                Operator := Operator^.Left ;
            end ;
            Operator^.Left := Subexpression ;
        end ;
        X := Cvt_String( Parser.Token ) ;
        if( ( X = '' ) or ( pos( X[ 1 ], ')}]' ) = 0 ) ) then { No closing parenthesis}
        begin
            Parser.Put_Token( Temp_String( X ) ) ;
            _Error := Error_Missing_Right_Parenthesis ;
            Get_Subexpression := nil ;
            Zero_Expression_Tree( Subexpression ) ;
            exit ;
        end ;
        goto Operator_Loop ;
    end ;

    if( X = '+' ) then { Unary plus }
    begin
        goto Loop ; { Ignore these }
    end ;

    if( X = '-' ) then { Unary minus }
    begin
        Temp := Get_Subexpression ; { Get subexpression }
        if( Temp = nil ) then
        begin
            _Error := Error_Syntax_Error ;
            Get_Subexpression := Subexpression ;
            exit ;
        end ;
        new( Operator ) ;
        Operator^.Back := nil ;
        Operator^.Left := Temp ;
        Operator^.Right := nil ;
        Operator^.Node_Type := Node_Operator ;
        Operator^.Precedence := 28000 ;
        Operator^.Operator := Op_Negate_Stack ;
        Temp^.Back := Operator ;
        if( Subexpression = nil ) then { Nothing in tree yet }
        begin
            Subexpression := Operator ;
        end else
        begin
            Temp := Subexpression ;
            while( Temp^.Left <> nil ) do
            begin
                Temp := Temp^.Left ;
            end ;
            Temp^.Left := Operator ;
            Operator^.Back := Temp ;
        end ;
        goto Operator_Loop ;
    end ;

    Parser.Put_Token( Temp_String( X ) ) ;
    if( Number ) then { Numeric literal }
    begin
        new( Temp ) ;
        Temp^.Node_Type := Node_Literal ;
        X := Literal_Text ;
        Temp^.Left := nil ;
        Temp^.Right := nil ;
        Temp^.Back := nil ;
        Temp^.Literal := strnew( strpCopy( AA, Literal_Text ) ) ;
        Subexpression := Temp ;
        goto Operator_Loop ;
    end ;

    { Must be a variable, function, or an error }
    X := Cvt_String( Parser.Token ) ;
    Get_Symbol( X, Flags, Address, Value, _Result ) ;
    if( _Result <> 0 ) then
    begin
        _Error := Error_Syntax_Error ;
        Get_Subexpression := Subexpression ;
        exit ;
    end ;
    if( ( Flags and SF_Binary_Operator ) = SF_Right_Unary_Operator ) then
    begin
        new( Operator ) ;
        Operator^.Back := nil ;
        Operator^.Left := nil ;
        Operator^.Node_Type := Node_Operator ;
        Operator^.Precedence := Value ;
        Operator^.Operator := Address ;
        Temp := Get_Subexpression ; { Get subexpression }
        if( Temp = nil ) then
        begin
            _Error := Error_Syntax_Error ;
            Get_Subexpression := Subexpression ;
            exit ;
        end ;
        Operator^.Right := Temp ;
        Temp^.Back := Operator ;
        if( Subexpression = nil ) then { Nothing in tree yet }
        begin
            Subexpression := Operator ;
        end else
        begin
            Temp := Subexpression ;
            while( Temp^.Left <> nil ) do
            begin
                Temp := Temp^.Left ;
            end ;
            Temp^.Left := Operator ;
            Operator^.Back := Temp ;
        end ;
        goto Operator_Loop ;
    end ;
    if( ( Flags and SF_Function ) <> 0 ) then
    { Function reference.  Function references have two pieces of code:
      1) invocation of the function (including the call, and creation of
         temporary string results), and 2) parameter passing. }
    begin
        { Generate parameter passing code }
        Function_Reference( Value ) ; { Process function parameters }

        { Create new expression node for function call... }
        new( Temp ) ;
        Temp^.Node_Type := Node_Function ;
        Temp^.Left := nil ;
        Temp^.Right := nil ;
        Temp^.Back := nil ;
        Temp^.Operator := Address ;
        Subexpression := Temp ;
        goto Operator_Loop ;
    end ;
    Parser.Put_Token( Temp_String( X ) ) ;
    X := Cvt_String( Parser.Token ) ; { Pull back from token stream }

    new( Temp ) ;
    Temp^.Node_Type := Node_Function ;
    Temp^.Back := nil ;
    Temp^.Left := nil ;
    Temp^.Right := nil ;
    Temp^.Operator := Address ;
    if( Subexpression = nil ) then { Nothing in tree yet }
    begin
        Subexpression := Temp ;
    end else
    begin
        Operator := Subexpression ;
        while( Operator^.Left <> nil ) do
        begin
            Operator := Operator^.Left ;
        end ;
        Operator^.Left := Temp ;
        Temp^.Back := Operator ;
    end ;

    { Process all left-unary operators }
Operator_Loop:
    X := Cvt_String( Parser.Token ) ;
    if( length( X ) > 0 ) then
    begin
        Get_Symbol( X, Flags, Address, Value, _Result ) ;
        if(
            ( _Result = 0 )
            and
            ( ( Flags and SF_Binary_Operator ) = SF_Left_Unary_Operator )
          ) then { Left-unary operator }
        begin
            { Append unary operator to subexpression tree }
            new( Operator ) ;
            Operator^.Precedence := Value ;
            Operator^.Left := Subexpression ;
            Operator^.Right := nil ;
            Operator^.Back := nil ;
            Subexpression^.Back := Operator ;
            Operator^.Node_Type := Node_Operator ;
            Operator^.Operator := Address ;
            Subexpression := Operator ;
            goto Operator_Loop ; { Go back for next unary operator }
        end ;
        Parser.Put_Token( Temp_String( X ) ) ; { Push token }
    end ;

    Get_Subexpression := Subexpression ;
end ; { TExpression_Evaluator.Get_Subexpression }


function TExpression_Evaluator.Parse_Expression : pExpression_Node ;
{ Parse an expression into an expression tree, and return pointer to root }

label Operator_Loop ;

var Operator_Precedence : integer ;
    Operator : string ;
    Left_Subexpression, Current, Previous, Root : pExpression_Node ;
    Flags, Address, Value : longint ;
    _Result : longint ;

begin
    { Perform expression setup }
    Current := nil ;
    Root := nil ;

    { Get first subexpression }
    Left_Subexpression := Get_Subexpression ;

    { Process until end of expression }
    while( Left_Subexpression <> nil ) do
    begin
Operator_Loop:
        { Get operator }
        Operator := Cvt_String( Parser.Token ) ;
        if( length( Operator ) = 0 ) then
        begin
            _Result := 1 ;
            Flags := 0 ;
        end else
        begin
            Get_Symbol( Operator, Flags, Address, Value, _Result ) ;
        end ;
        if(
            ( _Result <> 0 )
            or
            ( ( Flags and SF_Binary_Operator ) = 0 )
          ) then { Invalid operator }
        begin
            Parser.Put_Token( Temp_String( Operator ) ) ; { Return token }
            if( Current <> nil ) then
            begin
                Current^.Right := Left_Subexpression ;
                Left_Subexpression^.Back := Current ;
            end else
            begin
                Root := Left_Subexpression ;
            end ;
            _Error := Error_Syntax_Error ;
            Parse_Expression := Root ;
            exit ;
        end ;
        if( ( Flags and SF_Binary_Operator ) <> SF_Binary_Operator ) then
        { Unary operator or missing binary operator }
        begin
            if( ( Flags and SF_Binary_Operator ) <> SF_Left_Unary_Operator ) then
            { Left-unary is OK }
            begin
                Parser.Put_Token( Temp_String( Operator ) ) ; { Return token }
                if( Current = nil ) then
                begin
                    Parse_Expression := Left_Subexpression ;
                end else
                begin
                    Current^.Right := Left_Subexpression ;
                    Left_Subexpression^.Back := Current ;
                    Parse_Expression := Root ;
                end ;
                _Error := Error_Syntax_Error ;
                exit ;
            end ;
        end ;

        { Create node for operator }
        if( Root = nil ) then { First thing in tree }
        begin
            new( Root ) ;
            Current := Root ;
            Previous := nil ;
        end else
        begin
            Previous := Current ; { Save previous current node }
            new( Current ) ;
        end ;
        Current^.Precedence := Value ;
        Current^.Left := Left_Subexpression ;
        Current^.Right := nil ;
        Current^.Back := Previous ;
        Current^.Node_Type := Node_Operator ;
        Current^.Operator := Address ;
        Left_Subexpression^.Back := Current ;

        { Insert operator node into existing tree }
        if( Previous <> nil ) then { Tree exists }
        begin
            if( Previous^.Left = nil ) then
            begin
                Previous^.Left := Current ;
            end else
            begin
                Previous^.Right := Current ;
            end ;
        end ; { if( Previous <> nil ) then }

        { Handle precedence }
        if( _Result = 0 ) then
        begin
            Operator_Precedence := Value ;
            if( ( Current^.Back <> nil ) and { There was a previous operator }
                ( Current^.Back^.Precedence >= Operator_Precedence )
                { Current operator has less precedence than the last
                  (Operators with same precedence evaluate from left to right,
                   thus, the following has an implied lower precedence) }
               ) then
            begin
                { Back up to the proper node }
                Previous := Current ;
                while(
                       ( Previous^.Back <> nil )
                       and
                       ( Previous^.Back^.Precedence >= Operator_Precedence )
                     ) do
                begin
                    Previous := Previous^.Back ;
                end ;

                { Move current node to proper location in tree... }
                { Unlink from current location }
                if( Current^.Back^.Left = Current ) then
                begin
                    Current^.Back^.Left := Current^.Left ;
                end else
                begin
                    Current^.Back^.Right := Current^.Left ;
                end ;
                if( Current^.Left <> nil ) then
                begin
                    Current^.Left^.Back := Current^.Back ;
                end ;
                Current^.Back := Previous^.Back ;

                { Relink to new location }
                if( Previous = Root ) then { Current becomes root node }
                begin
                    Current^.Left := Root ;
                    Root^.Back := Current ;
                    Root := Current ;
                    Current^.Back := nil ;
                end else
                begin
                    if( Previous^.Back^.Left = Previous ) then
                    begin
                        Previous^.Back^.Left := Current ;
                    end else
                    begin
                        Previous^.Back^.Right := Current ;
                    end ;
                    Current^.Left := Previous ;
                    Previous^.Back := Current ;
                end ;
            end ; { Current operator has less precedence than the last }
        end ; { if( _Result = 0 ) }

        { Get next subexpression and loop }
        Left_Subexpression := Get_Subexpression ;
    end ; { while( Left_Subexpression <> nil ) do }
    Parse_Expression := Root ;
end ; { TExpression_Evaluator.Parse_Expression }


function TExpression_Evaluator.Expression : boolean ; { Generate code for an expression. }

var _Result : pExpression_Node ;

begin
    randomize ;
    _Result := Parse_Expression ;
    Expression := ( _Result <> nil  ) ;
    if( _Result <> nil ) then
    begin
        { Optimize the expression tree }
        _Error := 0 ;
        Optimize_Tree( _Result ) ;
        if( _Error <> 0 ) then { Some error during optimization }
        begin
            exit ;
        end ;

        { Generate code }
        Generate_Tree( _Result ) ;
    end ;
end ; { TExpression_Evaluator.Expression }


function TExpression_Evaluator.Evaluate( X : string ) : boolean ;

var Dummy : integer ;
    E : extended ;

begin
    { Setup... }
    Stack[ 0 ] := 0 ; { Clear stack }
    _Error := 0 ;

    val( X, E, Dummy ) ;
    if( Dummy = 0 ) then { A numeric constant }
    begin
        Push( E ) ;
        Evaluate := True ;
        exit ;
    end ;

    Parser.Init ;
    Parser.Put_Token( Temp_String( X ) ) ;
    Whole_Length := length( X ) ;
    Evaluate := Expression ;
    if(
        ( _Error = 0 )
        and
        ( ( Cvt_String( Parser.Token ) <> '' ) or ( Stack[ 0 ] <> 1 ) )
      ) then { Left-over unparsable text }
    begin
        _Error := Error_Syntax_Error ;
    end ;
    if( _Error <> 0 ) then
    begin
        Evaluate := False ;
    end ;
    _Position := Whole_Length - length( Cvt_String( Parser.Grab_Line ) ) ;
    Parser.Done ;
end ; { TExpression_Evaluator.Evaluate }


function TExpression_Evaluator.Get_Result : extended ;

begin
    Get_Result := Stack[ 1 ] ;
end ;


function TExpression_Evaluator.Last_Error : integer ;

begin
    Last_Error := _Error ;
end ;


function TExpression_Evaluator.Get_Position : integer ;

begin
    Get_Position := _Position ;
end ;


procedure TExpression_Evaluator.Put_Token( const X : string ) ;

begin
    Parser.Put_Token( Temp_String( X ) ) ;
end ;


function TExpression_Evaluator.Token : string ;

begin
    Token := Cvt_String( Parser.Token ) ;
end ;


procedure Eval( Expression : string ; var _Result, Imaginary : Extended ;
    var E, P : Integer ) ;

var Evaluator : TExpression_Evaluator ;

begin
    Imaginary := 0 ;
    Evaluator.Init ;
    if( Evaluator.Evaluate( Expression ) ) then
    begin
        _Result := Evaluator.Get_Result ;
        E := 0 ;
        P := 0 ;
    end else
    begin
        _Result := 0 ;
        P := Evaluator.Get_Position ;
        E := Evaluator.Last_Error ;
    end ;
    Evaluator.Done ;
end ;



end.
