{
        Program Name : IO_Panel
        Package Name : LB
        Purpose      : TPanel for input/output
        Institution  : Conroy & Conroy Co.
        Date Written : 12-Jan-1999
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1999 by Alan Conroy.  Released to the public domain.
        
        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TIO_Panel provides an I/O display.
}

unit IO_Panel ;

interface

uses // Borland...
     Windows,
     Messages,
     SysUtils,
     Classes, // TNotifyEvent
     Graphics,
     Controls,
     Forms,
     Dialogs,
     StdCtrls, // TListBox
     ExtCtrls, // TPanel

     // C&C...
     Output_Panel, // TOutput_Panel
     Standard ; // TInteger_List

type TIO_Panel = class( TOutput_Panel )
                     public // Constructors and destructors...
                         constructor Create( AOwner : TComponent ) ; override ;
                         destructor Destroy ; override ;

                     private // Instance data...
                         // Lists...
                         Input_Start : integer ; // Starting column of user input
                         Saved_Input : TStringList ; // Saved user input for recall

                         // Buffers...
                         Input_Buffer : string ;
                         Input_Buffer_Position : integer ; // Caret position in input buffer

                         // Misc data...
                         Last_Recall : integer ; // For up/down arrows

                         // Selection stuff...
                         Selection_Left, Selection_Right : integer ;
                         Selection : boolean ;
                         Anchor : integer ; // Place where selection started

                     public // API...
                         Overwrite_Mode : boolean ; // True if overwrite, false if insert

                         // Selection-related...
                         procedure Begin_Selection ;
                         procedure End_Selection ;
                         procedure Extend_Selection ;

                         // Position movements...
                         procedure Move_BOL ;
                         procedure Move_EOL ;
                         procedure Move_Left ;
                         procedure Move_Right ;

                         // Input methods...
                         procedure Delete_Current ;
                         procedure Force_Key( Key : char ) ;
                         function Input : string ;
                         procedure Recall_Next ;
                         procedure Recall_Previous ;
                         procedure Reset_Input ;
                         procedure Save_Pending ;
                 end ;

implementation

uses // Borland...
     Clipbrd, // Clipboard

     // C&C...
     ASCIIDef ; // CR

// TIO_Panel methods...

{$IFDEF DEBUG}
procedure Check_IO_Panel( O : TComponent ) ;

begin
    if( ( O = nil ) or ( not ( O is TIO_Panel ) ) ) then
    begin
        raise Exception.Create( 'Invalid typecast or call to method of nil instance' ) ;
    end ;
end ;
{$ENDIF}


// Constructors and destructors...

constructor TIO_Panel.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    Saved_Input := TStringList.Create ;

    Last_Recall := 0 ;
end ;


destructor TIO_Panel.Destroy ;

begin
    Saved_Input.Free ;
    Saved_Input := nil ;

    inherited Destroy ;
end ;


// API...

procedure TIO_Panel.Begin_Selection ;

begin
    if( not Selection ) then
    begin
        Selection := True ;
        Selection_Right := Input_Buffer_Position ;
        Selection_Left := Selection_Right ;
        Anchor := Selection_Right ;
    end ;
end ;


procedure TIO_Panel.End_Selection ;

begin
    if( Selection ) then
    begin
        Selection := False ;
        Invalidate ;
    end ;
end ;


procedure TIO_Panel.Extend_Selection ;

begin
    if( Selection ) then
    begin
        if( Input_Buffer_Position < Anchor ) then
        begin
            Selection_Left := Input_Buffer_Position ;
            Selection_Right := Anchor ;
        end else
        begin
            Selection_Right := Input_Buffer_Position ;
            Selection_Left := Anchor ;
        end ;
        Invalidate ;
    end ;
end ;


procedure TIO_Panel.Move_BOL ;

begin
    Input_Buffer_Position := 0 ;
end ;


procedure TIO_Panel.Move_EOL ;

begin
    Input_Buffer_Position := length( Input_Buffer ) ;
end ;


procedure TIO_Panel.Move_Left ;

begin
    if( Input_Buffer_Position > 0 ) then
    begin
        dec( Input_Buffer_Position ) ; // Move left
    end ;
end ;


procedure TIO_Panel.Move_Right ;

begin
    if( Input_Buffer_Position < length( Input_Buffer ) ) then
    begin
        inc( Input_Buffer_Position ) ; // Move right
    end ;
end ;


procedure TIO_Panel.Delete_Current ;

begin
    if( Selection ) then
    begin
        Selection := False ;
        if( Selection_Left <> Selection_Right ) then
        begin
            Input_Buffer_Position := Selection_Left ;

            Input_Buffer := copy( Input_Buffer, 1, Selection_Left ) +
                copy( Input_Buffer, Selection_Right + 1, length( Input_Buffer ) ) ;
        end ;
        Invalidate ;
    end else
    if( Input_Buffer_Position > 0 ) then // Something to delete
    begin
        // Modify the input buffer
        Input_Buffer := copy( Input_Buffer, 1, Input_Buffer_Position ) +
            copy( Input_Buffer, Input_Buffer_Position + 2, length( Input_Buffer ) ) ;

        // Show to user
        Invalidate ;
    end ;
    Replace( Input_Start, -1, Input_Buffer ) ;
end ;


procedure TIO_Panel.Force_Key( Key : char ) ;

begin
{$IFDEF DEBUG}
    Check_IO_Panel( self ) ;
{$ENDIF}
    // Modify input buffer
    if( ( Key = #255 ) or ( Key = BS ) ) then // Delete previous character
    begin
        if( Selection ) then
        begin
            Selection := False ;
            if( Selection_Left <> Selection_Right ) then
            begin
                Input_Buffer :=
                    copy( Input_Buffer, 1, Selection_Left - 1 ) +
                    copy( Input_Buffer, Selection_Right + 1, length( Input_Buffer ) ) ;
            end ;
        end else
        if( Input_Buffer_Position > 0 ) then // Something to delete
        begin
            Input_Buffer :=
                copy( Input_Buffer, 1, Input_Buffer_Position - 1 ) +
                copy( Input_Buffer, Input_Buffer_Position + 1, length( Input_Buffer ) ) ;
            dec( Input_Buffer_Position ) ;
        end ;
    end else
    if( Key = #3 ) then // Copy to clipboaard
    begin
        if( Selection ) then
        begin
            Clipboard.AsText :=
                copy( Input_Buffer, Selection_Left, Selection_Right - Selection_Left + 1) ;
        end ;
    end else
    if( Key <> CR ) then
    begin
        if( Selection ) then
        begin
            Selection := False ;
            if( Selection_Left <> Selection_Right ) then
            begin
                Input_Buffer_Position := Selection_Left ;
                Input_Buffer := copy( Input_Buffer, 1, Selection_Left ) +
                    copy( Input_Buffer, Selection_Right + 1, length( Input_Buffer ) ) ;
            end ;
        end else
        if( Overwrite_Mode ) then
        begin
            // Restore buffer, delete next character, and remove input again
            Delete_Current ;
        end ;
        Input_Buffer := copy( Input_Buffer, 1, Input_Buffer_Position ) +
            Key +
            copy( Input_Buffer, Input_Buffer_Position + 1, length( Input_Buffer ) ) ;
        inc( Input_Buffer_Position ) ;
        Output( Key, Font.Color ) ;
    end ;

    Replace( Input_Start, -1, Input_Buffer ) ;

    // Show to user
    Invalidate ;

    // Final stuff
    Last_Recall := Saved_Input.Count + 1 ;
end ; // TIO_Panel.Force_Key


function TIO_Panel.Input : string ;

begin
    Save_Pending ;
    Result := Input_Buffer ;
    Reset_Input ;
end ;


procedure TIO_Panel.Recall_Next ;

begin
    if( Last_Recall < Saved_Input.Count ) then
    begin
        inc( Last_Recall ) ;
        Input_Buffer := Saved_Input[ Last_Recall - 1 ] ;
    end else
    begin
        Input_Buffer := '' ;
        Last_Recall := Saved_Input.Count + 1 ;
    end ;
    Input_Buffer_Position := length( Input_Buffer ) ;
    Replace( Input_Start, -1, Input_Buffer ) ;
    Invalidate ;
end ;


procedure TIO_Panel.Recall_Previous ;

begin
    if( Last_Recall > 0 ) then
    begin
        dec( Last_Recall ) ;
        if( Last_Recall > 0 ) then
        begin
            Input_Buffer := Saved_Input[ Last_Recall - 1 ] ;
        end else
        begin
            Input_Buffer := '' ;
        end ;
    end else
    begin
        Input_Buffer := '' ;
    end ;
    Input_Buffer_Position := length( Input_Buffer ) ;
    Replace( Input_Start, -1, Input_Buffer ) ;
    Invalidate ;
end ;


procedure TIO_Panel.Reset_Input ;

begin
    Input_Buffer_Position := 0 ;
    Input_Buffer := '' ;
    Overwrite_Mode := False ;
    Input_Start := CCPos ;
    Selection := False ;
    Invalidate ;
end ;


procedure TIO_Panel.Save_Pending ;

begin
    if( length( Input_Buffer ) > 0 ) then
    begin
        Saved_Input.Add( Input_Buffer ) ;
    end ;
end ;



end.
