{
        Program Name : SCULIB
        Package Name : SCU
        Purpose      : System Configuration Utility Library
        Institution  : Conroy & Conroy Co.
        Date Written : 16-April-2001
        Written By   : Alan Conroy
        Version      : 1.1B

	Copyright (C) 2001-2003 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  12-June-2002  EAC         Added contextual interface.    *
        *  22-Oct-2003   EAC         Fixed some pathing bugs.       *
        *  23-Oct-2003               Added singleton support.       *
        *                            Fix uninitialized Ignore_Alarms*
        *                            which could cause alarms to be *
        *                            incorrectly ignored.           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  SCU is the System Configuration Utility, used to manage software
	configuration.  Conceptually, this is similar to the Windows
	registry, but without the problems thereof.

          All data values start with a two-byte type code.

          Singleton is set only if we have a single system-wide instance of SCU
        (SCU32SRV server).
}

unit SCULib ;

interface

{$I EDEFINES.INC}

uses _Common, { TObject }
     _DebugIn, { PDebug_Interface }
     _UEHDefs, { TUEC }
     UEHDefs, { TStandard_Error_Interface }
     Files, { PSFile }
     FSFH, { TXFile_System }
     Loggers, { PSLogger }
     SCU ; { T_SCU }

const SCUErr_Facility = 22 ;
      SCUErr_Success = 0 ;
      SCUErr_Value_Not_Found = 1 ;
      SCUErr_Invalid_Size = 2 ;
      SCUErr_Exists = 3 ; { Key or value already exists }
      SCUErr_Alarm_Failure = 4 ; { Operation caused an alarm which user used to abort operation }
      SCUErr_Bad_Data = 5 ; { Bad data (used by SCU32.DLL) }

{$IFDEF DELPHI}
type TSCU = class( T_SCU )
                public { Constructors and destructors... }
                    constructor Construct ;
                    destructor Done ;
{$ELSE}
type TSCU = object( TObject )
{$ENDIF}
		private { Instance data... }
		    _File_System : PXFile_System ; { File system for global SCU.DAT }
		    _System_Path : string ; { Path and filename for SCU.DAT }
		    _Max_Parent_Nest : integer ;
                    Find_Record : TFind_Record ;
                    Ignore_Alarms : boolean ; { True to ignore alarms }
                    Temp_Lookup : string ;
                    _Contextual : boolean ; { True if running contextually }
                    Startup_Error : TUEC ;
                    _Reference_Count : integer ;

		private { Internal routines... }
                    function Check_Delete( Path : string ) : boolean ;
                    function Check_Write( Path : string ) : boolean ;
                    function Check_Read( Path : string ) : boolean ;
		    function Find_Path( User : string ; Path : string ;
                        var Fil : PSFile ) : string ;
                    procedure Lock_SCU ;
                    function Open_Existing_File( var User,
                        Path : string ) : PSFile ;
                    function Read_String( const Path : string ) : string ;
                    procedure Unlock_SCU ;

		    { Instance support... }
                    function _Init( const Name : string ;
                        Extend : longint ) : TUEC ;

		    { Low-level access... }
		    function _Write( Path : string ; Data : pointer ;
			Size : integer ) : TUEC ;
		    function _Read( Path : string ; Data : pointer ;
			Size : integer ) : TUEC ;
		    function _Read_Size( Path : string ) : integer ;
		    function _Read_Type( Path : string ) : integer ;
		    function _Exists( Path : string ) : boolean ;
		    function _Key_Exists( Path : string ) : boolean ;

                    { Look up the specified path and wildcard for the specified
                      type of data (SCUDT_Unknown for everything, SCUDT_Key for
                      keys only. }
		    function _Lookup( Path : string ; DT : integer ) : string ;

		    function _Delete( Path : string ; Logger : PSLogger ;
                        var Abort : boolean ) : TUEC ;
		    function _Create( Path : string ; DT : longint ) : TUEC ;
                    function _Create_And_Write( Path : string ; DT : longint ;
                        Data : pointer ; Size : integer ) : TUEC ;
                    function _Rename( Old, New : string ) : TUEC ;

		    { General API... }
		    function _Set_Value( User, Path : string ; Data : pointer ;
			Size : integer ) : TUEC ;
		    function _Get_Value( User, Path : string ; Data : pointer ;
			Size : integer ) : TUEC ;
		    function _Value_Size( User, Path : string ) : longint ;
		    function _Value_Type( User, Path : string ) : longint ;
		    function _Value_Exists( User, Path : string ) : boolean ;
		    function _Delete_Value( User, Path : string ) : TUEC ;
		    function _Create_Value( User, Path : string ;
                        DT : longint ) : TUEC ;
                    function _Read_Value( User, Path : string ;
                        var DT, Size : longint ; var Buffer : PChar ) : TUEC ;

                public { Public instance support }
                    Singleton : boolean ;

                    procedure Attach ; override ; stdcall ;
                    procedure Detach ; override ; stdcall ;
		    function Initialize : TUEC ; override ; stdcall ;
		    procedure Terminate ; override ; stdcall ;
                    function Init( Name : PChar ) : TUEC ; override ; stdcall ;

                public { Low-level API... }
		    function Write( Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Read( Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Read_Size( Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Read_Type( Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Exists( Path : PChar ) : boolean ;
                        override ; stdcall ;
		    function Key_Exists( Path : PChar ) : boolean ;
                        override ; stdcall ;

		    function Lookup( Path : PChar ; DT : longint ) : PChar ;
                        override ; stdcall ;

		    function Delete( Path : PChar ) : TUEC ;
                        override ; stdcall ;
		    function Create( Path : PChar ; DT : longint ) : TUEC ;
                         override ; stdcall ;
                    function Create_And_Write( Path : PChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : TUEC ;
                        override ; stdcall ;
                    function Rename( Old, New : PChar ) : TUEC ;
                        override ; stdcall ;

                public { Standard API... }
		    function Set_Value( User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Get_Value( User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Value_Size( User, Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Value_Type( User, Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Value_Exists( User, Path : PChar ) : boolean ;
                        override ; stdcall ;
		    function Delete_Value( User, Path : PChar ) : TUEC ;
                        override ; stdcall ;
		    function Create_Value( User, Path : PChar ;
                        DT : longint ) : TUEC ; override ; stdcall ;
                    function Verify : TUEC ; override ; stdcall ;
                    function Debugger : PDebug_Interface ; override ; stdcall ;
                    function Version : longint ; override ; stdcall ;

                    function Read_Value( User, Path : PChar ;
                        var DT, Size : longint ; var Buffer : PChar) : TUEC ;
                        override ; stdcall ;

                    function Init_And_Extend( Name : PChar ;
                        Size : longint ) : TUEC ; override ; stdcall ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes. }

                    function Delete_With_Callback( Path : PChar ;
                        Logger : PSLogger ) : TUEC ; override ; stdcall ;

                    procedure Shutdown ; override ; stdcall ;

                { V1.2 interface: }
                public { Contextual version of instance support... }
                    function Init_Context( Name : PChar ; Flags : longint ) : longint ;
                        override ; stdcall ;
                    procedure Terminate_Context( Context : longint ) ;
                        override ; stdcall ;
                    function Init_And_Extend_Context( Name : PChar ;
                        Size : longint ; Flags : longint ) : longint ;
                        override ; stdcall ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes. }

                public { Contextual version of Low-level API... }
		    function Write_Context( Context : longint ; Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Read_Context( Context : longint ; Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Read_Size_Context( Context : longint ; Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Read_Type_Context( Context : longint ; Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Exists_Context( Context : longint ; Path : PChar ) : boolean ;
                        override ; stdcall ;
		    function Key_Exists_Context( Context : longint ; Path : PChar ) : boolean ;
                        override ; stdcall ;

		    function Lookup_Context( Context : longint ; Path : PChar ; DT : longint ) : PChar ;
                        override ; stdcall ;

		    function Delete_Context( Context : longint ; Path : PChar ) : TUEC ;
                        override ; stdcall ;
		    function Create_Context( Context : longint ; Path : PChar ; DT : longint ) : TUEC ;
                         override ; stdcall ;
                    function Create_And_Write_Context( Context : longint ; Path : PChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : TUEC ;
                        override ; stdcall ;
                    function Rename_Context( Context : longint ; Old, New : PChar ) : TUEC ;
                        override ; stdcall ;

                public { Contextual version of Standard API... }
		    function Set_Value_Context( Context : longint ; User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Get_Value_Context( Context : longint ; User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; override ; stdcall ;
		    function Value_Size_Context( Context : longint ; User, Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Value_Type_Context( Context : longint ; User, Path : PChar ) : longint ;
                        override ; stdcall ;
		    function Value_Exists_Context( Context : longint ; User, Path : PChar ) : boolean ;
                        override ; stdcall ;
		    function Delete_Value_Context( Context : longint ; User, Path : PChar ) : TUEC ;
                        override ; stdcall ;
		    function Create_Value_Context( Context : longint ; User, Path : PChar ;
                        DT : longint ) : TUEC ; override ; stdcall ;
                    function Verify_Context( Context : longint ) : TUEC ;
                        override ; stdcall ;
                    function Debugger_Context( Context : longint ) : PDebug_Interface ;
                        override ; stdcall ;

                    function Read_Value_Context( Context : longint ; User, Path : PChar ;
                        var DT, Size : longint ; var Buffer : PChar) : TUEC ;
                        override ; stdcall ;

                    function Delete_With_Callback_Context( Context : longint ;
                        Path : PChar ; Logger : PSLogger ) : TUEC ;
                        override ; stdcall ;
            end ; { TSCU }

const SCUDT_Undefined = 0 ;
      SCUDT_Key = 1 ;
      SCUDT_String = 2 ;
      SCUDT_Integer = 3 ; { 4-bit signed }
      SCUDT_Boolean = 4 ;
      SCUDT_Float = 5 ; { Double-precision }
      SCUDT_Binary = 6 ;
      SCUDT_All = -1 ; { All types }

{$IFDEF DELPHI}
function Create_SCU_Interface : T_SCU ;
procedure Free_SCU_Interface( X : T_SCU ) ;
{$ENDIF}

implementation

uses { Borland... }
     Windows, { MessageBox }
     Sysutils, { Uppercase }

     { C&C... }
     Collect, { TString_Collection }
     CommonUt, { Edit }
     Compatib, { Set_Length }
     DebugInt, { Delphi_Debug_Wrapper }
     ErrMastr, { Register_Error_Interface }
     FSError, { FSErr_FileExists }
     Heaps, { THeap }
     NUM1S, { Num1 }
     Standard ; { Set_Length }

{ Unit API... }

{$IFDEF DELPHI}
function Create_SCU_Interface : T_SCU ;

begin
    Create_SCU_Interface := TSCU.Construct ;
end ;


procedure Free_SCU_Interface( X : T_SCU ) ;

begin
    X.Free ;
end ;
{$ENDIF}


{ Unit utility routines... }

{ Make sure the passed user name has no leading backslash and one ending one. }
function Normalize_User( User : string ) : string ;

begin
    if( length( User ) > 0 ) then
    begin
        if( User[ 1 ] = '\' ) then
        begin
            User := copy( User, 2, length( User ) ) ;
        end ;
        if( pos( '\', User ) > 0 ) then
        begin
            Normalize_User := '' ;
            exit ;
        end ;
	if( copy( User, length( User ), 1 ) <> '\' ) then
	begin
	    User := User + '\' ;
	end ;
    end ;
    Normalize_User := User ;
end ;


function Strip_Root( const S : string ) : string ;

var Dummy : integer ;

begin
    if( pos( ':', S ) > 0 ) then { A full path }
    begin
        Dummy := pos( '\SCU.DAT\', S ) ;
        Strip_Root := copy( S, Dummy + 9, length( S ) ) ;
    end else
    begin
        Strip_Root := S ;
    end ;
end ;



{ TSCU methods... }

{ Constructors and destructors... }

constructor TSCU.Construct ;

begin
    Initialize_Find_Record( Find_Record ) ;
end ;


destructor TSCU.Done ;

begin
    Uninitialize_Find_Record( Find_Record ) ;

{$IFDEF DELPHI}
    inherited Destroy ;
{$ENDIF}
end ;


{ Internal routines... }

{ Checks for alarm for specified path.  If it does not exist, or deletes
  allowed, return True, else prompt user. }
function TSCU.Check_Delete( Path : string ) : boolean ;

var S : string ;

begin
    Result := True ; { Assume success }
    if( Ignore_Alarms ) then
    begin
        exit ;
    end ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
        Path := '\' + Path ;
    end ;
    Path := Edit( Path, 32 ) ;
    if( copy( Path, 1, 8 ) = '\ALARMS\' ) then
    begin
        if( not _Key_Exists( Path ) ) then
        begin
            exit ;
        end ;
        setlength( S, _Read_Size( Path ) ) ;
        _Read( Path, @S[ 1 ], length( S ) ) ;
        if( pos( 'C', S ) = 0 ) then
        begin
            exit ; { Control (of alarms) allowed }
        end ;
        Result :=
            MessageBox( 0, PChar( 'Attempt to delete alarm for software configuration key: ' + copy( Path, 8, length( Path ) ) + '.  Allow deletion?' ), 'SCU', MB_ICONEXCLAMATION or MB_YESNO ) = idYes ;
    end else
    begin
        if( not _Key_Exists( '\ALARMS' + Path ) ) then
        begin
            exit ;
        end ;
        setlength( S, _Read_Size( '\ALARMS' + Path ) ) ;
        _Read( '\ALARMS' + Path, @S[ 1 ], length( S ) ) ;
        if( pos( 'D', S ) = 0 ) then
        begin
            exit ; { Deletion allowed }
        end ;
        Result :=
            MessageBox( 0, PChar( 'Attempt to delete software configuration key: ' + Path + '.  Allow deletion?' ), 'SCU', MB_ICONEXCLAMATION or MB_YESNO ) = id_Yes ;
    end ;
end ; { TSCU.Check_Delete }


{ Checks for alarm for specified path.  If not exists, or writes allowed,
  return True, else prompt user. }
function TSCU.Check_Write( Path : string ) : boolean ;

var S : string ;

begin
    Result := True ; { Assume success }
    if( Ignore_Alarms ) then
    begin
        exit ;
    end ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
        Path := '\' + Path ;
    end ;
    Path := Edit( Path, 32 ) ;
    if( copy( Path, 1, 8 ) = '\ALARMS\' ) then
    begin
        if( not _Key_Exists( Path ) ) then
        begin
            exit ;
        end ;
        setlength( S, _Read_Size( Path ) ) ;
        _Read( Path, @S[ 1 ], length( S ) ) ;
        if( pos( 'C', S ) = 0 ) then
        begin
            exit ; { Control (of alarms) allowed }
        end ;
        Result :=
            MessageBox( 0, PChar( 'Attempt to change alarm for software configuration key: ' + copy( Path, 8, length( Path ) ) + '.  Allow change?' ), 'SCU', MB_ICONEXCLAMATION or MB_YESNO ) = idYes ;
    end else
    begin
        if( not _Key_Exists( '\ALARMS' + Path ) ) then
        begin
            exit ;
        end ;
        setlength( S, _Read_Size( '\ALARMS' + Path ) ) ;
        _Read( '\ALARMS' + Path, @S[ 1 ], length( S ) ) ;
        if( pos( 'W', S ) = 0 ) then
        begin
            exit ; { Write allowed }
        end ;
        Result :=
            MessageBox( 0, PChar( 'Attempt to change software configuration key: ' + Path + '.  Allow change?' ), 'SCU', MB_ICONEXCLAMATION or MB_YESNO ) = id_Yes ;
    end ;
end ; { TSCU.Check_Write }


{ Checks for alarm for specified path.  If not exists, or reads allowed,
  return True, else prompt user. }
function TSCU.Check_Read( Path : string ) : boolean ;

var S : string ;

begin
    Result := True ; { Assume success }
    if( Ignore_Alarms ) then
    begin
        exit ;
    end ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
        Path := '\' + Path ;
    end ;
    Path := uppercase( Path ) ;
    if( copy( Path, 1, 8 ) = '\ALARMS\' ) then
    begin
        exit ; { Reads of alarms are always allowed }
    end else
    begin
        if( not _Key_Exists( '\ALARMS' + Path ) ) then
        begin
            exit ;
        end ;
        setlength( S, _Read_Size( '\ALARMS' + Path ) ) ;
        _Read( '\ALARMS' + Path, @S[ 1 ], length( S ) ) ;
        if( pos( 'R', S ) = 0 ) then
        begin
            exit ; { Read allowed }
        end ;
        Result :=
            MessageBox( 0, PChar( 'Attempt to read software configuration key: ' + Path + '.  Allow read?' ), 'SCU', MB_ICONEXCLAMATION or MB_YESNO ) = id_Yes ;
    end ;
end ; { TSCU.Check_Read }


{ Find a path - assumes User is normalized (as per Normalize_User).  Path is
  function return.  Fil is ignored on entry and is a pointer to the file
  matching the path on return. }
function TSCU.Find_Path( User : string ; Path : string ; var Fil : PSFile ) : string ;

var Dummy : integer ;
    F : PSFile ;
    Nested : integer ;
    S, S1 : string ;
    System_Path : string ;

    function Lookup_Group( const Group : string ) : string ;

    var Dummy : integer ;
        Res : TUEC ;
        S, S1 : string ;

    begin
	if( Nested > _Max_Parent_Nest ) then
	begin
	    Lookup_Group := '' ;
	    exit ;
	end ;
	inc( Nested ) ;
	S1 := System_Path + '\Groups\' + Group + '\Groups\*.*' ;
	S1 := _Lookup( S1, SCUDT_Key ) ;
	S := System_Path + '\Groups\' ;
	while( length( S1 ) > 0 ) do
	begin
	    { Get next parent group... }
	    Dummy := length( S1 ) ;
	    while( ( Dummy > 0 ) and ( S1[ Dummy ] <> '\' ) ) do
	    begin
		dec( Dummy ) ;
	    end ;
	    S1 := copy( S1, Dummy + 1, length( S1 ) ) ; { Get group name }

	    { See if path found in this group... }
	    F := Open( S + S1 + Path, FSOM_Input, Res ) ;
	    if( F <> nil ) then
	    begin
		F^.Close ;
		Lookup_Group := S + S1 + Path ;
		exit ;
	    end ;

	    { Try this group's parent(s)... }
	    S1 := Lookup_Group( S1 ) ;
	    if( length( S ) > 0 ) then
	    begin
		F^.CLose ;
		Lookup_Group := S ;
		exit ;
	    end ;

	    { Try next parent in this group }
	    S1 := _Lookup( '', SCUDT_Key ) ;
	end ;
	Lookup_Group := '' ; { Not found }
	dec( Nested ) ;
    end ; { TSCU.Find_Path.Lookup_Group }

var Count : integer ;
    Res : TUEC ;

begin { TSCU.Find_Path }
    { Find a path in the SCU distributed database.  The rules are to look in
      the following places until a matching path is found:
	1. Look in <user-default-path>\<path>.  This step is skipped if
           User is not defined.  "user-default-path" is found in
           Windows\SCU.DAT\Users\<user>\SCU Path.
        2. Look in Windows\SCU.DAT\Users\<user>\<path>.  This step is skipped if
           User is not defined.
	3. Iterate through SCU.DAT\Users\<user>\Groups\* using the found
	   group names to look in SCU.DAT\Groups\<group>\<path>.  This step is
           skipped if User is not defined.
	4. Iterate through the groups from step 2, looking for parent groups
	   (mentioned via SCU.DAT\Groups\<group>\Groups\*).  This step is
           skipped if step 2 is not skipped.
	5. Look in SCU.DAT\<path>
	6. If SCU.DAT\Network\SCU_Path is defined, use the value as the name of
	   the SCU.DAT file and goto step 2. }

    { Prepare parameters... }
    if( ( length( Path ) > 0 ) and ( Path[ 1 ] <> '\' ) ) then
    begin
        Path := '\' + Path ; { Path must start with backslash }
    end ;
    if( copy( User, 1, 1 ) = '\' ) then
    begin
        User := copy( User, 2, length( User ) ) ;
    end ;
    if( copy( User, length( User ), 1 ) = '\' ) then
    begin
        User := copy( User, 1, length( User ) - 1 ) ;
    end ;

    { Set default values... }
    if( length( User ) > 0 ) then
    begin
        Find_Path := _System_Path + '\Users\' + User + Path ;
    end else
    begin
        Find_Path := _System_Path + Path ;
    end ;
    Fil := nil ;

    { Now search for existing value... }
    System_Path := _System_Path ;
    if( length( User ) > 0 ) then
    begin
        { Step 1... }
        S := Read_String( System_Path + '\Users\' + User + '\SCU Path' ) ;
        if( length( S ) > 0 ) then
        begin
            if( copy( S, length( S ), 1 ) <> '\' ) then
            begin
                S := S + '\' ;
            end ;
            S := S + 'SCU.DAT' + Path ;
            Fil := Open( S, FSOM_Input, Res ) ;
            if( Fil <> nil ) then
            begin
                Find_Path := S ;
                exit ;
            end ;
        end ; { if( length( S ) > 0 ) }
    end ; { if( length( User ) > 0 ) }
    while( True ) do
    begin
	if( length( User ) > 0 ) then
	begin
	    { Step 2... }
	    S := System_Path + '\Users\' + User + Path ;
	    Fil := Open( S, FSOM_Input, Res ) ;
	    if( Fil <> nil ) then
	    begin
		Find_Path := S ;
		exit ;
	    end ;

	    { Step 3... }
	    S := System_Path + '\Users\' + User + '\Groups\*.*' ;
	    S1 := _Lookup( S, SCUDT_Key ) ;
	    S := System_Path + '\Groups\' ;
	    while( length( S1 ) > 0 ) do
	    begin
		Dummy := length( S1 ) ;
		while( ( Dummy > 0 ) and ( S1[ Dummy ] <> '\' ) ) do
		begin
		    dec( Dummy ) ;
		end ;
		S1 := copy( S1, Dummy + 1, length( S1 ) ) ; { Get group name }
		Fil := Open( S + S1 + Path, FSOM_Input, Res ) ;
		if( Fil <> nil ) then
		begin
		    Find_Path := S + S1 + Path ;
		    exit ;
		end ;
		S1 := _Lookup( '', SCUDT_Key ) ;
	    end ;

	    { Step 4... }
	    S := System_Path + '\Users\' + User + '\Groups\*.*' ;
	    S1 := _Lookup( S, SCUDT_Key ) ;
	    while( length( S1 ) > 0 ) do
	    begin
		Dummy := length( S1 ) ;
		while( ( Dummy > 0 ) and ( S1[ Dummy ] <> '\' ) ) do
		begin
		    dec( Dummy ) ;
		end ;
		S1 := copy( S1, Dummy + 1, length( S1 ) ) ; { Get group name }
		Nested := 0 ; { Prevent infinite loops }
		S := Lookup_Group( S1 ) ;
		if( length( S ) > 0 ) then
		begin
		    Find_Path := S ;
		    exit ;
		end ;
		S1 := _Lookup( '', SCUDT_Key ) ;
	    end ;
        end ; { if( length( User ) > 0 )  }

        { Step 5... }
        S := System_Path + Path ;
        Fil := Open( S, FSOM_Input, Res ) ;
        if( Fil <> nil ) then
        begin
            Find_Path := S ;
            exit ;
        end ;

        { Step 6... }
        F := Open( System_Path + '\Network\SCU_Path', FSOM_Input, Res ) ;
        if( F = nil ) then
        begin
            exit ;
        end ;
        if( ( F^.Get_Size > 259 ) or ( F^.Get_Size < 3 ) ) then { Network path is too long or too short }
        begin
            F^.Close ;
            exit ;
        end ;
        Dummy := 0 ;
        F^.Blockread( Dummy, 2, Count ) ;
        if( Dummy <> SCUDT_String ) then { Not a string }
        begin
            F^.Close ;
            exit ;
        end ;
        F^.Seek( 2 ) ;
        SetLength( System_Path, F^.Get_Size - 2 ) ;
        F^.Blockread( System_Path[ 1 ], length( System_Path ), Count ) ;
        F^.Close ;
        if( copy( System_Path, length( System_Path ), 1 ) <> '\' ) then
        begin
            System_Path := System_Path + '\' ;
        end ;
        System_Path := System_Path + 'SCU.DAT' ;
    end ; { while( True ) }
end ; { TSCU.Find_Path }


procedure TSCU.Lock_SCU ;

begin
    if( ( not ( _Contextual or Singleton ) ) and ( _File_System <> nil  ) ) then
    begin
        _File_System^.Lock( _System_Path ) ;
    end ;
end ;


function TSCU.Open_Existing_File( var User, Path : string ) : PSFile ;

var Res : TUEC ;

begin
    User := Normalize_User( User ) ;
    Path := Find_Path( User, Path, Result ) ;
    if( Result = nil ) then
    begin
        Result := Open( Path, FSOM_Input, Res ) ;
    end ;
    if( Result <> nil ) then
    begin
        Result^.Seek( 0 ) ;
    end ;
end ;


function TSCU.Read_String( const Path : string ) : string ;

var Dummy : integer ;
    S : string ;
    U : TUEC ;

begin
    Dummy := _Read_Size( Path ) ;
    if( Dummy < 1 ) then
    begin
        Read_String := '' ;
        exit ;
    end ;
    Set_Length( S, Dummy ) ;
    U := _Read( Path, @S[ 1 ], Dummy ) ;
    Read_String := S ;
end ;


procedure TSCU.Unlock_SCU ;

begin
    if( ( not ( _Contextual or Singleton ) ) and ( _File_System <> nil  ) ) then
    begin
        _File_System^.Unlock( _System_Path ) ;
    end ;
end ;


{ Instance support... }

procedure Initialize_Find_Record( var Find_Record : TFind_Record ) ;

begin
    Files.Initialize_Find_Record( Find_Record ) ;
    Find_Record.Separator := #0 ; { Don't use file lists }
end ;


procedure TSCU.Attach ;

begin
    inc( _Reference_Count ) ;
end ;


procedure TSCU.Detach ;

begin
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
        Terminate ;
        Free ;
    end ;
end ;


function TSCU.Initialize : TUEC ;

begin
{$IFDEF DELPHI}
    System.Initialize( _System_Path ) ;
    System.Initialize( Temp_Lookup ) ;
    Singleton := False ;
    Ignore_Alarms := False ;

    { Get parent of Windows system directory... }
    SetLength( _System_Path, MAX_PATH ) ;
    SetLength( _System_Path, GetSystemDirectory( PChar( _System_Path ), MAX_PATH ) ) ;
    Setlength( _System_Path, length( _System_Path ) - 1 ) ; { In case it ends with a backslash }
    while( pos( _System_Path[ length( _System_Path ) ], '\:' ) = 0 ) do
    begin
        Setlength( _System_Path, length( _System_Path ) - 1 ) ;
    end ;
{$ELSE}
    _System_Path := 'C:\WINDOWS\' ;
{$ENDIF}
    _System_Path := _System_Path + 'SCU.DAT' ;
    Initialize := _Init( _System_Path, -1 ) ;
end ;


function TSCU._Init( const Name : string ; Extend : longint ) : TUEC ;

var Dummy : integer ;
    _Find_Record : TFind_Record ;
    F : file ;
    FH : TFH ;
    Heap : TSHeap ;
    P : longint ;
    Res : TUEC ;

begin
    _System_Path := Name ;
    _Max_Parent_Nest := 32 ;
    assign( F, _System_Path ) ;
    {$I-}
    reset( F ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy = 2 ) then { File doesn't exist }
    begin
	Initialize_File_System( _System_Path, 'Software Configuration Utility', 0, Res ) ;
	if( Res.Code <> 0 ) then
	begin
	    _Init := Res ;
	    exit ;
	end ;
        if( Extend > 0 ) then { User wants to pre-extend file }
        begin
            Heap.Init ;
            FH.Init( 256, 0, _System_Path, Res, @Heap ) ;
            if( Res.Code = 0 ) then { Succesfully opened file as file heap }
            begin
                P := FH.Allocate( Extend ) ;
                if( P <> 0 ) then { Successfully allocated space }
                begin
                    FH.Deallocate( P, Extend ) ;
                end ;
                FH.Own_Allocator := False ;
                FH.Done ;
            end ;
            Heap.Done ;
        end ;
        _File_System := PXFile_System( Open_File_System( Name, 0, Res ) ) ;
        Dummy := _File_System^.MkDir( '\Network' ) ; { Do this first to start directory tree's life balanced }
        if( Dummy <> 0 ) then
        begin
            _File_System^.Last_Error( Result ) ;
            exit ;
        end ;
        Dummy := _File_System^.MkDir( '\Groups' ) ;
        if( Dummy <> 0 ) then
        begin
            _File_System^.Last_Error( Result ) ;
            exit ;
        end ;
        Dummy := _File_System^.MkDir( '\Alarms' ) ;
        if( Dummy <> 0 ) then
        begin
            _File_System^.Last_Error( Result ) ;
            exit ;
        end ;
        Dummy := _File_System^.MkDir( '\Users' ) ;
        if( Dummy <> 0 ) then
        begin
            _File_System^.Last_Error( Result ) ;
            exit ;
        end ;
        Res.Facility := 1 ;
        Res.Code := 0 ;
        if( Singleton ) then
        begin
            Ignore_Alarms := True ; { No alarms exist at this point }
        end ;
    end else
    begin
        _File_System := PXFile_System( Open_File_System( Name, 0, Res ) ) ;
        Startup_Error := Res ;
        if( Singleton and ( Res.Code = 0 ) ) then
        begin
            Initialize_Find_Record( _Find_Record ) ;
            Ignore_Alarms := ( _File_System.FindFirst( '\Alarms\*.*', -1, _Find_Record ) <> 0 ) ;
            Uninitialize_Find_Record( _Find_Record ) ;
        end ;
    end ; { if( Dummy = 2 ) }
    _Init := Res ;

    { Other setup }
    if( Find_Record.Initialized = 0 ) then
    begin
        Uninitialize_Find_Record( Find_Record ) ;
    end ;
    Initialize_Find_Record( Find_Record ) ;
end ; { TSCU._Init }


procedure TSCU.Terminate ;

begin
    Uninitialize_Find_Record( Find_Record ) ;
    if( ( not _Contextual ) and ( _File_System <> nil ) ) then
    begin
        Close_File_System( _File_System^.Path ) ;
    end ;
end ;


function TSCU.Version : longint ;

begin
    Version := 12 ; { V1.2 }
end ;


{ Low-level API... }

function TSCU._Write( Path : string ; Data : pointer ;
    Size : integer ) : TUEC ;

var Count : integer ;
    F : PSFile ;
    Res : TUEC ;

begin
    if( Size < 0 ) then
    begin
	Res.Facility := SCUErr_Facility ;
	Res.Code := SCUErr_Invalid_Size ;
	_Write := Res ;
	exit ;
    end ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    if( not Check_Write( Path ) ) then
    begin
        exit ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    F := Open( _System_Path + Path, FSOM_Input, Res ) ;
    if( F = nil ) then
    begin
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        if( Res.Code = 2 ) then
        begin
            Res.Facility := SCUErr_Facility ;
            Res.Code := SCUErr_Value_Not_Found ;
        end ;
        _Write := Res ;
        exit ;
    end ;
    F^.Seek( 2 ) ;
    F^.Blockwrite( Data^, Size, Count ) ;
    F^.Truncate( Size + 2 ) ;
    F^.Close ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
    Res.Facility := SCUErr_Facility ;
    Res.Code := SCUErr_Success ;
    _Write := Res ;
end ; { TSCU._Write }


function TSCU._Read( Path : string ; Data : pointer ; Size : integer ) : TUEC ;

var Count : integer ;
    F : PSFile ;
    Res : TUEC ;

begin
    if( Size < 0 ) then
    begin
	Res.Facility := SCUErr_Facility ;
	Res.Code := SCUErr_Invalid_Size ;
	_Read := Res ;
	exit ;
    end ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    if( not Check_Read( Path ) ) then
    begin
        exit ;
    end ;
    F := Open( _System_Path + Path, FSOM_Input, Res ) ;
    if( F = nil ) then
    begin
	if( Res.Code = 2 ) then
	begin
	    Res.Facility := SCUErr_Facility ;
	    Res.Code := SCUErr_Value_Not_Found ;
	end ;
	_Read := Res ;
	exit ;
    end ;
    F^.Seek( 2 ) ;
    F^.Blockread( Data^, Size, Count ) ;
    F^.Close ;
    Res.Facility := SCUErr_Facility ;
    Res.Code := SCUErr_Success ;
    _Read := Res ;
end ; { TSCU._Read }


function TSCU._Read_Size( Path : string ) : integer ;

var F : PSFile ;
    Res : TUEC ;

begin
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    F := Open( _System_Path + Path, FSOM_Input, Res ) ;
    if( F = nil ) then
    begin
	_Read_Size := -1 ; { Some error }
	exit ;
    end ;
    _Read_Size := F^.Get_Size - 2 ;
    F^.Close ;
end ; { TSCU._Read_Size }


function TSCU._Read_Type( Path : string ) : integer ;

var A : word ;
    Count, Dummy : longint ;
    F : PSFile ;
    _Find_Record : TFind_Record ;
    Res : TUEC ;
    S : string ;

begin
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    F := Open( _System_Path + Path, FSOM_Input, Res ) ;
    if( F = nil ) then
    begin
        Initialize_Find_Record( _Find_Record ) ;
        S := Files.Lookup( _System_Path + Path, _Find_Record ) ;
	if( length( S ) = 0 ) then
        begin
            _Read_Type := -1 ; { Some error }
        end else
        begin
            _Read_Type := SCUDT_Key ;
        end ;
        Uninitialize_Find_Record( _Find_Record ) ;
	exit ;
    end ;
    A := F^.Get_Attributes ;
    if( ( A and FSFF_Directory ) <> 0 ) then
    begin
        _Read_Type := SCUDT_Key ;
        exit ;
    end ;
    Dummy := 0 ;
    F^.Blockread( Dummy, 2, Count ) ;
    F^.Close ;
    _Read_Type := Dummy ;
end ; { TSCU._Read_Type }


function TSCU._Key_Exists( Path : string ) : boolean ;

var _Find_Record : TFind_Record ;
    S : string ;

begin
    Initialize_Find_Record( _Find_Record ) ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    S := Files.Lookup( _System_Path + Path, _Find_Record ) ;
    _Key_Exists := ( length( S ) > 0 ) ;
    Uninitialize_Find_Record( _Find_Record ) ;
end ;


function TSCU._Exists( Path : string ) : boolean ;

var F : PSFile ;
    Res : TUEC ;

begin
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    F := Open( _System_Path + Path, FSOM_Input, Res ) ;
    if( F = nil ) then
    begin
	_Exists := False ;
	exit ;
    end ;
    _Exists := True ;
    F^.CLose ;
end ;


function TSCU._Delete( Path : string ; Logger : PSLogger ;
    var Abort : boolean ) : TUEC ;

var _Find_Record : TFind_Record ;
    List : TString_Collection ;
    Loop : integer ;
    Res : TUEC ;
    S : string ;

begin
    { Setup... }
    fillchar( Res, sizeof( Res ), 0 ) ;
    _Delete := Res ;
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    if( copy( Path, length( Path ), 1 ) = '\' ) then
    begin
	Path := copy( Path, 1, length( Path ) - 1 ) ;
    end ;

    if( not Check_Delete( Path ) ) then
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Alarm_Failure ;
        _Delete := Res ;
        exit ;
    end ;

    { Delete all subordinate data... }

    { First build list of all matches... }
    Initialize_Find_Record( _Find_Record ) ;
    S := Files.Lookup( _System_Path + Path + '\...\*.*', _Find_Record ) ;
    List.Init( 0, 0 ) ;
    while( length( S ) > 0 ) do
    begin
        S := copy( S, length( _System_Path ) + 1, length( S ) ) ; { Trim system path }
	List.Insert( S ) ;
        S := Files.Lookup( '', _Find_Record ) ; { Find next match }
    end ;

    { Now delete each item in list... }
    for Loop := 0 to List.Count - 1 do
    begin
        S := List.At( Loop ) ;
        if( not Singleton ) then
        begin
            Lock_SCU ;
        end ;
        Res := _Delete( S, Logger, Abort ) ; { Delete subordinate data to S }
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        if(
            ( Res.Code <> 0 )
            and
            (
              ( Res.Facility <> SCUErr_Facility )
              or
              ( Res.Code <> SCUErr_Value_Not_Found )
            )
          ) then
        begin
            _Delete := Res ;
            Uninitialize_Find_Record( _Find_Record ) ;
	    List.Done ;
            exit ;
        end ;
        if( not Singleton ) then
        begin
            Lock_SCU ;
        end ;
        _File_System^.Erase( S ) ; { Delete S itself }
        _File_System^.Last_Error( Res ) ;
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        if( Res.Code <> 0 ) then { Deletion failure - exit to prevent infinite loop }
        begin
            _Delete := Res ;
            Uninitialize_Find_Record( _Find_Record ) ;
            List.Done ;
            exit ;
        end ;
	if( Logger <> nil ) then
        begin
            Logger.Log_Message( PChar( S ), 0, Abort ) ;
        end ;
        if( Abort ) then
        begin
            Uninitialize_Find_Record( _Find_Record ) ;
            List.Done ;
            exit ;
        end ;
    end ; { while( length( S ) > 0 )  }
    List.Done ;

    { Now delete specified data... }
    S := Files.Lookup( _System_Path + Path, _Find_Record ) ;
    if( length( S ) > 0 ) then
    begin
        S := copy( S, length( _System_Path ) + 1, length( S ) ) ; { Trim system path }
        if( not Singleton ) then
        begin
            Lock_SCU ;
        end ;
        _File_System^.Erase( S ) ;
        _File_System^.Last_Error( Res ) ;
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
	if( ( Logger <> nil ) and ( Res.Code = 0 ) ) then
        begin
            Logger.Log_Message( PChar( S ), 0, Abort ) ;
            if( Abort ) then
            begin
                Uninitialize_Find_Record( _Find_Record ) ;
                exit ;
            end ;
        end ;
    end else
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Value_Not_Found ;
    end ;
    Uninitialize_Find_Record( _Find_Record ) ;

    _Delete := Res ;
end ; { TSCU._Delete }


function TSCU._Lookup( Path : string ; DT : integer ) : string ;

var Key_Only : boolean ;
    S : string ;

label Try_Again ;

begin
    Key_Only := ( DT = SCUDT_Key ) ;
    if( ( length( Path ) > 0 ) and ( Path[ 1 ] <> '\' ) ) then
    begin
	Path := '\' + Path ;
    end ;
Try_Again:
    if( length( Path ) > 0 ) then
    begin
        S := Files.Lookup( _System_Path + Path, Find_Record ) ;
    end else
    begin
        S := Files.Lookup( '', Find_Record ) ;
    end ;
    S := copy( S, length( _System_Path ) + 1, length( S ) ) ; { Trim system path }
    if( ( length( S ) > 0 ) and ( DT <> SCUDT_All ) ) then
    begin
        if( ( ( Find_Record.Found_Attributes and FSFF_Directory ) <> 0 ) <> Key_Only ) then
        begin
            Path := '' ;
            goto Try_Again ; { Not a directory when we were looking for one, or vice versa }
        end ;
    end ;
    _Lookup := S ;
end ; { TSCU._Lookup }


function TSCU._Create( Path : string ; DT : longint ) : TUEC ;

var Dummy : integer ;
    F : PSFile ;
    Res : TUEC ;

begin
    if( copy( Path, 1, 1 ) <> '\' ) then
    begin
	Path := '\' + Path ;
    end ;
    if( DT = SCUDT_Key ) then { Creating a new key }
    begin
        Dummy := _File_System^.MkDir( Path ) ;
        if( Dummy <> 0 ) then
        begin
            _File_System^.Last_Error( Res ) ;
            if(
                ( Res.Facility = FSErr_Facility )
                and
                ( Res.Code = FSErr_FileExists )
              ) then
            begin
                Res.Facility := SCUErr_Facility ;
                Res.Code := SCUErr_Exists ;
            end ;
            _Create := Res ;
        end ;
        Res.Facility := SCUErr_Facility ;
        Res.Code := 0 ;
        exit ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    F := Open( _System_Path + Path, FSOM_Output, Res ) ;
    if( F = nil ) then
    begin
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        if( Res.Code = FSErr_FileExists ) then
        begin
            Res.Facility := SCUErr_Facility ;
            Res.Code := SCUErr_Exists ;
        end ;
        _Create := Res ;
        exit ;
    end ;
    F^.Blockwrite( DT, 2, Dummy ) ;
    F^.Close ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
    Res.Facility := SCUErr_Facility ;
    Res.Code := SCUErr_Success ;
    _Create := Res ;
end ; { TSCU._Create }


function TSCU._Rename( Old, New : string ) : TUEC ;

var Res : TUEC ;

begin
    if( _Key_Exists( New ) ) then
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Exists ;
        _Rename := Res ;
        exit ;
    end ;
    if( not _Key_Exists( Old ) ) then
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Value_Not_Found ;
        _Rename := Res ;
        exit ;
    end ;
    if( copy( New, 1, 1 ) <> '\' ) then
    begin
        New := '\' + New ;
    end ;
    if( copy( Old, 1, 1 ) <> '\' ) then
    begin
        Old := '\' + Old ;
    end ;
    if( not Check_Write( Old ) ) then
    begin
        exit ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    _File_System^.Rename( Old, New ) ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
    _File_System^.Last_Error( Res ) ;
    _Rename := Res ;
end ; { TSCU._Rename }


function TSCU._Create_And_Write( Path : string ; DT : longint ; Data : pointer ;
    Size : integer ) : TUEC ;

var Saved : boolean ;
    UEC : TUEC ;

begin
    UEC := _Create( Path, DT ) ;
    if( UEC.Code <> 0 ) then
    begin
        _Create_And_Write := UEC ;
        exit ;
    end ;
    Saved := Ignore_Alarms ;
    Ignore_Alarms := True ;
    _Create_And_Write := _Write( Path, Data, Size ) ;
    Ignore_Alarms := Saved ;
end ;


{ General API... }

function TSCU._Set_Value( User, Path : string ; Data : pointer ;
    Size : integer ) : TUEC ;

var Count : longint ;
    F : PSFile ;
    Res : TUEC ;

begin
    if( Size < 0 ) then
    begin
	Res.Facility := SCUErr_Facility ;
	Res.Code := SCUErr_Invalid_Size ;
	_Set_Value := Res ;
	exit ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    F := Open_Existing_File( User, Path ) ;
    if( not Check_Write( Strip_Root( Path ) ) ) then
    begin
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        F^.Close ;
        exit ;
    end ;
    if( F <> nil ) then
    begin
	F^.Seek( 2 ) ;
	F^.Blockwrite( Data^, Size, Count ) ;
	Res := F^.IO_Error ;
	F^.Close ;
    end else
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Value_Not_Found ;
    end ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
    _Set_Value := Res ;
end ; { TSCU._Set_Value }


function TSCU._Get_Value( User, Path : string ; Data : pointer ;
    Size : integer ) : TUEC ;

var Count : longint ;
    F : PSFile ;
    Res : TUEC ;

begin
    F := Open_Existing_File( User, Path ) ;
    if( not Check_Read( Strip_Root( Path ) ) ) then
    begin
        F^.Close ;
        exit ;
    end ;
    if( F <> nil ) then
    begin
	F^.Seek( 2 ) ;
	F^.Blockread( Data^, Size, Count ) ;
	Res := F^.IO_Error ;
	F^.CLose ;
    end else
    begin
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Value_Not_Found ;
    end ;
    _Get_Value := Res ;
end ; { TSCU._Get_Value }


function TSCU._Value_Size( User, Path : string ) : longint ;

var F : PSFile ;

begin
    F := Open_Existing_File( User, Path ) ;
    if( not Check_Read( Strip_Root( Path ) ) ) then
    begin
        F^.Close ;
        _Value_Size := -1 ;
        exit ;
    end ;
    if( F <> nil ) then
    begin
	_Value_Size := F^.Get_Size - 2 ;
	F^.Close ;
    end else
    begin
	_Value_Size := -1 ;
    end ;
end ; { TSCU._Value_Size }


function TSCU._Value_Type( User, Path : string ) : longint ;

var Count : longint ;
    Dummy : longint ;
    F : PSFile ;

begin
    _Value_Type := -1 ; { Default to invalid }
    F := Open_Existing_File( User, Path ) ;
    if( F <> nil ) then
    begin
        if( ( F^.Get_Attributes and FSFF_Directory ) <> 0 ) then
        begin
            Dummy := SCUDT_Key ;
        end else
        begin
            Dummy := 0 ;
            F^.Blockread( Dummy, 2, Count ) ;
            if( F^.IO_Error.Code <> 0 ) then
            begin
                Dummy := -1 ;
            end ;
        end ;
        _Value_Type := Dummy ;
	F^.Close ;
    end ;
end ; { TSCU._Value_Type }


function TSCU._Value_Exists( User, Path : string ) : boolean ;

var F : PSFile ;

begin
    F := Open_Existing_File( User, Path ) ;
    _Value_Exists := ( F <> nil ) ;
    if( F <> nil ) then
    begin
        F^.Close ;
    end ;
end ;


function TSCU._Delete_Value( User, Path : string ) : TUEC ;

var F : PSFile ;
    Res : TUEC ;
    S : string ;

begin
    User := Normalize_User( User ) ;
    if( copy( Path, 1, 1 ) = '\' ) then
    begin
        Path := copy( Path, 2, length( Path ) ) ;
    end ;
    if( length( User ) > 0 ) then
    begin
        S := _System_Path + '\Users\' + User + Path ;
        if( not Check_Delete( 'Users\' + User + Path ) ) then
        begin
            exit ;
        end ;
    end else
    begin
        S := _System_Path + '\' + Path ;
        if( not Check_Delete( Path ) ) then
        begin
            exit ;
        end ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    F := Open( S, FSOM_Input, Res ) ;
    if( Res.Code = 0 ) then
    begin
	F^.Close ;
        _File_System^.Erase( Path ) ;
        _File_System^.Last_Error( Res ) ;
    end ;
    _Delete_Value := Res ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
end ; { TSCU._Delete_Value }


function TSCU._Create_Value( User, Path : string ; DT : longint ) : TUEC ;

var Count : longint ;
    Dummy : integer ;
    F : PSFile ;
    Res : TUEC ;
    S : string ;

begin
    User := Normalize_User( User ) ;
    if( copy( Path, 1, 1 ) = '\' ) then
    begin
        Path := copy( Path, 2, length( Path ) ) ;
    end ;
    if( not Singleton ) then
    begin
        Lock_SCU ;
    end ;
    if( DT = SCUDT_Key ) then { Creating a new key }
    begin
        if( length( User ) = 0 ) then
        begin
            Dummy := _File_System^.MkDir( '\' + Path ) ;
        end else
        begin
            Dummy := _File_System^.MkDir( '\Users\' + User + Path ) ;
        end ;
        if( Dummy <> -1 ) then
        begin
            _File_System^.Last_Error( Res ) ;
        end else
        begin
            Res.Facility := SCUErr_Facility ;
            Res.Code := SCUErr_Success ;
        end ;
        _Create_Value := Res ;
        if( not Singleton ) then
        begin
            Unlock_SCU ;
        end ;
        exit ;
    end ;
    if( length( User ) = 0 ) then
    begin
        S := _System_Path + '\' + Path ;
    end else
    begin
        S := _System_Path + '\Users\' + User + Path ;
    end ;
    F := Open( S, FSOM_Output, Res ) ;
    if( Res.Code = 0 ) then
    begin
	F^.Blockwrite( DT, 2, Count ) ;
	Res := F^.IO_Error ;
	F^.Close ;
    end else
    begin
        if( Res.Code = FSErr_FileExists ) then
        begin
            Res.Facility := SCUErr_Facility ;
            Res.Code := SCUErr_Exists ;
        end ;
    end ;
    if( not Singleton ) then
    begin
        Unlock_SCU ;
    end ;
    _Create_Value := Res ;
end ; { TSCU._Create_Value }


var Read_Value_Result : string ;
    Terminal_NULL : char ;

function TSCU._Read_Value( User, Path : string ; var DT, Size : longint ;
    var Buffer : PChar ) : TUEC ;
{ Reads the specified value.  If result indicates a success, DT is the type of
  the data, Size is the actual data size, and Buffer is a pointer to a buffer
  containing the data.  DT, Size, and Buffer are ignored on call, and set on
  return.

  This function serves as a single call to: 1) determine if the value exists,
  2) determine what type it is, and 3) read the value's data. }

var Count, Dummy : integer ;
    F : PSFile ;
    Res : TUEC ;

begin
    F := Open_Existing_File( User, Path ) ;
    if( F <> nil ) then
    begin
        if( ( F^.Get_Attributes and FSFF_Directory ) <> 0 ) then
        begin
            Dummy := SCUDT_Key ;
            Size := 0 ;
            fillchar( Res, sizeof( Res ), 0 ) ;
        end else
        begin
            Dummy := 0 ;
            F^.Blockread( Dummy, 2, Count ) ;
            Res := F^.IO_Error ;
            if( Res.Code <> 0 ) then
            begin
                Dummy := SCUDT_Undefined ;
                Size := 0 ;
                Buffer := nil ;
            end else
            begin
                Size := F^.Get_Size - 2 ;
{$IFDEF DELPHI}
                SetLength( Read_Value_Result, Size ) ;
                Buffer := PChar( Read_Value_Result ) ;
{$ELSE}
                if( Size > 255 ) then
                begin
                    Size := 255 ;
                end ;
                if( Size < 255 ) then
                begin
                    Buffer[ 255 ] := #0 ;
                end ;
                Buffer := @Read_Value_Result[ 1 ] ;
{$ENDIF}
                F^.Blockread( Buffer^, Size, Count ) ;
                Size := Count ; { Actual amount read }
                Res := F^.IO_Error ;
            end ;
        end ; { if( ( F^.Get_Attributes and FSFF_Directory ) <> 0 ) }
        DT := Dummy ;
        F^.Close ;
    end else
    begin
        DT := SCUDT_Undefined ;
        Size := -1 ;
        Buffer := nil ;
        Res.Facility := SCUErr_Facility ;
        Res.Code := SCUErr_Value_Not_Found ;
    end ; { if( F <> nil ) }
    _Read_Value := Res ;
end ; { TSCU._Read_Value }


{ Standard API... }

function TSCU.Init( Name : PChar ) : TUEC ;

var X : string ;

begin
    X := Name ;
    Init := _Init( X, -1 ) ;
end ;


function TSCU.Write( Path : PChar ; Data : pointer ;
    Size : integer ) : TUEC ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Write := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    Write := _Write( X, Data, Size ) ;
end ;


function TSCU.Read( Path : PChar ; Data : pointer ;
    Size : integer ) : TUEC ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Read := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    Read := _Read( Path, Data, Size ) ;
end ;


function TSCU.Read_Size( Path : PChar ) : integer ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Read_Size := -1 ;
        exit ;
    end ;
    X := Path ;
    Read_Size := _Read_Size( X ) ;
end ;


function TSCU.Read_Type( Path : PChar ) : integer ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Read_Type := -1 ;
        exit ;
    end ;
    X := Path ;
    Read_Type := _Read_Type( X ) ;
end ;


function TSCU.Exists( Path : PChar ) : boolean ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Exists := False ;
        exit ;
    end ;
    X := Path ;
    Exists := _Exists( X ) ;
end ;


function TSCU.Key_Exists( Path : PChar ) : boolean ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Key_Exists := False ;
        exit ;
    end ;
    X := Path ;
    Key_Exists := _Key_Exists( X ) ;
end ;


function TSCU.Lookup( Path : PChar ; DT : integer ) : PChar ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Temp_Lookup := '' ;
    end else
    begin
        X := Path ;
        Temp_Lookup := _Lookup( X, DT ) ;
    end ;
    Lookup := PChar( Temp_Lookup ) ;
end ;


function TSCU.Delete( Path : PChar ) : TUEC ;

var B : boolean ;
    X : string ;

begin
    if( _File_System = nil ) then
    begin
        Delete := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    B := False ;
    Delete := _Delete( X, nil, B ) ;
end ;


function TSCU.Create( Path : PChar ; DT : longint ) : TUEC ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Create := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    Create := _Create( X, DT ) ;
end ;


function TSCU.Create_And_Write( Path : PChar ; DT : longint ;
    Data : pointer ; Size : integer ) : TUEC ;

var X : string ;

begin
    if( _File_System = nil ) then
    begin
        Create_And_Write := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    Create_And_Write := _Create_And_Write( X, DT, Data, Size ) ;
end ;


function TSCU.Rename( Old, New : PChar ) : TUEC ;

begin
    if( _File_System = nil ) then
    begin
        Rename := Startup_Error ;
        exit ;
    end ;
    Rename := _Rename( Old, New ) ;
end ;


function TSCU.Set_Value( User, Path : PChar ; Data : pointer ;
    Size : integer ) : TUEC ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Set_Value := Startup_Error ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Set_Value := _Set_Value( S_User, S_Path, Data, Size ) ;
end ;


function TSCU.Get_Value( User, Path : PChar ; Data : pointer ;
    Size : integer ) : TUEC ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Get_Value := Startup_Error ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Get_Value := _Get_Value( S_User, S_Path, Data, Size ) ;
end ;


function TSCU.Value_Size( User, Path : PChar ) : longint ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Value_Size := -1 ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Value_Size := _Value_Size( S_User, S_Path ) ;
end ;


function TSCU.Value_Type( User, Path : PChar ) : longint ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Value_Type := -1 ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Value_Type := _Value_Type( S_User, S_Path ) ;
end ;


function TSCU.Value_Exists( User, Path : PChar ) : boolean ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Value_Exists := False ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Value_Exists := _Value_Exists( S_User, S_Path ) ;
end ;


function TSCU.Delete_Value( User, Path : PChar ) : TUEC ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Delete_Value := Startup_Error ;
        exit ;
    end ;
    S_User := User ;
    S_Path := Path ;
    Delete_Value := _Delete_Value( S_User, S_Path ) ;
end ;


function TSCU.Create_Value( User, Path : PChar ; DT : longint ) : TUEC ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Create_Value := Startup_Error ;
        exit ;
    end ;
    S_User := User ; { This extra step is needed to work around a Delphi bug }
    S_Path := Path ;
    Create_Value := _Create_Value( S_User, S_Path, DT ) ;
end ;


function TSCU.Read_Value( User, Path : PChar ; var DT, Size : longint ;
    var Buffer : PChar ) : TUEC ;

var S_User, S_Path : string ;

begin
    if( _File_System = nil ) then
    begin
        Read_Value := Startup_Error ;
        exit ;
    end ;
    S_User := User ; { This extra step is needed to work around a Delphi bug }
    S_Path := Path ;
    Read_Value := _Read_Value( User, Path, DT, Size, Buffer ) ;
end ;


function TSCU.Verify : TUEC ;

begin
    if( _File_System = nil ) then
    begin
        Verify := Startup_Error ;
        exit ;
    end ;
    _File_System^.Verify( _System_Path, Result ) ;
end ;


function TSCU.Debugger : PDebug_Interface ;

begin
    Debugger := nil ;
end ;


function TSCU.Init_And_Extend( Name : PChar ; Size : longint ) : TUEC ;

begin
    Init_And_Extend := _Init( Name, Size ) ;
end ;


function TSCU.Delete_With_Callback( Path : PChar ; Logger : PSLogger ) : TUEC ;

var B : boolean ;
    X : string ;

begin
    if( _File_System = nil ) then
    begin
        Delete_With_Callback := Startup_Error ;
        exit ;
    end ;
    X := Path ;
    B := False ;
    Delete_With_Callback := _Delete( X, Logger, B ) ;
end ;


procedure TSCU.Shutdown ;

begin
    // This routine left intentionally blank.
end ;


{ V1.2 interface: }

{ Contextual version of instance support... }

function TSCU.Init_Context( Name : PChar ; Flags : longint ) : longint ;

begin
    _Contextual := True ;
    Init( Name ) ;
    Init_Context := longint( _File_System ) ;
    if( ( Flags and 1 ) <> 0 ) then
    begin
        _File_System.Set_Cache( -1 ) ;
    end ;
end ;


procedure TSCU.Terminate_Context( Context : longint ) ;

begin
    Close_File_System( PXFile_System( Context )^.Path ) ;
end ;


function TSCU.Init_And_Extend_Context( Name : PChar ;
    Size : longint ; Flags : longint ) : longint ;

begin
    _Contextual := True ;
    Init_And_Extend( Name, Size ) ;
    Init_And_Extend_Context := longint( _File_System ) ;
    if( ( Flags and 1 ) <> 0 ) then
    begin
        _File_System.Set_Cache( -1 ) ;
    end ;
end ;


{ Contextual version of Low-level API... }

function TSCU.Write_Context( Context : longint ; Path : PChar ; Data : pointer ;
    Size : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Write_Context := Write( Path, Data, Size ) ;
end ;


function TSCU.Read_Context( Context : longint ; Path : PChar ; Data : pointer ;
    Size : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Read_Context := Read( Path, Data, Size ) ;
end ;


function TSCU.Read_Size_Context( Context : longint ; Path : PChar ) : longint ;

begin
    _File_System := PXFile_System( Context ) ;
    Read_Size_Context := Read_Size( Path ) ;
end ;

function TSCU.Read_Type_Context( Context : longint ; Path : PChar ) : longint ;

begin
    _File_System := PXFile_System( Context ) ;
    Read_Type_Context := Read_Type( Path ) ;
end ;


function TSCU.Exists_Context( Context : longint ; Path : PChar ) : boolean ;

begin
    _File_System := PXFile_System( Context ) ;
    Exists_Context := Exists( Path ) ;
end ;


function TSCU.Key_Exists_Context( Context : longint ; Path : PChar ) : boolean ;

begin
    _File_System := PXFile_System( Context ) ;
    Key_Exists_Context := Key_Exists( Path ) ;
end ;


function TSCU.Lookup_Context( Context : longint ; Path : PChar ;
    DT : longint ) : PChar ;

begin
    _File_System := PXFile_System( Context ) ;
    Lookup_Context := Lookup( Path, DT ) ;
end ;


function TSCU.Delete_Context( Context : longint ; Path : PChar ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Delete_Context := Delete( Path ) ;
end ;


function TSCU.Create_Context( Context : longint ; Path : PChar ;
    DT : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Create_Context := Create( Path, DT ) ;
end ;


function TSCU.Create_And_Write_Context( Context : longint ; Path : PChar ;
    DT : longint ; Data : pointer ; Size : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Create_And_Write_Context := Create_And_Write( Path, DT, Data, Size ) ;
end ;


function TSCU.Rename_Context( Context : longint ; Old, New : PChar ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Rename_Context := Rename( Old, New ) ;
end ;


{ Contextual version of Standard API... }

function TSCU.Set_Value_Context( Context : longint ; User, Path : PChar ;
    Data : pointer ; Size : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Set_Value_Context := Set_Value( User, Path, Data, Size ) ;
end ;


function TSCU.Get_Value_Context( Context : longint ; User, Path : PChar ;
    Data : pointer ; Size : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Get_Value_Context := Get_Value( User, Path, Data, Size ) ;
end ;


function TSCU.Value_Size_Context( Context : longint ;
    User, Path : PChar ) : longint ;

begin
    _File_System := PXFile_System( Context ) ;
    Value_Size_Context := Value_Size( User, Path ) ;
end ;


function TSCU.Value_Type_Context( Context : longint ;
    User, Path : PChar ) : longint ;

begin
    _File_System := PXFile_System( Context ) ;
    Value_Type_Context := Value_Type( User, Path ) ;
end ;


function TSCU.Value_Exists_Context( Context : longint ;
    User, Path : PChar ) : boolean ;

begin
    _File_System := PXFile_System( Context ) ;
    Value_Exists_Context := Value_Exists( User, Path ) ;
end ;


function TSCU.Delete_Value_Context( Context : longint ;
    User, Path : PChar ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Delete_Value_Context := Delete_Value( User, Path ) ;
end ;


function TSCU.Create_Value_Context( Context : longint ; User, Path : PChar ;
    DT : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Create_Value_Context := Create_Value( User, Path, DT ) ;
end ;


function TSCU.Verify_Context( Context : longint ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    _File_System.Verify( _File_System^.Path, Result ) ;
end ;


function TSCU.Debugger_Context( Context : longint ) : PDebug_Interface ;

begin
    _File_System := PXFile_System( Context ) ;
    Debugger_Context := Delphi_Debug_Wrapper( _File_System.Debugger ) ;
end ;


function TSCU.Read_Value_Context( Context : longint ; User, Path : PChar ;
    var DT, Size : longint ; var Buffer : PChar) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Read_Value_Context := Read_Value( User, Path, DT, Size, Buffer ) ;
end ;


function TSCU.Delete_With_Callback_Context( Context : longint ;
    Path : PChar ; Logger : PSLogger ) : TUEC ;

begin
    _File_System := PXFile_System( Context ) ;
    Delete_With_Callback_Context := Delete_With_Callback( Path, Logger ) ;
end ;


{$IFDEF DELPHI}
type tSCU_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ;
                                    override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                  end ;
     pSCU_EI = tSCU_EI ;
{$ELSE}
type tSCU_EI = object( TSStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    virtual ;
                                function Facility_Version : longint ; virtual ;
                                function Version : longint ; virtual ;
                  end ;
     pSCU_EI = ^tSCU_EI ;
{$ENDIF}

{ tSCU_EI method... }

function tSCU_EI.Get_Facility : longint ;

begin
    Get_Facility := SCUErr_Facility ;
end ;


const SCU_Facility_Name : PChar = 'SCU' ;

function tSCU_EI.Get_Name : PChar ;

begin
    Get_Name := SCU_Facility_Name ;
end ;


var SCU_Error_Text : array[ 0..127 ] of char ;

function tSCU_EI.Translate_Error( Code : longint ) : PChar ;

var SCU_Error : string ;

begin
    case Code of
	SCUErr_Success : SCU_Error := 'Success' ;
        SCUErr_Value_Not_Found : SCU_Error := 'Value not found' ;
        SCUErr_Invalid_Size : SCU_Error := 'Invalid size' ;
        SCUErr_Exists : SCU_Error := 'Item exists' ;
        SCUErr_Alarm_Failure : SCU_Error := 'Alarm failure' ;
        SCUErr_Bad_Data : SCU_Error := 'Bad data format' ;
	else SCU_Error := 'Unknown error number ' + num1( Code ) ;
    end ;
    SCU_Error := SCU_Error + #0 ;
    move( SCU_Error[ 1 ], SCU_Error_Text[ 0 ], length( SCU_Error ) ) ;
    Translate_Error := @SCU_Error_Text ;
end ; { tSCU_EI.Translate_Error }


{$IFDEF DELPHI}
procedure tSCU_EI.Done ;

begin
    { Done by unit }
end ;
{$ENDIF}


function tSCU_EI.Facility_Version : longint ;

begin
    Facility_Version := 1 ;
end ;


function tSCU_EI.Version : longint ;

begin
    Version := 10 ; { Compliant with UEH V1.0 specification }
end ;


{$IFDEF DELPHI}
var SCUEI : pSCU_EI = nil ;
{$ELSE}
var SCUEI : pSCU_EI ;
{$ENDIF}

procedure Unit_Initialization ;

begin
    Terminal_NULL := #0 ;
{$IFDEF DELPHI}
    SCUEI := tSCU_EI.Create ;
    Register_Error_Interface( PSError_Interface( SCUEI ) ) ;
{$ELSE}
    new( SCUEI, Init ) ;
    Register_Error_Interface( SCUEI ) ;
{$ENDIF}
end ;


procedure Unit_Deallocation ;

begin
{$IFDEF DELPHI}
    Unregister_Error_Interface( SCUEI.Get_Facility ) ;
    SCUEI.Free ;
    SCUEI := nil ;
{$ELSE}
    Unregister_Error_Interface( SCUEI^.Get_Facility ) ;
    dispose( SCUEI ) ;
    SCUEI := nil ;
{$ENDIF}
end ; { Unit_Deallocation }


{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialization ;
{$IFDEF DELPHI}

finalization
    Unit_Deallocation ;
{$ENDIF}
end.
