/* 
              Program Name : TSRS
              Package Name : LB
              Purpose      : TSR support definitions
              Institution  : Conroy & Conroy Co.
              Date Written : 21-Jan-1992
              Written By   : Alan Conroy
              Version      : 2
              Revision     : 1
              Edit         :

	TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

	  These definitions are for support of TSR applications.

        *********************************************************
        *                                                       *
        *            F I L E   D E S C R I P T I O N S          *
        *                                                       *
        *********************************************************

 */

/* Standard type declarations */

/* FCB (File Control Block) for V1.n */
typdef struct FCBV1
{
    unsigned char Extended ; /* FFH if extended, offset -7 */
    unsigned char Filler[ 5 ] ;
    unsigned char Extended_Attribute ; /* Offset -1 */
    unsigned char Drive_Number ; /* Offset 0 */
    char File_Name[ 11 ] ;
    short int Current_Block ;
    short int Record_Size ;
    long File_Size ;
    short int Date ;
    short int Time ;
    unsigned char Flags ;
    short int Absolute_Cluster ;
    short int Starting_Cluster ;
    short int Relative_Cluster ;
    unsigned char Filler1 ;
    unsigned char Current_Record ;
    long Record_Number ;
} ;

/* FCB (File Control Block) for V2.n */
typedef struct FCBV2
{
    unsigned char Extended ; /* FFH if extended, offset -7 */
    unsigned char Filler[ 5 ] ;
    unsigned char Extended_Attribute ; /* Offset -1 */
    unsigned char Drive_Number ; /* Offset 0 */
    char File_Name[ 11 ] ;
    short int Current_Block ;
    short int Record_Size ;
    long File_Size ;
    short int Date ;
    short int Time ;
    unsigned char Flags ;
    short int Starting_Cluster ;
    unsigned char Filler1[ 5 ] ;
    unsigned char Current_Record ;
    long Record_Number ;
} ;

/* FCB (File Control Block) for V3.n, SHARE loaded, local file */
typedef struct FCB_Local
{
    unsigned char Extended ; /* FFH if extended, offset -7 */
    unsigned char Filler[ 5 ] ;
    unsigned char Extended_Attribute ; /* Offset -1 */
    unsigned char Drive_Number ; /* Offset 0 */
    char File_Name[ 11 ] ;
    short int Current_Block ;
    short int Record_Size ;
    long File_Size ;
    short int Date ;
    short int Time ;
    unsigned char SFT_Number ;
    unsigned char Flags ;
    short int Starting_Cluster ;
    short int SHARE_Offset ;
    unsigned char Attributes ;
    unsigned char Filler1 ;
    unsigned char Current_Record ;
    long Record_Number ;
} ;

/* FCB (File Control Block) for V3.n, SHARE loaded, remote file */
typedef struct FCB_Remote
{
    unsigned char Extended ; /* FFH if extended, offset -7 */
    unsigned char Filler[ 5 ] ;
    unsigned char Extended_Attribute ; /* Offset -1 */
    unsigned char Drive_Number ; /* Offset 0 */
    char File_Name[ 11 ] ;
    short int Current_Block ;
    short int Record_Size ;
    long File_Size ;
    short int Date ;
    short int Time ;
    short int Directory_Sector ;
    short int Relative_Cluster ;
    short int Absolute_Cluster ;
    unsigned char Current_Record ;
    long Record_Number ;
} ;

/* FCB (File Control Block) for V3.n, SHARE not loaded */
typedef struct FCBV3
{
    unsigned char Extended ; /* FFH if extended, offset -7 */
    unsigned char Filler[ 5 ] ;
    unsigned char Extended_Attribute ; /* Offset -1 */
    unsigned char Drive_Number ; /* Offset 0 */
    char File_Name[ 11 ] ;
    short int Current_Block ;
    short int Record_Size ;
    long File_Size ;
    short int Date ;
    short int Time ;
    unsigned char Flags ;
    short int Starting_Cluster ;
    short int Directory_Sector ;
    unsigned char Directory_Entry ;
    unsigned char Current_Record ;
    long Record_Number ;
} ;

/* PSP (Program Segment Prefix */
typedef struct *tPSP pPSP ;
typedef struct tPSP
{
    short int INT_20H ;
    short int Size ;
    unsigned char Filler ;
    unsigned char FAR_Jump[ 5 ] ;
    long INT_22H ;
    long INT_23H ;
    long INT_24H ;
    short int Parent ;
    unsigned char Files[ 20 ] ;
    short int Environment ;
    long SS_SP ;
    short int Max_Files ;
    void *File_Table ;
    void *Previous_PSP ;
    unsigned char Filler1[ 20 ] ;
    unsigned char Dispatch[ 3 ] ;
    unsigned char Filler2[ 9 ] ;
    unsigned char FCB1[ 16 ] ;
    unsigned char FCB2[ 20 ] ;
    char Command_Tail[ 128 ] ;
} ;

/* DPB (Drive Parameter Block) For DOS V1.n */
typedef struct DPBV1
{
    unsigned char Drive ;
    unsigned char Unit_Number ;
    short int Bytes_Per_Sector ;
    unsigned char Largest_Sector ;
    unsigned char Shift_Count ;
    short int Reserved_Sectors ;
    unsigned char FAT_Copies ;
    short int Root_Directory_Entries ;
    short int First_Data_Sector ;
    short int Max_Cluster ;
} ;

/* DPB (Drive Parameter Block) For DOS V2.n */
typedef struct DPBV2
{
    unsigned char Drive ;
    unsigned char Unit_Number ;
    short int Bytes_Per_Sector ;
    unsigned char Largest_Sector ;
    unsigned char Shift_Count ;
    short int Reserved_Sectors ;
    unsigned char FAT_Copies ;
    short int Root_Directory_Entries ;
    short int First_Data_Sector ;
    short int Max_Cluster ;
    unsigned char FAT_Sectors ;
    short int Root_Sector ;
    void *Driver ;
    unsigned char Media_ID ;
    unsigned char Rebuild ;
    void *Next_Driver ;
    short int Current_Directory_Cluster ;
    char Current_Path[ 64 ] ;
} ;

/* DPB (Drive Parameter Block) For DOS V3.n */
typedef struct DPBV3
{
    unsigned char Drive ;
    unsigned char Unit_Number ;
    short int Bytes_Per_Sector ;
    unsigned char Largest_Sector ;
    unsigned char Shift_Count ;
    short int Reserved_Sectors ;
    unsigned char FAT_Copies ;
    short int Root_Directory_Entries ;
    short int First_Data_Sector ;
    short int Max_Cluster ;
    unsigned char FAT_Sectors ;
    short int Root_Sector ;
    void *Driver ;
    unsigned char Media_ID ;
    unsigned char Rebuild ;
    void *Next_Driver ;
    short int Search_Start ;
    short int Free_Clusters 
} ;

/* DPB (Drive Parameter Block) For DOS V4.n */
typedef struct DPBV4
{
    unsigned char Drive ;
    unsigned char Unit_Number ;
    short int Bytes_Per_Sector ;
    unsigned char Largest_Sector ;
    unsigned char Shift_Count ;
    short int Reserved_Sectors ;
    unsigned char FAT_Copies ;
    short int Root_Directory_Entries ;
    short int First_Data_Sector ;
    short int Max_Cluster ;
    short int FAT_Sectors ;
    short int Root_Sector ;
    void *Driver ;
    unsigned char Media_ID ;
    unsigned char Rebuild ;
    void *Next_Driver ;
    short int Search_Start ;
    short int Free_Clusters ;
} ;

/* BPB (BIOS Parameter Block), DOS V2.n */
/* As it appears on disks without 8 sectors/track, in sector 0 */
typedef struct BPBV2
{
    unsigned char Filler[ 3 ] ;
    char System_ID[ 8 ] ;
    short int Bytes_Per_Sector ;
    unsigned char Sectors_Per_Cluster ;
    short int Reserved_Sectors ;
    unsigned char FAT_Count ;
    short int Root_Directory_Entries ;
    short int Total_Sectors ;
    unsigned char Media_ID ;
    short int FAT_Sectors ;
    short int Sectors_Per_Track ;
    short int Heads ;
    long Hidden_Sectors ;
} ;

/* BPB (BIOS Parameter Block), DOS V3.n */
/* As it appears on disks without 8 sectors/track, in sector 0 */
typedef struct BPBV3
{
    unsigned char Filler[ 3 ] ;
    char System_ID[ 8 ] ;
    short int Bytes_Per_Sector ;
    unsigned char Sectors_Per_Cluster ;
    short int Reserved_Sectors ;
    unsigned char FAT_Count ;
    short int Root_Directory_Entries ;
    short int Total_Sectors ;
    unsigned char Media_ID ;
    short int FAT_Sectors ;
    short int Sectors_Per_Track ;
    short int Heads ;
    long Hidden_Sectors 
} ;

/* BPB (BIOS Parameter Block), DOS V4.n */
typedef struct BPBV4
{
    unsigned char Filler0[ 3 ] ;
    char System_ID[ 8 ] ;

    short int Bytes_Per_Sector ;
    unsigned char Sectors_Per_Cluster ;
    short int Reserved_Sectors ;
    unsigned char FAT_Count ;
    short int Root_Directory_Entries ;
    short int Total_Sectors ;
    unsigned char Media_ID ;
    short int FAT_Sectors ;
    long Extended_Total_Sectors ;
    unsigned char Filler[ 6 ] ;
    short int Cylinders ;
    unsigned char Device_Type ;
    short int Attributes ;
} ;

/* BPB (BIOS Parameter Block), Special - for DDT entries */
typedef struct BPB
{
    short int Bytes_Per_Sector ;
    unsigned char Sectors_Per_Cluster ;
    short int Reserved_Sectors ;
    unsigned char FAT_Count ;
    short int Root_Directory_Entries ;
    short int Total_Sectors ;
    unsigned char Media_ID ;
    short int FAT_Sectors ;
    short int Sectors_Per_Track ;
    short int Heads ;
    short int Hidden_Sectors ;
} ;

/* DDT (Drive Data Table), DOS V3.3 */
typedef struct DDTV3
{
    void *Next ;
    unsigned char Unit_Number ;
    unsigned char Drive_Number ;
    BIOS_PB : BPB;
    unsigned char Filler ;
    short int Open ;
    char Volume_Label[ 11 ] ;
    unsigned char Null ;
    unsigned char Device_Type ;
    short int Flags ;
    short int Cylinders ;
    Max_Capacity : BPB;
    unsigned char Filler1[ 3 ] ;
    char File_System[ 9 ] ;
    unsigned char Null1 ;
    long Last_Access ;
} ;

/* DDT (Drive Data Table), DOS V4.n */
typedef struct DDTV4
{
    void *Next ;
    unsigned char Unit_Number ;
    unsigned char Drive_Number ;
    BIOS_PB : BPB;
    unsigned char Filler[ 9 ] ;
    unsigned char Device_Type ;
    short int Flags ;
    short int Cylinders ;
    Max_Capacity : BPB;
    unsigned char Filler1[ 13 ] ;
    long Last_Access ;
    char Volume_Label[ 11 ] ;
    unsigned char Null ;
    long Serial_Number ;
    char File_System[ 8 ] ;
    unsigned char Null1 ;
} ;

/* DPL (DOS Parameter List) */
typedef union DPL
{
    struct
    {
	short int AX ;
	short int BX ;
	short int CX ;
	short int DX ;
	short int SI ;
	short int DI ;
	short int DS ;
	short int ES ;
	short int Filler ;
	short int ID ;
	short int Process ;
    } ;
    struct
    {
	unsigned char AL ;
	unsigned char AH ;
	unsigned char BL ;
	unsigned char BH ;
	unsigned char CL ;
	unsigned char CH ;
	unsigned char DL ;
	unsigned char DH ;
    } ;
} ;

/* Disk Serial Number Record */
typedef struct SN_Record
{
	    short int Level ;
	    long Serial_Number ;
	    char Volume_Label[ 11 ] ;
	    char File_System[ 8 ] 
} ;

/* KEYB internal data, DOS V3.3 */
typedef struct KEYBV3
{
    void *INT_09 ;
    void *INT_2F ;
    unsigned char Filler[ 6 ] ;
    short int Flags ;
    unsigned char Filler1[ 6 ] ;
    char Country_ID[ 2 ] ;
    short int Current_Code_Page ;
    short int First ;
    short int Ptr ;
    unsigned char Filler2[ 2 ] ;
    short int Xlate ;
    short int Last ;
    unsigned char Filler3[ 9 ] 
} ;

/* KEYB internal data, DOS V4.n */
typedef struct KEYBV4
{
    void *INT_09 ;
    void *INT_2F ;
    unsigned char Filler[ 6 ] ;
    short int Flags ;
    unsigned char Filler1[ 6 ] ;
    char Country_ID[ 2 ] ;
    short int Current_Code_Page ;
    unsigned char Filler2[ 2 ] ;
    short int First ;
    short int Ptr ;
    unsigned char Filler3[ 2 ] ;
    short int Xlate ;
    short int Last ;
    unsigned char Filler4[ 9 ]  ;
} ;

/* EXEC Parameter Block */
typedef struct EPB
{
    short int Environment ;
    void *Command_Line ;
    void *FCB1 ;
    void *FCB2 ;
    long Initial_SS_SP ;
    long Entry_Pointer 
} ;

/* SDB (Search Data Block) */
typedef struct SDB
{
    unsigned char Drive_Letter ;
    char Search_Template[ 11 ] ;
    unsigned char Search_Attributes ;
    short int Directory_Entry ;
    short int Parent_Cluster ;
    unsigned char Filler[ 5 ] ;
    unsigned char Found_Attribute ;
    short int Time ;
    short int Date ;
    long Size ;
    char Name[ 13 ]  ;
} ;

/* List of Lists (LOL), DOS V2.n */
typedef struct LOLV2
{
    unsigned char Filler[ 14 ] ; /* Offset -16 */
    short int MCB ; /* Offset -2 */
    void *DPB ; /* Offset 0 */
    void *SFT ;
    void *CLOCK_Driver ;
    void *CON_Driver ;
    unsigned char Last_Drive ; /* Offset 10H */
    short int Max_Bytes_Per_Block ;
    void *First_Disk_Buffer ;
    unsigned char NUL_Driver_Header[ 18 ]  ;
} ;

/* List of Lists (LOL), DOS V3.0 */
typedef struct LOLV30
{
    unsigned char Filler[ 4 ] ; /* Offset -16 */
    short int Retry_Count ; /* Offset -12 */
    short int Retry_Delay ; /* Offset -10 */
    void *Current_Disk_Buffer ; /* Offset -8 */
    short int CON_Input ; /* Offset -4 */
    short int MCB ; /* Offset -2 */
    void *DPB ; /* Offset 0 */
    void *SFT ;
    void *CLOCK_Driver ;
    void *CON_Driver ;
    unsigned char Device_Count ; /* Offset 10H */
    short int Max_Bytes_Per_Block ;
    void *First_Disk_Buffer ;
    void *CDS ;
    unsigned char Last_Drive ;
    void *STRING_Workspace ;
    short int STRING_Size ;
    void *FCB ;
    short int Protected_FCBs ;
    unsigned char NUL_Driver_Header[ 18 ] 
} ;

/* List of Lists (LOL), DOS V3.1+ */
typedef struct LOLV3
{
    unsigned char Filler[ 4 ] ; /* Offset -16 */
    short int Retry_Count ; /* Offset -12 */
    short int Retry_Delay ; /* Offset -10 */
    void *Current_Disk_Buffer ; /* Offset -8 */
    short int CON_Input ; /* Offset -4 */
    short int MCB ; /* Offset -2 */
    void *DPB ; /* Offset 0 */
    void *SFT ;
    void *CLOCK_Driver ;
    void *CON_Driver ;
    short int Max_Bytes_Per_Block ; /* Offset 10H */
    void *First_Disk_Buffer ;
    void *CDS ;
    void *FCB ;
    short int Protected_FCBs ;
    unsigned char Device_Count ;
    unsigned char Last_Drive ;
    unsigned char NUL_Driver_Header[ 18 ] ;
    unsigned char JOIN_Count ;
} ;

/* List of Lists (LOL), DOS V4.n */
typedef struct LOLV4
{
    unsigned char Filler[ 14 ] ; /* Offset -16 */
    short int MCB ; /* Offset -2 */
    void *DPB ; /* Offset 0 */
    void *SFT ;
    void *CLOCK_Driver ;
    void *CON_Driver ;
    short int Max_Bytes_Per_Block ; /* Offset 10H */
    void *Disk_Buffer ;
    void *CDS ;
    void *FCB ;
    short int Protected_FCBS ;
    unsigned char Device_Count ;
    unsigned char Last_Drive ;
    unsigned char NUL_Driver_Header[ 18 ] ;
    unsigned char JOIN_Count ;
    short int Program_Names ;
    void *IFS_Resident ;
    void *IFS_Chain ;
    short int Buffers_X ;
    short int Buffers_Y ;
    unsigned char Boot_Drive ;
    unsigned char Filler1 ;
    short int Extended_RAM_Size 
} ;

/* MCB (Memory Control Block) */
typedef struct MCB
{
    unsigned char Block_Type ;
    short int Owner ;
    short int Size ;
    unsigned char Filler[ 3 ] ;
    char Name[ 8 ]  ;
} ;

/* SCB (Subsegment Control Block) */
typedef struct SCB
{
    char Subsegment_Type ;
    short int Start ;
    short int Size ;
    unsigned char Filler[ 3 ] ;
    char Name[ 8 ] 
} ;

/* STACKS Code Segment */
typedef struct SCS
{
    short int Filler ;
    short int Stacks ;
    short int SCB_Size ;
    short int Stack_Size ;
    void *Data_Segment ;
    short int SCB ;
    short int Last ;
    short int Next ;
} ;

/* STACKS array elements */
typedef struct SAE
{
    unsigned char Status ;
    unsigned char Filler ;
    short int Previous_SP ;
    short int Previous_SS ;
    short int Top ;
} ;

/* Sharing Record */
typedef struct Sharing_Record
{
    unsigned char Flag ;
    short int Size ;
    unsigned char Checksum ;
    short int Lock_Record ;
    void *SFT ;
    short int Sequence ;
    char Pathname ; /* First byte of variable area */
} ;

/* Lock record */
typedef struct Lock_Record
{
    short int Next ;
    long Start ;
    long Ending ;
    void *SFT ;
    short int Owner 
} ;

/* SFT (System File Table), DOS V2.n */
typedef union SFTV2
{
	struct
	{
		unsigned char Handle_Count ;
		unsigned char Open_Mode ;
		unsigned char Attributes ;
		unsigned char Drive ;
		char Name[ 11 ] ;
		short int Filler[ 2 ] ;
		long Size ;
		short int Date ;
		short int Time ;
		unsigned char Device_Attribute ;
		void *Driver ;
		short int Absolute_Cluster ;
		short int Filler1 ;
		long Current_Position ;
	} ;
	struct
	{
		unsigned char Filler2[ 28 ] ;
		short int Start ;
		short int Relative_Cluster ;
	} ;
} ;

/* SFT (System File Table), DOS V3.n */
typedef struct SFTV3
{
    short int Handle_Count ;
    short int Open_Mode ;
    unsigned char Attributes ;
    short int Device_Info ;
    void *Driver ;
    short int Start_Cluster ;
    short int Time ;
    short int Date ;
    long Size ;
    long Current_Position ;
    short int Relative_Cluster ;
    short int Absolute_Cluster ;
    short int Dir_Sector ;
    unsigned char Directory_Entry ;
    char FCB_FN[ 11 ] ;
    void *Previous_SFT ;
    short int Owner ;
    short int Sharing_Record 
} ;

/* SFT (System File Table), DOS V4.n */
typedef union SFTV4
{
	struct
	{
		short int Handle_Count ;
		short int Open_Mode ;
		unsigned char Attributes ;
		short int Device_Info ;
		void *Driver ;
		short int Start_Cluster ;
		short int Time ;
		short int Date ;
		long Size ;
		void *REDIRIFS ;
		unsigned char Filler[ 3 ] ;
		char FCB_FN[ 11 ] ;
		void *Previous_SFT ;
		short int Machine_Number ;
		short int Owner ;
		short int Sharing_Record ;
		short int Absolute_Cluster ;
		void *IFS_Driver ;
	} ;
	struct
	{
		unsigned char Filler1[ 25 ] ;
		short int Relative_Cluster ;
		short int Directory_Sector ;
		unsigned char Directory_Entry ;
	} ;
} ;

/* CDS (Current Directory Structure), DOS 3.n */
typedef union CDSV3
{
	struct
	{
		char Curr_Path[ 67 ] ;
		short int Flags ;
		void *DPB ;
		void *REDIRIFS ;
		short int Parameter ;
		short int Root_OFS ;
	} ;
	struct
	{
		unsigned char Filler[ 73 ] ;
		short int Start ;
		short int Filler0 ;
	} ;
} ;

/* CDS (Current Directory Structure), DOS 4.n */
typedef union CDSV4
{
	struct
	{
		char Curr_Path[ 67 ] ;
		short int Flags ;
		void *DPB ;
		void *REDIRIS ;
		short int Parameter ;
		short int Root_OFS ;
		unsigned char Filler1 ;
		void *IFS_Driver ;
		short int Filler2 ;
	} ;
	struct
	{
		unsigned char Filler[ 73 ] ;
		short int Start ;
		short int Filler0 ;
	} ;
} 

/* DOS Directory entry */
typedef struct Dir_Rec
{
    char Name[ 11 ] ;
    /* First byte is $E5 if entry is free and $00 if end of entries */
    unsigned char Attributes ;
    unsigned char Filler[ 10 ] ;
    short int Update_Time ;
    short int Update_Date ;
    short int Start_Cluster ;
    long Size ;
} ;

/* FindFirst/Next data block for SDAs */
typedef struct SDB_Rec
{
    unsigned char Drive_Letter ;
    char Search_Template[ 11 ] ;
    unsigned char Search_Attributes ;
    short int Directory_Entry ;
    short int Parent_Cluster ;
    unsigned char Filler[ 5 ] 
} ;

/* SDA (Swappable DOS Area), DOS 3.n */
typedef struct SDAV3
{
    unsigned char Critical_Error_Flag ;
    unsigned char InDOS_Flag ;
    unsigned char Error_Drive ;
    unsigned char Error_Locus ;
    short int Extended_Error_Code ;
    unsigned char Suggested_Action ;
    unsigned char Error_Class ;
    void *Error_ES_DI ;
    void *Current_DTA ;
    short int Current_PSP ;
    short int Stored_SP ;
    short int Return_Code ;
    unsigned char Current_Drive ;
    unsigned char Break_Flag ;
    /* Swap following only if in DOS */
    short int Stored_AX ;
    short int Sharing_PSP ;
    short int Sharing_Number ;
    short int First_MB ;
    short int Best_MB ;
    short int Last_MB ;
    unsigned char Filler[ 10 ] ;
    unsigned char DD ;
    unsigned char MM ;
    short int YY_1980 ;
    short int DD_1980 ;
    unsigned char Day_Of_Week ;
    unsigned char Filler1[ 3 ] ;
    unsigned char Driver_Request_Header[ 26 ] ;
    void *Entry_Point ;
    unsigned char Driver_Request_Header1[ 22 ] ;
    unsigned char Driver_Request_Header2[ 22 ] ;
    unsigned char PSP_Type ;
    unsigned char Filler2[ 7 ] ;
    unsigned char CLOCK_Record[ 6 ] ;
    unsigned char Filler3[ 2 ] ;
    char FN1[ 128 ] ;
    char FN2[ 128 ] ;
    SDB : SDB_Rec;
    Found_File : Dir_Rec;
    Drive_CDSCopy : CDSV3;
    char FCB_FN1[ 11 ] ;
    unsigned char Filler4 ;
    char FCB_FN2[ 11 ] ;
    unsigned char Filler5[ 9 ] ;
    unsigned char Search_Attributes ;
    unsigned char FCB_Type ;
    unsigned char Extended_Attributes ;
    unsigned char Open_Mode ;
    unsigned char Filler6[ 3 ] ;
    unsigned char DOS_Flag ;
    unsigned char Filler7[ 9 ] ;
    unsigned char Termination_Type ;
    unsigned char Filler8 ;
    unsigned char Replace_Byte ;
    void *Error_DPB ;
    void *Stack_Frame ;
    short int Stored_SP1 ;
    void *DPB_Ptr ;
    unsigned char Filler9[ 8 ] ;
    unsigned char Media_ID ;
    unsigned char Filler10[ 5 ] ;
    void *Current_SFT ;
    void *Drive_CDSPTR ;
    void *Caller_FCB ;
    short int Filler11 ;
    short int Temp ;
    void *JFT ;
    short int FN1_CSOFS ;
    short int FN2_CSOFS ;
    unsigned char Filler12[ 18 ] ;
    long Current_Offset ;
    unsigned char Filler13[ 12 ] ;
    long Appended ;
    void *Disk_Buffer ;
    void *SFT ;
    short int Stored_BX ;
    short int Stored_DS ;
    short int Temp1 ;
    void *Call_Frame ;
    Ren_SrcFile : SDB_Rec;
    Ren_File : Dir_Rec;
    unsigned char Error_Stack[ 331 ] ;
    unsigned char Disk_Stack[ 384 ] ;
    unsigned char Char_Stack[ 384 ] 
} ;

/* SDA (Swappable DOS Area), DOS 3.3 */
typedef struct SDAV33
{
    unsigned char Critical_Error_Flag ;
    unsigned char InDOS_Flag ;
    unsigned char Error_Drive ;
    unsigned char Error_Locus ;
    short int Extended_Error_Code ;
    unsigned char Suggested_Action ;
    unsigned char Error_Class ;
    void *Error_ES_DI ;
    void *Current_DTA ;
    short int Current_PSP ;
    short int Stored_SP ;
    short int Return_Code ;
    unsigned char Current_Drive ;
    unsigned char Break_Flag ;
    /* Swap following only if in DOS */
    short int Stored_AX ;
    short int Sharing_PSP ;
    short int Sharing_Number ;
    short int First_MB ;
    short int Best_MB ;
    short int Last_MB ;
    unsigned char Filler[ 10 ] ;
    unsigned char DD ;
    unsigned char MM ;
    short int YY_1980 ;
    short int DD_1980 ;
    unsigned char Day_Of_Week ;
    unsigned char Filler1[ 3 ] ;
    unsigned char Driver_Request_Header[ 26 ] ;
    void *Entry_Point ;
    unsigned char Driver_Request_Header1[ 22 ] ;
    unsigned char Driver_Request_Header2[ 22 ] ;
    unsigned char PSP_Type ;
    unsigned char Filler2[ 7 ] ;
    unsigned char CLOCK_Record[ 6 ] ;
    unsigned char Filler3[ 2 ] ;
    char FN1[ 128 ] ;
    char FN2[ 128 ] ;
    SDB : SDB_Rec;
    Found_File : Dir_Rec;
    Drive_CDSCopy : CDSV3;
    char FCB_FN1[ 11 ] ;
    unsigned char Filler4 ;
    char FCB_FN2[ 11 ] ;
    unsigned char Filler5[ 9 ] ;
    unsigned char Search_Attributes ;
    unsigned char FCB_Type ;
    unsigned char Extended_Attributes ;
    unsigned char Open_Mode ;
    unsigned char Filler6[ 3 ] ;
    unsigned char DOS_Flag ;
    unsigned char Filler7[ 9 ] ;
    unsigned char Termination_Type ;
    unsigned char Filler8 ;
    unsigned char Replace_Byte ;
    void *Error_DPB ;
    void *Stack_Frame ;
    short int Stored_SP1 ;
    void *DPB_Ptr ;
    unsigned char Filler9[ 8 ] ;
    unsigned char Media_ID ;
    unsigned char Filler10[ 5 ] ;
    void *Current_SFT ;
    void *Drive_CDSPTR ;
    void *Caller_FCB ;
    short int Filler11 ;
    short int Temp ;
    void *JFT ;
    short int FN1_CSOFS ;
    short int FN2_CSOFS ;
    unsigned char Filler12[ 18 ] ;
    long Current_Offset ;
    unsigned char Filler13[ 12 ] ;
    long Appended ;
    void *Disk_Buffer ;
    void *SFT ;
    short int Stored_BX ;
    short int Stored_DS ;
    short int Temp1 ;
    void *Call_Frame ;
    Ren_SrcFile : SDB_Rec;
    Ren_File : Dir_Rec;
    unsigned char Error_Stack[ 331 ] ;
    unsigned char Disk_Stack[ 384 ] ;
    unsigned char Char_Stack[ 384 ] ;
    unsigned char Flag ;
    unsigned char Drive ;
    short int Filler14 
} ;

/* SDA (Swappable DOS Area), DOS 4.n */
typedef struct SDAV4
{
    unsigned char Critical_Erro_Flag ;
    unsigned char InDOS_Flag ;
    unsigned char Error_Drive ;
    unsigned char Error_Locus ;
    short int Extended_Error_Code ;
    unsigned char Suggested_Action ;
    unsigned char Error_Class ;
    void *Error_ES_DI ;
    void *Current_DTA ;
    short int Current_PSP ;
    short int Stored_SP ;
    short int Return_Code ;
    unsigned char Current_Drive ;
    unsigned char Break_Flag ;
    unsigned char Filler[ 2 ] ;
    /* Swap following only if in DOS */
    short int Stored_AX ;
    short int Sharing_PSP ;
    short int Sharing_Number ;
    short int First_MB ;
    short int Best_MB ;
    short int Last_MB ;
    unsigned char Filler1[ 10 ] ;
    unsigned char DD ;
    unsigned char MM ;
    short int YY_1980 ;
    short int DD_1980 ;
    unsigned char Day_Of_Week ;
    unsigned char Filler2[ 3 ] ;
    unsigned char Driver_Request_Header[ 30 ] ;
    void *Entry_Point ;
    unsigned char Driver_Request_Header1[ 22 ] ;
    unsigned char Driver_Request_Header2[ 30 ] ;
    unsigned char Filler3[ 6 ] ;
    unsigned char CLOCK_Record[ 6 ] ;
    unsigned char Filler4[ 2 ] ;
    char FN1[ 128 ] ;
    char FN2[ 128 ] ;
    SDB : SDB_Rec;
    Found_File : Dir_Rec;
    Drive_CDSCopy : CDSV4;
    char FCB_FN1[ 11 ] ;
    unsigned char Filler5 ;
    char FCB_FN2[ 11 ] ;
    unsigned char Filler6[ 11 ] ;
    unsigned char Search_Attributes ;
    unsigned char Open_Mode ;
    unsigned char Filler7[ 3 ] ;
    unsigned char DOS_Flag ;
    unsigned char Filler8[ 9 ] ;
    unsigned char Termination_Type ;
    unsigned char Filler9[ 3 ] ;
    void *Error_DPB ;
    void *Stack_Frame ;
    short int Stored_SP1 ;
    void *DPB_Pointer ;
    short int Disk_Buffer ;
    unsigned char Filler10[ 8 ] ;
    unsigned char Media_ID ;
    unsigned char Filler11 ;
    void *Filler12 ;
    void *Current_SFT ;
    void *Drive_CDSPTR ;
    void *Caller_FCB ;
    unsigned char Filler13[ 4 ] ;
    void *JFT ;
    short int FN1_CSOFS ;
    short int FN2_CSOFS ;
    unsigned char Filler14[ 12 ] ;
    short int Directory_Cluster ;
    unsigned char Filler15[ 10 ] ;
    long Offset ;
    short int Filler16 ;
    short int Partial_Bytes ;
    short int Sectors ;
    unsigned char Filler17[ 6 ] ;
    long Appended_Bytes ;
    void *Disk_Buffer1 ;
    void *SFT1 ;
    short int Store_BX ;
    short int Store_DS ;
    short int Temp ;
    void *Previous_Call_Frame ;
    unsigned char Filler18[ 5 ] ;
    void *Filler19 ;
    short int Special_Open_Action ;
    short int SPop_Attr ;
    short int SPop_Mode ;
    unsigned char Filler20[ 9 ] ;
    short int Save_DS ;
    unsigned char Filler21[ 5 ] ;
    void *Filename ;
    void *Filler22 ;
    short int Ssve_SS ;
    short int Save_SP ;
    unsigned char Stack_Switch ;
    Ren_SrcFile : SDB_Rec;
    Ren_File : Dir_Rec;
    unsigned char Error_Stack[ 331 ] ;
    unsigned char Disk_Stack[ 384 ] ;
    unsigned char Char_Stack[ 384 ] ;
    unsigned char AH_Flag ;
    unsigned char Drive ;
    unsigned char Flag ;
    unsigned char Filler23[ 9 ] ;
} ;

typedef struct OS /* Structure for manipulation of pointers */
{
    short int Offset ;
    short int Segment ;
} ;

typedef union XRegisters /* Extended Registers type */
{
	struct
	{
		short int BP ;
		short int ES ;
		short int DS ;
		short int DI ;
		short int SI ;
		short int DX ;
		short int CX ;
		short int BX ;
		short int AX ;
		short int SS ;
		short int SP ;
		short int CS ;
		short int IP ;
		short int FLAGS ;
	} ;
	struct
	{
		unsigned char Fill1a ;
		unsigned char Fill1b ;
		unsigned char Fill2a ;
		unsigned char Fill2b ;
		unsigned char Fill3a ;
		unsigned char Fill3b ;
		unsigned char Fill4a ;
		unsigned char Fill4b ;
		unsigned char Fill5a ;
		unsigned char Fill5b ;
		unsigned char DL ;
		unsigned char DH ;
		unsigned char CL ;
		unsigned char CH ;
		unsigned char BL ;
		unsigned char BH ;
		unsigned char AL ;
		unsigned char AH ;
	} ;
} ;

/* CTR (Clock Transfer Record) */
typedef struct CTR
{
    short int DD_1980 ;
    unsigned char Minutes ;
    unsigned char Hours ;
    unsigned char Hundreths ;
    unsigned char Seconds ;
} ;

/* DDH (Device Driver Header) */
typedef union DDH
{
	struct
	{
		void *Next ;
		short int Attributes ;
		short int Exit_Point ;
		short int Entry_Point ;
		unsigned char Sub_Units ;
		unsigned char Filler[ 7 ] ;
		short int Reserved ;
		char Drive ;
		unsigned char Units ;
		unsigned char Signature[ 6 ] ;
	} ;
	struct
	{
		unsigned char Filler1[ 10 ] ;
		char Name[ 8 ] ;
	} ;
} ;

/* DDB (DOS Disk Buffer), DOS V2.n */
typedef struct DDBV2
{
    void *Next ;
    unsigned char Drive ;
    unsigned char Filler[ 3 ] ;
    short int Sector ;
    unsigned char Copies ;
    unsigned char Offset ;
    void *DPB ;
    unsigned char Start ;
} ;

/* DDB (DOS Disk Buffer), DOS V3.n */
typedef struct DDBV3
{
    void *Next ;
    unsigned char Drive ;
    unsigned char Flags ;
    short int Sector ;
    unsigned char Copies ;
    unsigned char Offset ;
    void *DPB ;
    short int Filler ;
    unsigned char Start ;
} ;

/* DDB (DOS Disk Buffer), DOS V4.n */
typedef struct DDBV4
{
    short int Next ;
    short int Previous ;
    unsigned char Drive ;
    unsigned char Flags ;
    long Sector ;
    unsigned char Copies ;
    short int Offset ;
    void *DPB ;
    short int Count ;
    unsigned char Filler ;
    unsigned char Start ;
} ;

/* DBHCH (Disk Buffer Hash Chain Head), DOS V4.n */
typedef struct DBHCH
{
    short int Page ;
    void *Least_Used ;
    unsigned char Dirty_Buffers ;
    unsigned char Filler ;
} ;

/* DBI (Disk Buffer Info), DOS V4.0 */
typedef struct DBIV40
{
    void *DBHCH ;
    short int Chain_Count ;
    void *Lookahead ;
    unsigned char Flag ;
    short int Handle ;
    short int Page ;
    short int Filler ;
    short int Segment ;
    short int Filler1 ;
    short int Mapping[ 4 ] ;
} ;

/* DBI (Disk Buffer Info), DOS V4.01 */
typedef struct DBIV41
{
    void *DBHCH ;
    short int Chain_Count ;
    void *Lookahead ;
    short int Lookahead_Count ;
    unsigned char Flag ;
    short int Segment ;
    short int Page ;
    short int Segment1 ;
    short int Page1 ;
    unsigned char Frames ;
    short int Workspace ;
    short int Handle ;
    short int Page2 ;
    short int Filler ;
    short int Frame ;
    short int Filler1 ;
    unsigned char Flag1 ;
} ;
