{
        Program Name : VCLDebug
        Package Name : LB
        Purpose      : Debug interfaces for VCL classes
        Institution  : Conroy & Conroy Co.
        Date Written : 25-May-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit VCLDebug ;

interface

uses // Borland...
     Windows, // TRect
     Classes, // TComponent
     Controls, // TControl

     // Other...
     _DebugIn, { Text_Debugger }
     DebugInt ; { TText_Debugger }

type TInherited_Debug_Iterator = class( TText_Debugger )
                                     protected // Instance data...
                                         Initialized : boolean ;
                                         Children : TStringList ;

                                     public // Constructors and destructors...
                                         constructor Create ;

                                         destructor Destroy ; override ;

                                     public // API...
                                         function Child( Index : longint ) : PDebug_Interface ;
                                             override ;

                                         function Count : integer ; override ;

                                         procedure Setup ; virtual ;
                                 end ;

type TComponent_Debug_Iterator = class( TInherited_Debug_Iterator )
                                     public // API...
                                         Component : TComponent ;
                                         
                                         procedure Setup ; override ;

                                         function Count : integer ; override ;
                                 end ;

type TComponent_List_Debug_Iterator = class( TInherited_Debug_Iterator )
                                           public // API...
                                               Component : TComponent ;

                                               procedure Setup ; override ;

                                               function Count : integer ; override ;
                                       end ;

type TControl_Debug_Iterator = class( TComponent_Debug_Iterator )
                                   public // API...
                                       Control : TControl ;

                                       procedure Setup ; override ;

                                       function Count : integer ; override ;
                               end ;

function Component_Description( C : TComponent ) : string ;
function Component_Debugger( C : TComponent ) : PDebug_Interface ;

function Align_To_String( X : TAlign ) : string ;
function Anchors_To_String( X : TAnchors ) : string ;
function BiDiMode_To_String( X : TBiDiMode ) : string ;
function ComponentState_To_String( X : TComponentState ) : string ;
function ComponentStyle_To_String( X : TComponentStyle ) : string ;
function ControlState_To_String( X : TControlState ) : string ;
function ControlStyle_To_String( X : TControlStyle ) : string ;
function Cursor_To_String( X : TCursor ) : string ;
function DockOrientation_To_String( X : TDockOrientation ) : string ;
function Point_To_String( X : TPoint ) : string ;
function Rect_To_String( X : TRect ) : string ;

implementation

uses // Borland...
     SysUtils, // inttostr

     // C&C...
     CommonUt, // Pointer_To_String
     UStrings ; // Pointer_To_String

function Align_To_String( X : TAlign ) : string ;

begin
    case X of
        alNone: Result := 'alNone' ;
        alTop: Result := 'alTop' ;
        alBottom: Result := 'alBottom' ;
        alLeft: Result := 'alLeft' ;
        alRight: Result := 'alRight' ;
        alClient: Result := 'alClient' ;
        else Result := '* unknown *' ;
    end ;
end ;


function Anchors_To_String( X : TAnchors ) : string ;

    function Add( const S : string ) : string ;

    begin
        if( length( Result ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + S ;
    end ;

begin
    Result := '' ;
    if( akTop in X ) then
    begin
       Add( 'akTop' ) ;
    end ;
    if( akLeft in X ) then
    begin
       Add( 'akLeft' ) ;
    end ;
    if( akRight in X ) then
    begin
       Add( 'akRight' ) ;
    end ;
    if( akBottom in X ) then
    begin
       Add( 'akBottom' ) ;
    end ;
end ;


function BiDiMode_To_String( X : TBiDiMode ) : string ;

begin
    case X of
        bdLeftToRight: Result := 'bdLeftToRight' ;
        bdRightToLeft: Result := 'bdRightToLeft' ;
        bdRightToLeftNoAlign: Result := 'bdRightToLeftNoAlign' ;
        bdRightToLeftReadingOnly: Result := 'bdRightToLeftReadingOnly' ;
        else Result := '* unknown *' ;
    end ;
end ;


function ComponentState_To_String( X : TComponentState ) : string ;

    function Add( const S : string ) : string ;

    begin
        if( length( Result ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + S ;
    end ;

begin
    Result := '' ;
    if( csLoading in X ) then
    begin
       Add( 'csLoading' ) ;
    end ;
    if( csReading in X ) then
    begin
       Add( 'csReading' ) ;
    end ;
    if( csWriting in X ) then
    begin
       Add( 'csWriting' ) ;
    end ;
    if( csDestroying in X ) then
    begin
       Add( 'csDestroying' ) ;
    end ;
    if( csDesigning in X ) then
    begin
       Add( 'csDesigning' ) ;
    end ;
    if( csAncestor in X ) then
    begin
       Add( 'csAncestor' ) ;
    end ;
    if( csUpdating in X ) then
    begin
       Add( 'csUpdating' ) ;
    end ;
    if( csFixups in X ) then
    begin
       Add( 'csFixups' ) ;
    end ;
    if( csFreeNotification in X ) then
    begin
       Add( 'csFreeNotification' ) ;
    end ;
    if( csInline in X ) then
    begin
       Add( 'csInline' ) ;
    end ;
    if( csDesignInstance in X ) then
    begin
       Add( 'csDesignInstance' ) ;
    end ;
    Result := '[' + Result + ']' ;
end ;


function ComponentStyle_To_String( X : TComponentStyle ) : string ;

    function Add( const S : string ) : string ;

    begin
        if( length( Result ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + S ;
    end ;

begin
    Result := '' ;
    if( csInheritable in X ) then
    begin
        Add( 'csInheritable' ) ;
    end ;
    if( csCheckPropAvail in X ) then
    begin
        Add( 'csCheckPropAvail' ) ;
    end ;
    Result := '[' + Result + ']' ;
end ;


function ControlState_To_String( X : TControlState ) : string ;

    function Add( const S : string ) : string ;

    begin
        if( length( Result ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + S ;
    end ;

begin
    if( csLButtonDown in X ) then
    begin
        Add( 'csLButtonDown' ) ;
    end ;
    if( csClicked in X ) then
    begin
        Add( 'csClicked' ) ;
    end ;
    if( csPalette in X ) then
    begin
        Add( 'csPalette' ) ;
    end ;
    if( csReadingState in X ) then
    begin
        Add( 'csReadingState' ) ;
    end ;
    if( csAlignmentNeeded in X ) then
    begin
        Add( 'csAlignmentNeeded' ) ;
    end ;
    if( csFocusing in X ) then
    begin
        Add( 'csFocusing' ) ;
    end ;
    if( csCreating in X ) then
    begin
        Add( 'csCreating' ) ;
    end ;
    if( csPaintCopy in X ) then
    begin
        Add( 'csPaintCopy' ) ;
    end ;
    if( csCustomPaint in X ) then
    begin
        Add( 'csCustomPaint' ) ;
    end ;
    if( csDestroyingHandle in X ) then
    begin
        Add( 'csDestroyingHandle' ) ;
    end ;
    if( csDocking in X ) then
    begin
        Add( 'csDocking' ) ;
    end ;
end ;


function ControlStyle_To_String( X : TControlStyle ) : string ;

    function Add( const S : string ) : string ;

    begin
        if( length( Result ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + S ;
    end ;

begin
    if( csAcceptsControls in X ) then
    begin
        Add( 'csAcceptsControls' ) ;
    end ;
    if( csCaptureMouse in X ) then
    begin
        Add( 'csCaptureMouse' ) ;
    end ;
    if( csDesignInteractive in X ) then
    begin
        Add( 'csDesignInteractive' ) ;
    end ;
    if( csClickEvents in X ) then
    begin
        Add( 'csClickEvents' ) ;
    end ;
    if( csFramed in X ) then
    begin
        Add( 'csFramed' ) ;
    end ;
    if( csSetCaption in X ) then
    begin
        Add( 'csSetCaption' ) ;
    end ;
    if( csOpaque in X ) then
    begin
        Add( 'csOpaque' ) ;
    end ;
    if( csDoubleClicks in X ) then
    begin
        Add( 'csDoubleClicks' ) ;
    end ;
    if( csFixedWidth in X ) then
    begin
        Add( 'csFixedWidth' ) ;
    end ;
    if( csFixedHeight in X ) then
    begin
        Add( 'csFixedHeight' ) ;
    end ;
    if( csNoDesignVisible in X ) then
    begin
        Add( 'csNoDesignVisible' ) ;
    end ;
    if( csReplicatable in X ) then
    begin
        Add( 'csReplicatable' ) ;
    end ;
    if( csNoStdEvents in X ) then
    begin
        Add( 'csNoStdEvents' ) ;
    end ;
    if( csDisplayDragImage in X ) then
    begin
        Add( 'csDisplayDragImage' ) ;
    end ;
    if( csReflector in X ) then
    begin
        Add( 'csReflector' ) ;
    end ;
    if( csActionClient in X ) then
    begin
        Add( 'csActionClient' ) ;
    end ;
    if( csMenuEvents in X ) then
    begin
        Add( 'csMenuEvents' ) ;
    end ;
end ;


function Cursor_To_String( X : TCursor ) : string ;

begin
    case X of
        crDefault: Result := 'crDefault' ;
        crNone: Result := 'crNone' ;
        crArrow: Result := 'crArrow' ;
        crCross: Result := 'crCross' ;
        crIBeam: Result := 'crIBeam' ;
        crSizeNESW: Result := 'crSizeNESW' ;
        crSizeNS: Result := 'crSizeNS' ;
        crSizeNWSE: Result := 'crSizeNWSE' ;
        crSizeWE: Result := 'crSizeWE' ;
        crUpArrow: Result := 'crUpArrow' ;
        crHourGlass: Result := 'crHourGlass' ;
        crDrag: Result := 'crDrag' ;
        crNoDrop: Result := 'crNoDrop' ;
        crHSplit: Result := 'crHSplit' ;
        crVSplit: Result := 'crVSplit' ;
        crMultiDrag: Result := 'crMultiDrag' ;
        crSQLWait: Result := 'crSQLWait' ;
        crNo: Result := 'crNo' ;
        crAppStart: Result := 'crAppStart' ;
        crHelp: Result := 'crHelp' ;
        crHandPoint: Result := 'crHandPoint' ;
        crSizeAll: Result := 'crSizeAll' ;
    end ;
    if( length( Result ) > 0 ) then
    begin
        Result := ' (' + Result + ')' ;
    end ;
    Result := inttostr( X ) + Result ;
end ;


function DockOrientation_To_String( X : TDockOrientation ) : string ;

begin
    case X of
        doNoOrient: Result := 'doNoOrient' ;
        doHorizontal: Result := 'doHorizontal' ;
        doVertical: Result := 'doVertical' ;
        else Result := '* unknown *' ;
    end ;
end ;


function Point_To_String( X : TPoint ) : string ;

begin
    Result := '(X:' + inttostr( X.X ) + ', Y:' + inttostr( X.Y ) + ')' ;
end ;


function Rect_To_String( X : TRect ) : string ;

begin
    Result := '(Left:' + inttostr( X.Left ) + ', Top:' + inttostr( X.Top ) +
        ', Right:' + inttostr( X.Right ) + ', Bottom:' + inttostr( X.Bottom ) +
        ')' ;
end ;


function Component_Description( C : TComponent ) : string ;

begin
    if( C = nil ) then
    begin
        Result := 'nil' ;
        exit ;
    end ;
    Result := '' ;
    try
        if( Length( C.Name ) > 0 ) then
        begin
            Result := C.Name ;
        end ;
        Result := '"' + Result + '" [' + C.classname + ']'
    except
    end ;
    Result := Result + ' ' + Pointer_To_String( pointer( C ) ) ;
end ;


function Component_Debugger( C : TComponent ) : PDebug_Interface ;

begin
    if( C = nil ) then
    begin
        Result := PDebug_Interface( TText_Debugger.Create ) ;
        exit ;
    end ;

    if( C is TControl ) then
    begin
        Result := PDebug_Interface( TControl_Debug_Iterator.Create ) ;
        TControl_Debug_Iterator( Result ).Control := TControl( C ) ;
    end else
    begin
        Result := PDebug_Interface( TComponent_Debug_Iterator.Create ) ;
        TComponent_Debug_Iterator( Result ).Component := C ;
    end ;
end ;


// TInherited_Debug_Iterator methods...

// Constructors and destructors...

constructor TInherited_Debug_Iterator.Create ;

begin
    inherited Create ;

    Children := TStringList.Create ;
end ;


destructor TInherited_Debug_Iterator.Destroy ;

var Loop : integer ;

begin
    for Loop := 0 to Children.Count - 1 do
    begin
        Children.Objects[ Loop ] := nil ;
    end ;
    Children.Free ;
    Children := nil ;

    inherited Destroy ;
end ;


// API...

function TInherited_Debug_Iterator.Child( Index : longint ) : PDebug_Interface ;

begin
    if( not Initialized ) then
    begin
        Setup ;
        Children.Sort ;
        Initialized := True ;
    end ;
    Result := PDebug_Interface( Children.Objects[ Index ] ) ;
end ;


function TInherited_Debug_Iterator.Count : integer ;

begin
    Count := 0 ;
end ;


procedure TInherited_Debug_Iterator.Setup ;

begin
    // This routine left intentionally blank
end ;


// TComponent_Debug_Iterator methods...

// API...

function TComponent_Debug_Iterator.Count : integer ;

begin
    Result := 10 + inherited Count ;
end ;


procedure TComponent_Debug_Iterator.Setup ;

var Index : integer ;
    Iterator : TText_Debugger ;
    S : string ;

begin
    inherited Setup ;

    for Index := 0 to 9 do
    begin
        Iterator := TText_Debugger.Create ;
        case Index of
            0 : try
                    S := 'ComObject = ' + Pointer_To_String( pointer( Component.ComObject ) ) ;
                except
                    S := 'ComObject = ()' ;
                end ;
            1 : S := 'ComponentIndex = ' + inttostr( Component.ComponentIndex ) ;
            2 : begin
                    Iterator.Free ;
                    Iterator := TComponent_List_Debug_Iterator.Create ;
                    TComponent_List_Debug_Iterator( Iterator ).Component := Component ;
                    S := 'Components' ;
                end ;
            3 : S := 'ComponentState = ' + ComponentState_To_String( Component.ComponentState ) ;
            4 : S := 'ComponentStyle = ' + ComponentStyle_To_String( Component.ComponentStyle ) ;
            5 : S := 'DesignInfo = ' + inttostr( Component.DesignInfo ) ;
            6 : S := 'Name = "' + Component.Name + '"' ;
            7 : begin
                    if( Component.Owner = nil ) then
                    begin
                        S := 'Owner = nil ' ;
                    end else
                    begin
                        Iterator.Free ;
                        Iterator := TText_Debugger( Component_Debugger( Component.Owner ) ) ;
                        S := 'Owner = ' + Component_Description( Component.Owner ) ;
                    end ;
                end ;
            8 : S := 'Tag = ' + inttostr( Component.Tag ) ;
            9 : S := 'VCLComObject = ' + Pointer_To_String( Component.VCLComObject ) ;
        end ; // case Index of
        Children.Add( S ) ;
        Children.Objects[ Children.Count - 1 ] := Iterator ;
        Iterator.Title := PChar( S ) ;
    end ; // for Index := 0 to 9 do
end ;



// TComponent_List_Debug_Iterator methods...

// API...

function TComponent_List_Debug_Iterator.Count : integer ;

begin
    Result := Component.ComponentCount + inherited Count ;
end ;


procedure TComponent_List_Debug_Iterator.Setup ;

var Index : integer ;
    Iterator : TText_Debugger ;

begin
    inherited Setup ;

    for Index := 0 to Component.ComponentCount - 1 do
    begin
        Iterator := TText_Debugger( Component_Debugger( Component.Components[ Index ] ) ) ;
        Iterator.Title := PChar( Component_Description( Component.Components[ Index ] ) ) ;
        Children.Add( Iterator.Title ) ;
        Children.Objects[ Children.Count - 1 ] := Iterator ;
    end ;
end ;



// TControl_Debug_Iterator methods...

// API...

function TControl_Debug_Iterator.Count : integer ;

begin
    Count := 28 + inherited Count ;
end ;


procedure TControl_Debug_Iterator.Setup ;

var Index : integer ;
    Iterator : TText_Debugger ;
    S : string ;

begin
    Component := Control ;
    inherited Setup ;

    for Index := 0 to 27 do
    begin
        Iterator := TText_Debugger.Create ;
        try
            case Index of
                0 : begin
                        Iterator.Free ;
                        Iterator := TText_Debugger( Component_Debugger( Control.Action ) ) ;
                        S := 'Action = ' + Component_Description( Control.Action ) ;
                    end ;
                1 : S := 'Align = ' + Align_To_String( Control.Align ) ;
                2 : S := 'Anchors = ' + Anchors_To_String( Control.Anchors ) ;
                3 : S := 'BiDiMode = ' + BiDiMode_To_String( Control.BiDiMode ) ;
                4 : S := 'BoundsRect = ' + Rect_To_String( Control.BoundsRect ) ;
                5 : S := 'ClientHeight = ' + inttostr( Control.ClientHeight ) ;
                6 : S := 'ClientOrigin = ' + Point_To_String( Control.ClientOrigin ) ;
                7 : S := 'ClientRect = ' + Rect_To_String( Control.ClientRect ) ;
                8 : S := 'ClientWidth = ' + inttostr( Control.ClientWidth ) ;
                //Constraints
                9 : S := 'ControlState = ' + ControlState_To_String( Control.ControlState ) ;
                10 : S := 'ControlStyle = ' + ControlStyle_To_String( Control.ControlStyle ) ;
                11 : S := 'Cursor = ' + Cursor_To_String( Control.Cursor ) ;
                12 : S := 'DockOrientation = ' + DockOrientation_To_String( Control.DockOrientation ) ;
                13 : S := 'Enabled = ' + Boolean_To_String( Control.Enabled ) ;
                14 : S := 'Floating = ' + Boolean_To_String( Control.Floating ) ;
                15 : S := 'Height = ' + inttostr( Control.Height ) ;
                16 : begin
                         Iterator.Free ;
                         Iterator := TText_Debugger( Component_Debugger( Control.HostDockSite ) ) ;
                         S := 'HostDockSite = ' + Component_Description( Control.HostDockSite ) ;
                     end ;
                17 : S := 'Hint = "' + Control.Hint + '"' ;
                18 : S := 'Left = ' + inttostr( Control.left ) ;
                19 : S := 'LRDockWidth = ' + inttostr( Control.LRDockWidth ) ;
                20 : begin
                         Iterator.Free ;
                         Iterator := TText_Debugger( Component_Debugger( Control.Parent ) ) ;
                         S := 'Parent = ' + Component_Description( Control.Parent ) ;
                     end ;
                21 : S := 'ShowHint = ' + Boolean_To_String( Control.ShowHint ) ;
                22 : S := 'TBDockHeight = ' + inttostr( Control.TBDockHeight ) ;
                23 : S := 'Top = ' + inttostr( Control.Top ) ;
                24 : S := 'UndockHeight = ' + inttostr( Control.UndockHeight ) ;
                25 : S := 'UndockWidth = ' + inttostr( Control.UndockWidth ) ;
                26 : S := 'Visible = ' + Boolean_To_String( Control.Visible ) ;
                27 : S := 'Width = ' + inttostr( Control.Width ) ;
            end ;
        except
	    on E : Exception do S := inttostr( Index ) + ' ' + E.Message ; 
        end ;
        Iterator.Title := PChar( S ) ;
        Children.Add( S ) ;
        Children.Objects[ Children.Count - 1 ] := Iterator ;
    end ;
end ;


end.
