{
              Program Name : Wildcard
              Package Name : LB
              Purpose      : Filename wildcard functions
              Institution  : Conroy & Conroy
              Date Written : 31-Mar-89
              Written By   : Alan Conroy
              Version      : 3.0A

              Copyright (C) 1989 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *  1-Nov-2002   EAC            Extract Lookup function	*
	*                              to files unit.           *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This function provides for Extended DOS Wildcard lookups.
}

unit Wildcard ;

interface

{$I EDEFINES.INC}

uses
{$IFDEF DELPHI}
     SysUtils,
{$ELSE}
     WinDOS, Strings,
  {$IFNDEF WINDOWS}
     Dos,
  {$ENDIF}
{$ENDIF}
     CommonUt, { FEdit }
     _Types ; { Small_String }

procedure Split( Path : String ; var Dir, Name, Ext : String ) ;

implementation

uses Compatib,
     TypeDefs ; { Small_String }

type PLookup = ^TLookup ;
     TLookup = object
		   private { Instance data... }
		     All_Subdirectories : boolean ; { True if at/past \...\ }
		     Child : PLookup ; { Child lookup object }
		     Current : String ; { Current lookup }
		     Doing_Files : boolean ; { True if files instead of dirs }
		     File_Stuff : TSearchrec ; { Information on file being returned }
		     Fixed : String ; { Non-wildcard part of specification }
		     Remainder : String ; { Portion of lookup spec, below us }
		     Specification : Small_String ; { Wildcard lookup specification }

		   public { Constructors and destructors... }
		     constructor Init( Spec : String ; All : boolean ) ;

		     destructor Done ;

		   public { Public API... }
		     function Next( Path : String ) : String ;
		     { Return next match, or null }
	       end ;

procedure Split( Path : String ; var Dir, Name, Ext : String ) ;
{ Split a path into directory, filename, and extension parts }

var Dummy : Integer ;

begin
    if( Path = '' ) then
    begin
        Dir := '' ;
        Name := '' ;
        Ext := '' ;
        exit ;
    end ;

    { Find extension... }
    Dummy := length( Path ) ;
    while ( Path[ Dummy ] <> '.' ) and ( Path[ Dummy ] <> '\' ) and
        ( Path[ Dummy ] <> ':' ) and ( Dummy > 1 ) do Dec( Dummy ) ;
    if Path[ Dummy ] <> '.' then {No extension found}
    begin
        Ext := '' ;
    end else
    begin
        Ext := Copy( Path, Dummy, 254 ) ;
        Path := Copy( Path, 1, Dummy - 1 ) ;
    end ;

    { Find name... }
    Dummy := length( Path ) ;
    while(  ( Dummy > 0 ) and ( Path[ Dummy ] <> '\' ) and ( Path[ Dummy ] <> ':' ) ) do
    begin
        Dec( Dummy ) ;
    end ;                                                        
    Name := Copy( Path, Dummy + 1, 254 ) ;
    Dir := Copy( Path, 1, Dummy ) ;
end ; { Split }


var Global_Lookup : PLookup ;
    File_List : String ;


{ TLookup methods... }

{ Constructors and destructors... }

constructor TLookup.Init( Spec : String ; All : boolean ) ;

var Dummy, Dummy1 : integer ;

begin
    { Make sure specification is in a standard format... }
    if( copy( Spec, length( Spec ) - 1, 2 ) = '\*' ) then
    begin
	Spec := Spec + '.*' ;
    end ;
    if( ( Spec[ length( Spec ) ] = '\' ) or ( Spec[ length( Spec ) ] = ':' ) ) then
    begin
	Spec := Spec + '*.*' ;
    end ;

    { Setup... }
{$IFDEF DELPHI}
    Initialize( Current ) ;
    Initialize( Fixed ) ;
    Initialize( Remainder ) ;
{$ELSE}
    Remainder := '' ;
    Fixed := '' ;
    Current := '' ;
{$ENDIF}
    Specification := Spec ;
    All_Subdirectories := All ;
    Child := nil ;
    Doing_Files := True ;

    { Find first wildcard... }
    Dummy := 1 ;
    while( ( Dummy <= length( Spec ) ) and ( pos( Spec[ Dummy ], '*%' ) = 0 ) ) do
    begin
	inc( Dummy ) ;
    end ;
    Dummy1 := pos( '...\', Spec ) ;
    if( ( Dummy1 < Dummy ) and ( Dummy1 > 0 ) ) then
    begin
	Dummy := Dummy1 ;
    end ;
    Dummy1 := Dummy ; { Save first wildcard position }

    if( Dummy <= length( Spec ) ) then { If a wildcard was found }
    begin
	{ Find end of portion with wildcard }
	while( ( Dummy <= length( Spec ) ) and ( pos( Spec[ Dummy ], ':\' ) = 0 ) ) do
	begin
	    inc( Dummy ) ;
	end ;
	inc( Dummy ) ; { Skip over colon or backslash }

	if( Dummy <= length( Spec ) ) then
	begin
	    Remainder := copy( Spec, Dummy, 254 ) ;
	    Specification := copy( Spec, 1, Dummy - 1 ) ;
	end ;
    end ;

    { Update Specification... }
    Dummy := pos( '...\', Specification ) ;
    if( Dummy > 0 ) then
    begin
	All_Subdirectories := True ;
	Specification := copy( Specification, 1, Dummy - 1 ) + '*.*' ;
    end ;

    { Updated Fixed portion... }
    Dummy := Dummy1 ; { Start at first wildcard }
    while( ( Dummy > 0 ) and ( pos( Specification[ Dummy ], ':\' ) = 0 ) ) do
    begin
	dec( Dummy ) ;
    end ;
    Fixed := copy( Specification, 1, Dummy ) ;
    Specification := copy( Specification, Dummy + 1, 254 ) ;
    Specification[ length( Specification ) + 1 ] := #0 ;
end ; { TLookup.Init }


destructor TLookup.Done ;

begin
    if( Child <> nil ) then
    begin
	dispose( Child, Done ) ;
    end ;
{$IFDEF DELPHI}
    Finalize( Current ) ;
    Finalize( Fixed ) ;
    Finalize( Remainder ) ;
{$ENDIF}
end ;


{ Public API... }

function TLookup.Next( Path : String ) : String ;
{ Return next match, or null.  First check at our level, if requested, then
  search sub-directories. }

label Child_Loop, Loop ;

var S : String ;
    Success : integer ;

begin
Child_Loop:
    if( Child <> nil ) then { At a lower level, pass request along }
    begin
	S := Path + Fixed + Current + '\' ;
	while( copy( S, length( S ) - 1, 2 ) = '\\' ) do
	begin
	    S := copy( S, 1, length( S ) - 1 ) ;
	end ;
	S := Child^.Next( S ) ;
	if( length( S ) = 0 ) then { Not found in child }
	begin
	    dispose( Child, Done ) ;
	    Child := nil ;
	end else
	begin
	    Next := S ;
	    exit ;
	end ;
    end ; { if( Child <> nil ) }

    { Find first/next match at our level... }

    { Check devices }
    if( Specification[ length( Specification ) ] = ':' ) then { Wildcard device }
    begin
	if( Current = 'Z:' ) then { No more devices }
	begin
	    Next := '' ;
	    exit ;
	end ;
	if( length( Current ) = 0 ) then
	begin
	    Current := 'A:' ;
	end else
	begin
	    Current := char( ord( Current[ 1 ] ) + 1 ) + ':' ;
	end ;
	if( length( Remainder ) = 0 ) then
	begin
	    S := '*.*' ;
	end else
	begin
	    S := Remainder ;
	end ;
	Child := new( PLookup, Init( S, All_Subdirectories ) ) ;
	S := Child^.Next( Path + Current ) ;
	if( length( S ) = 0 ) then { Nothing for this device }
	begin
	    Next := Next( Path ) ;
	end else
	begin
	    Next := Path + Fixed + S ;
	end ;
	exit ;
    end ;

    { Check files... }
    if(
	( length( Remainder ) = 0 )
	or
	(
	  ( pos( '\', Remainder ) = 0 ) { No more sub-directories }
	  and
	  ( All_Subdirectories ) { Doing \...\ }
	)
      ) then
    begin
	{ Leave Doing_Files flag alone }
    end else
    begin
	Doing_Files := False ;
    end ;
    if( Doing_Files ) then
    begin
	{ Find file spec in current directory... }
	if( Current = '' ) then { Need first lookup }
	begin
	    if( length( Remainder ) = 0 ) then
	    begin
		S := Path + Fixed + Specification ;
	    end else
	    begin
		S := Path + Fixed + Remainder ;
	    end ;
	    if( S[ length( S ) ] = '\' ) then
	    begin
		S := copy( S, 1, length( S ) - 1 ) ;
	    end ;
{$IFDEF DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
	    Success := FindFirst( S,
		faDirectory or faArchive or faReadOnly or faHidden or faSysFile,
		File_Stuff ) ;
{$IFDEF DELPHI}
  {$WARNINGS ON}
{$ENDIF}
	end else
	begin
	    Success := FindNext( File_Stuff ) ;
	end ;
	if( Success = 0 ) then { Found something }
	begin
{$IFDEF DELPHI}
	    Current := string( File_Stuff.Name ) ;
{$ELSE}
	    Current := strpas( File_Stuff.Name ) ;
{$ENDIF}
	    FEdit( Current, Current, 128 ) ;
	    Next := Path + Fixed + Current ;
	    exit ;
	end else
	begin
	    Current := '' ;
	end ;
	Doing_Files := False ;
    end ; { if( Doing_Files ) }

    { Check sub-directories... }
    if( ( length( Remainder ) = 0 ) and ( not All_Subdirectories ) ) then
    begin { We are done }
	Next := '' ;
	exit ;
    end ;
Loop:
    if( Specification = '\' ) then { Root directory (when followed by ...)}
    begin
	if( Current = '' ) then
	begin
	    Current := '\' ;
	    Child := new( PLookup, Init( Remainder, All_Subdirectories ) ) ;
	    goto Child_Loop ;	
	end else
	begin
	    Current := '' ;
	    Next := '' ;
	    exit ;
	end ;
    end ;
    if( Current = '' ) then
    begin
	S := Path + Fixed + Specification ;
	if( S[ length( S ) ] = '\' ) then
	begin
	    S := copy( S, 1, length( S ) - 1 ) ;
	end ;
	Success := FindFirst( S, faDirectory, File_Stuff ) ;
    end else
    begin
	Success := FindNext( File_Stuff ) ;
    end ;
    if( Success = 0 ) then { Found something }
    begin
{$IFDEF DELPHI}
	Current := string( File_Stuff.Name ) ;
{$ELSE}
	Current := strpas( File_Stuff.Name ) ;
{$ENDIF}
	if( copy( Current, 1, 1 ) = '.' ) then { Ignore . and .. }
	begin
	    goto Loop ;
	end ;
	if( ( File_Stuff.Attr and faDirectory ) = 0 ) then
	begin { Only want directories }
	    goto Loop ;
	end ;
	FEdit( Current, Current, 128 ) ;
	if( length( Remainder ) = 0 ) then
	begin
	    S := '*.*' ;
	end else
	begin
	    S := Remainder ;
	end ;
	if( All_Subdirectories ) then
	begin
	    S := '*.*\' + S ;
	end ;
	Child := new( PLookup, Init( S, All_Subdirectories ) ) ;
	goto Child_Loop ;
    end else
    begin
	Current := '' ;
	Next := '' ;
    end ;
end ; { TLookup.Next }


function Lookup( Filename : String ) : String ;
{ Do a wildcard lookup.  This function is passed a specification which
  consists of one or more file specifications, separated by commas or
  semicolons, for the first lookup attempt.  Succeeding lookups are
  specified with a null parameter.  This routine processes each file
  specification in the list until the lookup fails, then proceeds to the
  next one.  When the last lookup fails, the function returns null. }

var A, B : Integer ;
    Temp : String ;

begin
    if Length( Filename ) > 0 then {Starting new lookup}
    begin
        File_List := Filename ;
	Temp := '' ; { Assume failure of previous lookup }
	if( Global_Lookup <> nil ) then
	begin
	    dispose( Global_Lookup ) ;
	end ;
	Global_Lookup := new( PLookup, Init( Filename, False ) ) ;
    end else
    begin
	if( Global_Lookup = nil ) then
	begin
	    Lookup := '' ;
	    exit ;
	end ;
	Temp := Global_Lookup^.Next( '' ) ; { Lookup next matching file }
    end ;
    while ( Length( Temp ) = 0 ) And ( Length( File_List ) > 0 ) Do
    { While lookups fail }
    begin
	A := Pos( ',', File_List + ',' ) ; { Find end of next file specification }
	B := Pos( ';', File_List + ';' ) ;
        if( ( B < A ) and ( B > 0 ) ) then
        begin
            A := B ;
        end ;
        Temp := Copy( File_List, 1, A - 1 ) ; { Get file specification }
	File_List := Copy( File_List, A + 1, 255 ) ; { Strip from list }
	dispose( Global_Lookup, Done ) ;
	Global_Lookup := new( PLookup, Init( Temp, False ) ) ;
	Temp := Global_Lookup^.Next( '' ) ; { Do the lookup }
    end ;
    Lookup := Temp ; { Return result }
end ; { Lookup }



begin
    Global_Lookup := nil ;
    File_List := '' ;
end.
