{
        Program Name : _DevMan
        Package Name : LB
        Purpose      : Device Manager class definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 7-Jul-2006
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2006 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TDevice_Manager is a class for managing access to devices for
        Operating system objects.
}

unit _DevMan ;

interface

{$I EDEFINES.INC}

uses { Subroutine Library... }
     _Common, { TObject }
     COMInter ; { TCommon_COM_Interface }

{ Device Media IDs: }
const DMI_Unknown = -1 ; // Unknown media type
const DMI_Remote = 4 ; // Remote (network) drive.
const DMI_CDROM = 5 ; // CD or DVD.
const DMI_RAM = 6 ; // RAM disk.
const DMI_Removable = $F0 ; // 3.5" DS, 18 sectors (or other removable)
const DMI_Fixed = $F8 ; // Fixed disk
const DMI_Floppy1 = $F9 ; // 5.25", DS, 15 sectors or 3.5" DS, 9 sectors
const DMI_Floppy2 = $FC ; // 5.25" SS, 9 sectors
const DMI_Floppy3 = $FD ; // 5.25" DS, 9 sectors
const DMI_Floppy4 = $FE ; // 5.25" SS, 8 sectors
const DMI_Floppy5 = $FF ; // 5.25" DS, 8 sectors

{ Device File System Flags: }
const DFSF_Case_Sensitive = 1 ; // Specifies that searches are case-sensitive.
const DFSF_Preserve_Case = 2 ; // Preserves case in directory entries.
const DFSF_UNICODE = 4 ; // Uses Unicode characters in file and directory names.
const DFSF_Long_Filenames = $4000 ; // Supports new long filename functions.
const DFSF_Volume_Compressed = $8000 ; // The volume is compressed.

{$IFDEF DELPHI}
type TCOM_Device_Manager = class( TCommon_COM_Interface )
                               public { API... }
                                 function Disk_Clusters( Drive : PChar ) : int64 ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : PChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : PChar ) : PChar ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : PChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : PChar ) : boolean ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : PChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : PChar ) : longint ;
                                     virtual ; stdcall ; abstract ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown. }

                                 function Valid_Device( Device : PChar ) : boolean ;
                                     virtual ; stdcall ; abstract ;
                                 { Returns true if the passed device is valid. }
                           end ;

type TDevice_Manager = class( TDelphi_Object )
                           public { API... }
                                 function Disk_Clusters( Drive : string ) : int64 ;
                                     virtual ; abstract ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : string ) : longint ;
                                     virtual ; abstract ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : string ) : string ;
                                     virtual ; abstract ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : string ) : longint ;
                                     virtual ; abstract ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : string ) : boolean ;
                                     virtual ; abstract ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : string ) : longint ;
                                     virtual ; abstract ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : string ) : longint ;
                                     virtual ; abstract ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown. }

                                 function Valid_Device( Device : string ) : boolean ;
                                     virtual ; abstract ;
                                 { Returns true if the passed device is valid. }
                       end ;
{$ENDIF}

type TSDevice_Manager = object( TObject )
                            public { API... }
                                 function Disk_Clusters( Drive : string ) : int64 ;
                                     virtual ;
                                 { Returns number of clusters on specified disk
                                   drive.  Returns -1 if an error or unknown. }

                                 function Disk_Sector_Size( Drive : string ) : longint ;
                                     virtual ;
                                 { Returns size of disk sectors, in bytes.
                                   Returns -1 if an error or unknown. }

                                 function File_System( Device : string ) : string ;
                                     virtual ;
                                 { Returns the name of the file system used on
                                   the device.  Returns a null string if the
                                   device is invalid or the file system is
                                   unknown. }

                                 function File_System_Flags( Device : string ) : longint ;
                                     virtual ;
                                 { Returns the flags for the file system.  See
                                   DFSF_*. }

                                 function Fixed_Device( Device : string ) : boolean ;
                                     virtual ;
                                 { Returns true if the passed device is fixed
                                   (non-removable) }

                                 function Max_Filename_Length( Device : string ) : longint ;
                                     virtual ;
                                 { Returns the maximum filename length for this
                                   device's file system. }

                                 function Media_ID( Device : string ) : longint ;
                                     virtual ;
                                 { Return type of media in/on device.  -1 if
                                   error or unknown.  See Device Media Ids. }

                                 function Valid_Device( Device : string ) : boolean ;
                                     virtual  ;
                                 { Returns true if the passed device is valid. }
                        end ;
     PSDevice_Manager = ^TSDevice_Manager ;

implementation

{ TSDevice_Manager methods... }

function TSDevice_Manager.Disk_Clusters( Drive : string ) : int64 ;

begin
    Disk_Clusters := -1 ;
end ;


function TSDevice_Manager.Disk_Sector_Size( Drive : string ) : longint ;

begin
    Disk_Sector_Size := -1 ;
end ;


function TSDevice_Manager.File_System( Device : string ) : string ;

begin
    File_System := '' ;
end ;


function TSDevice_Manager.File_System_Flags( Device : string ) : longint ;

begin
    File_System_Flags := 0 ;
end ;


function TSDevice_Manager.Fixed_Device( Device : string ) : boolean ;

begin
    Fixed_Device := False ;
end ;


function TSDevice_Manager.Max_Filename_Length( Device : string ) : longint ;

begin
    Max_Filename_Length := 1 ;
end ;


function TSDevice_Manager.Media_ID( Device : string ) : longint ;

begin
    Media_ID := -1 ;
end ;


function TSDevice_Manager.Valid_Device( Device : string ) : boolean ;

begin
    Valid_Device := False ;
end ;


end.
