{
        Program Name : _ObjectM
        Package Name : Subroutine Library - Layer 1
        Purpose      : Object Manager pure virtual classes
        Institution  : Conroy & Conroy Co.
        Date Written : 15-October-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The classes in this unit define the Object Manager interface.
}

unit _ObjectM ;

interface

uses { C&C... }
     ComInter, { TCommon_COM_Interface }
     Common, { TBase_Object }
     _DebugIn, { PDebug_Interface }
     _FacMan, { PSFacility_Manager }
     _Streams ; { TSStream }

{$I EDEFINES.INC}

type TSObject_Manager = object( TBase_Object )
                            public { Unit debugger support... }
                                { Instance lifetime support... }

                                { Add an object }
                                procedure Add_Object( O : pointer ;
                                    Facility : integer ;
                                    Context : pointer ) ; virtual ;

                                { Return a debugger }
                                function Debugger : PSDebug_Interface ;
                                    virtual ;

                                { Return True if specified object is registered }
                                function Find_Object( O : pointer ;
                                    Facility : integer ) : boolean ;
                                    virtual ;

                                function Get_Context( Index : integer ) : pointer ;
                                    virtual ;

                                { Get facility associated with index }
                                function Get_Facility( Index : integer ) : integer ;
                                    virtual ;

                                { Get object associated with index }
                                function Get_Object( Index : integer ) : pointer ;
                                    virtual ;

                                { Remove a registered object }
                                procedure Remove_Object( O : pointer ;
                                    Facility : longint ) ; virtual ;

                                { Returns count of registered obejcts }
                                function Object_Count : longint ;
                                    virtual ;


                                { Facility support... }

                                { Register a facility }
                                procedure Register_Facility( Facility_Manager : PSFacility_Manager ) ;
                                    virtual ;

                                { Remove a registered facility }
                                procedure Remove_Facility( Facility_Manager : PSFacility_Manager ) ;
                                    virtual ;

                                { Get facility manager at an index }
                                function Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;
                                    virtual ;

                                { Construct an instance of a facility }
                                function Construct( Stream : PSStream ) : pointer ;
                                    virtual ;
                        end ; { TSObject_Manager }
     PSObject_Manager = ^TSObject_Manager ;

{$IFDEF DELPHI}
type TObject_Manager = class( TBase_Delphi_Object )
                           public { Unit debugger support... }
                               { Add an object }
                               procedure Add_Object( O : pointer ;
                                   Facility : integer ;
                                   Context : pointer ) ; virtual ; abstract ;

                               { Return True if specified object is registered }
                               function Find_Object( O : pointer ;
                                   Facility : integer ) : boolean ;
                                   virtual ; abstract ;

                               function Get_Context( Index : integer ) : pointer ;
                                   virtual ; abstract ;

                               { Get facility associated with index }
                               function Get_Facility( Index : integer ) : integer ;
                                   virtual ; abstract ;

                               { Get object associated with index }
                               function Get_Object( Index : integer ) : pointer ;
                                   virtual ; abstract ;

                               { Remove a registered object }
                               procedure Remove_Object( O : pointer ;
                                   Facility : longint ) ; virtual ; abstract ;

                               { Returns count of registered obejcts }
                               function Object_Count : longint ;
                                   virtual ; abstract ;

                                { Construct an instance of a facility }
                                function Construct( Stream : TStream ) : pointer ;
                                    virtual ; abstract ;

                                procedure Register_Facility( Facility_Manager : TFacility_Manager ) ;
                                    virtual ; abstract ;

                                { Remove a registered facility }
                                procedure Remove_Facility( Facility_Manager : TFacility_Manager ) ;
                                    virtual ; abstract ;

                                { Get facility manager at an index }
                                function Get_Facility_Manager( Index : longint ) : TFacility_Manager ;
                                    virtual ; abstract ;
                       end ;

     TCOM_Object_Manager = class( TCommon_COM_Interface )
                               public { Unit debugger support... }
                                   { Add an object }
                                   procedure Add_Object( O : pointer ;
                                       Facility : integer ;
                                       Context : pointer ) ;
                                       virtual ; stdcall ; abstract ;

                                   { Return True if specified object is registered }
                                   function Find_Object( O : pointer ;
                                       Facility : integer ) : boolean ;
                                       virtual ; stdcall ; abstract ;

                                   function Get_Context( Index : integer ) : pointer ;
                                       virtual ; stdcall ; abstract ;

                                   { Get facility associated with index }
                                   function Get_Facility( Index : integer ) : integer ;
                                       virtual ; stdcall ; abstract ;

                                   { Get object associated with index }
                                   function Get_Object( Index : integer ) : pointer ;
                                       virtual ; stdcall ; abstract ;

                                   { Remove a registered object }
                                   procedure Remove_Object( O : pointer ;
                                       Facility : longint ) ;
                                       virtual ; stdcall ; abstract ;

                                   { Returns count of registered objects }
                                   function Object_Count : longint ;
                                       virtual ; stdcall ; abstract ;

                                   { Construct an instance of a facility }
                                   function Construct( Stream : TCOM_Stream ) : pointer ;
                                       virtual ; stdcall ; abstract ;

                                   procedure Register_Facility( Facility_Manager : TCOM_Facility_Manager ) ;
                                       virtual ; stdcall ; abstract ;

                                   { Remove a registered facility }
                                   procedure Remove_Facility( Facility_Manager : TCOM_Facility_Manager ) ;
                                       virtual ; stdcall ; abstract ;

                                   { Get facility manager at an index }
                                   function Get_Facility_Manager( Index : longint ) : TCOM_Facility_Manager ;
                                       virtual ; stdcall ; abstract ;
                           end ; { TCOM_Object_Manager }
{$ENDIF}

implementation

uses { Layer 0... }
     _Common ; { Abstract_Error }

{ TSObject_Manager methods... }

procedure TSObject_Manager.Add_Object( O : pointer ; Facility : integer ;
    Context : pointer ) ;

begin
    Abstract_Error ;
end ;


function TSObject_Manager.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
    Abstract_Error ;
end ;


function TSObject_Manager.Find_Object( O : pointer ;
    Facility : integer ) : boolean ;

begin
    Find_Object := False ;
    Abstract_Error ;
end ;


function TSObject_Manager.Get_Context( Index : integer ) : pointer ;

begin
    Get_Context := nil ;
    Abstract_Error ;
end ;


function TSObject_Manager.Get_Facility( Index : integer ) : integer ;

begin
    Get_Facility := -1 ;
    Abstract_Error ;
end ;


function TSObject_Manager.Get_Object( Index : integer ) : pointer ;

begin
    Get_Object := nil ;
    Abstract_Error ;
end ;


procedure TSObject_Manager.Remove_Object( O : pointer ; Facility : longint ) ;

begin
    Abstract_Error ;
end ;


function TSObject_Manager.Object_Count : longint ;

begin
    Object_Count := 0 ;
    Abstract_Error ;
end ;


procedure TSObject_Manager.Register_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
    Abstract_Error ;
end ;


procedure TSObject_Manager.Remove_Facility( Facility_Manager : PSFacility_Manager ) ;

begin
    Abstract_Error ;
end ;


function TSObject_Manager.Get_Facility_Manager( Index : longint ) : PSFacility_Manager ;

begin
    Get_Facility_Manager := nil ;
    Abstract_Error ;
end ;


function TSObject_Manager.Construct( Stream : PSStream ) : pointer ;

begin
    Construct := nil ;
    Abstract_Error ;
end ;


  
end.

