/*
        Program Name : DebugInt
        Package Name : LB
        Purpose      : Debugger Interface definition
        Institution  : Conroy & Conroy Co.
        Date Written : 11-Jan-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This header provides a C++ definition of the TDebug_Interface
	class.

	Released to the public domain.

    TDebug_Interface is used when debugging code by providing a way of examining
    the internals of the software component at run-time.  For production
    versions of software, this is often not supported.  The interface assumes a
    hierarchical form with the root of a component returning children
    representing internal objects.  Each of those objects may have children, and
    so forth.

*/

#if !defined(__DEBUGINT_H)
    #define __DEBUGINT_H
    
    class TDebug_Interface
    {
        public: // API...
            virtual __stdcall int Child_Count() = 0 ;
            /* Returns the number of child components under this level.  If this is
               a terminal (leaf) node, the method returns 0. */

            virtual __stdcall char* Get_Title() = 0 ;
            // Returns text representing this level (such as "Cached data").

            virtual __stdcall bool Has_Children() = 0 ;
            // Returns True if we have child nodes under this level.

            virtual void __stdcall Set_Title( char* Value ) = 0 ;
            // Sets text representing this level.

            virtual void __stdcall Activate() = 0 ;
            /* Indicates that the user wished to activate this level.  The
               behavior is undefined and up to the implementor.  Typically, this
               does nothing unless this level is a terminal representing some
               complex data which is not viewable in a hierarchical form. */

            virtual bool __stdcall Can_Change() = 0 ;
            // Returns True if this item can be modified.

            virtual TDebug_Interface* __stdcall Child( int Index ) = 0 ;
            /* Returns a debug interface for the specified child level (first
               child level is 0). */

            virtual void __stdcall Kill() = 0 ;
            // Tells object to destroy itself

            virtual char* __stdcall Modify( char* Data, int Size ) = 0 ;
            /* Modifies data.  If Size is -1, Data is assumed to be
               null-terminated and its length is automatically determined.  The
               result will be NULL or a pointer to a null string to indicate
               success, and error text otherwise. */

            virtual int __stdcall Get_Context() = 0 ;
            // Retrives user-context.

            virtual void __stdcall Set_Context( int Value ) = 0 ;
            // Set user-context.
    }
#endif
