/*
        Program Name : SMU
        Package Name : LB
        Purpose      : Screen Management Utility
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Mar-89
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1989 by Alan Conroy.  Released to the public domain.

	Compilation Note: Must use following options if using 16-bit
		BCC command-line compiler: -B -ml

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          These routines provide complete screen management.
*/

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>
#include <stdlib.h>
#include <mem.h>
#include <graphics.h>

#include "clib.h"
#include "smu.h"

char Kb_Buffer[ 256 ] ;
int Line_Count ;

static struct Screen_Def /* Physical screen definition */
{
    char Data[ 26 ][ 81 ] ;
    char Window[ 26 ][ 81 ] ;
} ;
static struct Input_Record
{
    struct Input_Record *Last_Input, *Next_Input ;
    /* Pointers to next and previous lines */
    char Input_Line[ 256 ] ; /* Actual input */
} ;
static struct Smu_Root /* Root data for screen management */
{
    void *Driver ; /* Pointer to video driver.  NULL if none */
    short int Model ;
    /* Model of PC: 1=Unknown (use ANSI.SYS), 2=Rainbow,
       3=Monochrome PC, 4=CGA/EGA/VGA/SVGA in text mode PC,
       5=CGA/EGA/VGA/SVGA in graphics mode PC,
       6=Hercules graphics */
    short int Default ; /* Current default window */
    struct Window *Default_P ; /* Pointer to default window */
    struct Window *Next_Window ; /* Pointer to first window */
    short int Window_Count ; /* Total windows defined */
    short int Printer_Type ; /* Type of printer
				1 = Unknown
                                2 = LA50/LA75
                                3 = IBM Graphics printer */
    int Insert_Mode ; /* True if insert mode */
    short int Printer_Mode ; /* Print mode: 0 = Off, 1 = On, 2 = Disabled */
    short int Input_Buffer_Max ; /* Number of lines to save */
    struct Input_Record *Last_Input ;
    /* Pointer to list of previous input lines */
    short int Video_Mode ; /* Type of video mode (a la PC) */
    short int Foreground, Background ; /* Fg/Bg colors */
    short int Palatte ; /* Color pallate selected */
    short int Screen_Width ; /* Maximum size of screen in columns */
    int Flags ; /* General flags:
			AND 1-16 = reserved
                        AND 32 = Disable cursor
                        AND 64-32768 = reserved */
} ;

/* Characters which "print" */
static const short int Print_Characters[ 256 ] = { True, True, True, True, True, True, True, False,
                       /* 8-14 */     False, True, False, True, True, False, True,
                       /* 15-22 */    True, True, True, True, True, True, True, True,
                       /* 23-30 */    True, True, True, True, True, True, True, True,
                       /* 31-38 */    True, True, True, True, True, True, True, True,
                       /* 39-46 */    True, True, True, True, True, True, True, True,
                       /* 47-54 */    True, True, True, True, True, True, True, True,
                       /* 55-62 */    True, True, True, True, True, True, True, True,
                       /* 63-70 */    True, True, True, True, True, True, True, True,
                       /* 71-78 */    True, True, True, True, True, True, True, True,
                       /* 79-86 */    True, True, True, True, True, True, True, True,
                       /* 87-94 */    True, True, True, True, True, True, True, True,
                       /* 95-102 */   True, True, True, True, True, True, True, True,
                       /* 103-110 */  True, True, True, True, True, True, True, True,
                       /* 111-118 */  True, True, True, True, True, True, True, True,
                       /* 119-126 */  True, True, True, True, True, True, True, True,
                       /* 127-134 */  True, True, True, True, True, True, True, True,
                       /* 135-142 */  True, True, True, True, True, True, True, True,
                       /* 143-150 */  True, True, True, True, True, True, True, True,
                       /* 151-158 */  True, True, True, True, True, True, True, True,
                       /* 159-166 */  True, True, True, True, True, True, True, True,
                       /* 167-174 */  True, True, True, True, True, True, True, True,
                       /* 175-182 */  True, True, True, True, True, True, True, True,
                       /* 183-190 */  True, True, True, True, True, True, True, True,
                       /* 191-198 */  True, True, True, True, True, True, True, True,
                       /* 199-206 */  True, True, True, True, True, True, True, True,
                       /* 207-214 */  True, True, True, True, True, True, True, True,
                       /* 215-222 */  True, True, True, True, True, True, True, True,
                       /* 223-230 */  True, True, True, True, True, True, True, True,
                       /* 231-238 */  True, True, True, True, True, True, True, True,
                       /* 239-246 */  True, True, True, True, True, True, True, True,
                       /* 247-254 */  True, True, True, True, True, True, True, False,
                       /* 255 */      True } ;
/* Characters which affect cursor position */
static const short int Cursor_Characters[ 256 ] = { False, False, False, False, False, False,
                   /* 6-12 */     False, False, True, True, False, False, False,
                   /* 13-19 */    True, False, False, False, False, False, False,
                   /* 20-25 */    False, False, False, False, False, False,
                   /* 26-32 */    False, False, False, False, False, False, True,
                   /* 33-40 */    True, True, True, True, True, True, True, True,
                   /* 41-48 */    True, True, True, True, True, True, True, True,
                   /* 49-56 */    True, True, True, True, True, True, True, True,
                   /* 57-64 */    True, True, True, True, True, True, True, True,
                   /* 65-72 */    True, True, True, True, True, True, True, True,
                   /* 73-80 */    True, True, True, True, True, True, True, True,
                   /* 81-88 */    True, True, True, True, True, True, True, True,
                   /* 89-96 */    True, True, True, True, True, True, True, True,
                   /* 97-104 */   True, True, True, True, True, True, True, True,
                   /* 105-112 */  True, True, True, True, True, True, True, True,
                   /* 113-120 */  True, True, True, True, True, True, True, True,
                   /* 121-127 */  True, True, True, True, True, True, False,
                   /* 128-133 */  True, True, True, True, True, True,
                   /* 134-139 */  True, True, True, True, True, True,
                   /* 140-145 */  True, True, True, True, True, True,
                   /* 146-151 */  True, True, True, True, True, True,
                   /* 152-157 */  True, True, True, True, True, True,
                   /* 158-163 */  False, False, True, True, True, True,
                   /* 164-169 */  True, True, True, True, True, True,
                   /* 170-175 */  True, True, True, True, True, True,
                   /* 176-181 */  True, True, True, True, True, True,
                   /* 182-187 */  True, True, True, True, True, True,
                   /* 188-193 */  True, True, True, True, True, True,
                   /* 194-199 */  True, True, True, True, True, True,
                   /* 200-205 */  True, True, True, True, True, True,
                   /* 206-211 */  True, True, True, True, True, True,
                   /* 212-217 */  True, True, True, True, True, True,
                   /* 218-223 */  True, False, False, False, False, False,
                   /* 224-229 */  False, True, False, True, False, False,
                   /* 230-235 */  True, False, False, False, False, False,
                   /* 236-241 */  False, True, False, False, False, True,
                   /* 242-247 */  True, True, False, False, False, False,
                   /* 248-253 */  True, True, False, False, False, True,
                   /* 254-255 */  True, False } ;

static short int Attribute_Default ; /* Default attribute value */
static struct FILE *Printer_File ;
static int Dummy, Dummy1 ; /* Dummy loop variable */
static int Input_Row, Input_Column ; /* Row and Column at beginning of input */
static struct Screen_Def Screen ; /* Define screen */
static struct Smu_Root Screen_Root ;
static char Attribute_String[ 82 ], Screen_String[ 82 ] ;
static int Text_Height, Text_Width ; /* Size of font characters */
static char Output_Buffer[ 256 ] ;
static int Row_Start[ 25 ] ; /* Start of rows in memory */
static char *Window_Data ;
typedef char Character[ 2 ] ;
static Character far *Monochrome ;

static void Set_Curs( short int X, short int Y )
/* Display/remove cursor at location X,Y.  1,1 is home position */
{

int Line_Count, Lines_So_Far, Z, Rows ;
union REGS Reg ;

    if( Screen_Root.Video_Mode != 7 ) /* Non-text */
    {
        Reg.w.ax = 0x0900 ; /* Read character */
        int86( 0x10, &Reg, &Reg ) ; /* Get info */
        Z = Reg.w.ax ; /* Get character */
        Reg.w.ax = 0x0900 | ( Z & 255 ) ; /* Character to write */
        Reg.w.cx = 1 ; /* Multiplier */
        Reg.w.bx = Screen_Root.Foreground * 16 + Screen_Root.Background ;
        int86( 0x10, &Reg, &Reg ) ; /* Reverse it now */
    } ;
} ;

struct Window *Current_Window() /* Return current window pointer */
{
    return Screen_Root.Default_P ; /* Return pointer for current window */
} ;

int Window_Mapped( struct Window *Window_P )
/* Return TRUE/FALSE regarding whether or not window is mapped */
{
    if( Window_P->Map_Row > 0 )
    {
	return True ;
    }
    else
    {
	return False ;
    } ;
} ;

void Get_Default_Window( struct Window_Record *X )
{
    X->Pointer = Screen_Root.Default_P ;
    X->Id = Screen_Root.Default ; /* Return default window */
} ;

void Set_Default_Window( struct Window_Record X )
{
    Screen_Root.Default_P = X.Pointer ;
    Screen_Root.Default = X.Id ; /* Set default window */
} ;

static void Ibm_Pos( int X, int Y ) /* Position IBM cursor via BIOS */
{

union REGS R ;

    R.w.ax = 0x0200 ; /* Set cursor */
    R.w.bx = 0 ; /* Page */
    R.h.dh = X ;
    R.h.dl = Y ; /* Position */
    int86( 0x10, &R, &R ) ; /* Call BIOS */
} ;

short int Get_Model() /* Return system type */
{
    return Screen_Root.Model ;
} ;

void End_Smu() /* Nicely terminate SMU */
{

union REGS Reg ;

    if( Screen_Root.Model != 1 ) /* IBM */
    {
	Ibm_Pos( 24, 1 ) ;
    } ;
} ;

static void Init_Screen() /* Initialize depending upon hardware model */
{

int R ;
union REGS Reg ;

    Screen_Root.Video_Mode = 7 ; /* Text is default */
    Screen_Root.Screen_Width = 80 ; /* Default screen width */
    Screen_Root.Foreground = 0 ;
    Screen_Root.Background = 0 ;
    Screen_Root.Flags = 0 ;
    switch( Screen_Root.Model )
    {
	case 3 :
	case 4 : /* IBM PC in text mode */
	{
            Attribute_Default = 0x07 ;
            for( R = 1; R <= 24; R++ )
	    {
		Row_Start[ R ] = ( R - 1 ) * 80 ;
	    } ;
            Reg.w.ax = 0x0500 ;
            /* Set page to 0 (which initializes and clears screen) */
            int86( 0x10, &Reg, &Reg ) ; /* Do BIOS call */
	    Reg.w.ax = 0x0200 ;
	    Reg.w.bx = 0 ;
	    Reg.w.dx = 81 ;
	    int86( 0x10, &Reg, &Reg ) ; /* Disable cursor */
        } ;
     } ;
     if( Screen_Root.Model > 4 ) /* IBM PC graphics */
     {
         Reg.w.ax = 0x0F00 ;
         int86( 0x10, &Reg, &Reg ) ; /* Get video mode */
         Screen_Root.Video_Mode = Reg.h.al ;
         Text_Height = 7 ;
         Text_Width = 5 ;
     } ;
} ;

struct Window *Create_Window( int Row_Count, int Column_Count, int Flag )
/* Create a window */
{

struct Window *Last_Pointer, *Window_P ;
char *W_Pointer ;
int A, Size ;
short int Wid ;
int Wida[ 256 ] ;

    Size = Row_Count * Column_Count ;
    if( Size < Row_Count || Size < Column_Count )
    {
        puts( "?Window too large" ) ;
        return NULL ;
    } ;
    if( Screen_Root.Window_Count == 255 )
    {
        puts( "?Too many windows" ) ;
        return NULL ;
    } ;
    for( A = 1; A <= 255; A++ )
    {
	Wida[ A ] = False ; /* Prepare for loop */
    } ;
    Last_Pointer = NULL ;
    Window_P = Screen_Root.Next_Window ; /* Start with first window in list */
    while( Window_P != NULL ) /* Find available/used WIDs */
    {
         Wida[ Window_P->Id ] = True ;
         Last_Pointer = Window_P ; /* Last window pointed to */
         Window_P = Window_P->Next_Window ; /* Go to next window */
    } ;
    Wid = 0 ;
    A = 0 ;
    while( Wid == 0 ) /* Search until unused WID assigned */
    {
         A = A + 1 ;
         if( !Wida[ A ] )
	{
	    Wid = A ; /* If unused WID */
	} ;
    } ;
    Screen_Root.Window_Count = Screen_Root.Window_Count + 1 ;
    /* One more window */
    Window_P = malloc( 32767 ) ; /* Create window header */
    Window_P->Attribute_Pointer = malloc( Size ) ;
    Window_P->Data_Pointer = malloc( Size ) ; /* Create window */
    Window_P->Level = 0 ;
    Window_P->Rows = Row_Count ;
    Window_P->Columns = Column_Count ;
    Window_P->Map_Row = 0 ;
    Window_P->Map_Column = 0 ;
    Window_P->View_Row = 0 ;
    Window_P->View_Column = 0 ;
    Window_P->Flags = Flag | 2 ; /* Occluded */
    Window_P->Id = Wid ;
    Window_P->Next_Window = Screen_Root.Next_Window ;
    Window_P->Position_Row = 1 ; /* Current character position */
    Window_P->Position_Column = 1 ;
    Window_P->Low_Update = 32767 ;
    Window_P->High_Update = 0 ;
    Window_P->Size = Size ;
    Window_P->Attribute = Attribute_Default ;
    Screen_Root.Next_Window = Window_P ; /* Insert window in list */
    memset( Window_P->Attribute_Pointer, Attribute_Default, Size ) ;
    /* Null attributes */
    memset( Window_P->Data_Pointer, 32, Size ) ; /* Space fill */
    /* Now window is initialized */
    return Window_P ;
} ;


static void Do_Unmap( struct Window *Wp ) /* Physical unmap of a window */
{

int A, Offset ;
int Clear_Partial, Clear_Full ;
short int Last_Window ; /* Last offset used in Window_* arrays */
int Level, Level_Window ;
int R, C ; /* Row, column pointers */
struct Window *This_Window ;
struct Window *Window_List[ 255 ] ;
short int Window_Flag[ 255 ] ;
struct Window *Window_Pointer0, *Window_Pointer1 ;

    memset( Window_List, 0, 1020 ) ;
    memset( Window_Flag, 0, 255 ) ; /* Zero window flags */
    Last_Window = 0 ;
    Window_Pointer0 = Screen_Root.Next_Window ;
    /* Start with first window defined */
    while( Window_Pointer0 != NULL ) /* Until done with all windows */
    {
        if( Window_Pointer0 != Wp ) /* Not the same window we're unmapping */
        {
            if( Window_Pointer0->Map_Row < Wp->Map_Row + Wp->Map_Rows &&
                Window_Pointer0->Map_Column < Wp->Map_Column + Wp->Map_Columns &&
                Wp->Map_Row < Window_Pointer0->Map_Row + Window_Pointer0->Map_Rows
                && Wp->Map_Column <
                  Window_Pointer0->Map_Column + Window_Pointer0->Map_Columns )
                    { /* If window we are unmapping occludes another */
                      Window_List[ Last_Window ] = Window_Pointer0 ;
                      Last_Window = Last_Window + 1 ;
                      /* Build list of occluded windows */
                    } ;
        } ;
        Window_Pointer0 = Window_Pointer0->Next_Window ;
	/* Follow pointer to next defined window */
    } ;
    for( R = Wp->Map_Row; R <= Wp->Map_Row + Wp->Map_Rows - 1; R++ )
    {
    /* For all rows */
        for( C = Wp->Map_Column; C <= Wp->Map_Column + Wp->Map_Columns - 1; C++ )
        /* And columns */
        {
            Screen.Window[ R ][ C ] = 0 ;
	    /* Assume no window maps this cell anymore */
            Level = 0 ;
            Level_Window = - 1 ;
	    if( Last_Window > 0 )
	    {
                for( A = 0; A <= Last_Window - 1; A++ )
	        /* For all possible revealed windows */
                {
		    This_Window = Window_List[ A ] ;
		    /* Get pointer to next window */
		    if( This_Window->Level >= Level &&
			/* Closest to top so far */
			This_Window->Map_Row <= R &&
			This_Window->Map_Column <= C &&
	                This_Window->Map_Column + This_Window->Map_Columns > C &&
	                This_Window->Map_Row + This_Window->Map_Rows > R )
		  	/* and maps same row,column position */
		    {
			Level = This_Window->Level ;
			Screen.Window[ R ][ C ] = This_Window->Id ;
			Level_Window = A ;
		    } ;
              } ;
              /* At this point, Level_Window is the window offset located at
		 this cell */
            } ;
	    if( Level_Window >= 0 ) /* If a window maps this cell */
            {
                Offset = ( R - Window_List[ Level_Window ]->Map_Row +
                    Window_List[ Level_Window ]->View_Row - 1 ) *
                    Window_List[ Level_Window ]->Columns +
                    C - Window_List[ Level_Window ]->Map_Column +
                    Window_List[ Level_Window ]->View_Column - 1 ;
                if( Window_List[ Level_Window ]->Low_Update > Offset )
		{
                    Window_List[ Level_Window ]->Low_Update = Offset ;
		} ;
                if( Window_List[ Level_Window ]->High_Update < Offset )
		{
                    Window_List[ Level_Window ]->High_Update = Offset ;
		} ;
		Window_Flag[ Level_Window ] = 1 ;
		/* Window occlusion has changed */
            } ;
	} ;
    } ;
    if( Last_Window == 0 )
    {
	return ; /* No revealed windows */
    } ;

    for( A = 0; A <= Last_Window - 1; A++ )
    {
	if( Window_Flag[ A ] > 0 ) /* Check for occulsion */
        {
            Window_List[ A ]->Flags = Window_List[ A ]->Flags | 18 ;
	    Clear_Partial = False ;
	    Clear_Full = False ;
            for( R = Window_List[ A ]->Map_Row; R <=
              Window_List[ A ]->Map_Row + Window_List[ A ]->Map_Rows - 1; R++ )
              /* For all rows */
	    {
                for( C = Window_List[ A ]->Map_Column; C <=
                  Window_List[ A ]->Map_Column + Window_List[ A ]->Map_Columns - 1; C++ )
                /* And columns */
		{
                    if( Screen.Window[ R ][ C ] != Window_List[ A ]->Id )
		    {
			Clear_Partial = True ; /* Not 100% unoccluded */
		    }
		    else
		    {
			Clear_Full = True ; /* Not 100% occluded */
		    } ;
		} ;
	    } ;
	    if( Clear_Full )
	    {
		Window_List[ A ]->Flags =
		    Window_List[ A ]->Flags & ( ~2 ) ;
	    } ;
	    if( Clear_Partial )
	    {
		Window_List[ A ]->Flags =
		    Window_List[ A ]->Flags & ( ~16 ) ;
	    } ;
        } ;
    } ;
} ;


void Unmap_Window( struct Window *Window_P )
/* Unmap a window from the physical screen */
{
    if( ( Window_P->Flags & 2 ) == 2 )
    {
	return ; /* Window is 100% occluded */
    } ;
    Window_P->Flags = Window_P->Flags | 2 ; /* Window is occluded */
    Window_P->Flags = Window_P->Flags & ( ~16 ) ; /* Clear unoccluded flag */
    if( Window_P->Map_Row > 0 )
    {
	Do_Unmap( Window_P ) ; /* Unmap, if needed */
    } ;
    if( Screen_Root.Default_P == Window_P ) /* Unmapping default window */
    {
        if( Screen_Root.Next_Window == NULL )
	/* No other windows that could be mapped */
	{
	    Screen_Root.Default = 0 ;
	    Screen_Root.Default_P = NULL ;
	}
	else /* Map next window in list */
        {
            Screen_Root.Default = Screen_Root.Next_Window->Id ;
            Screen_Root.Default_P = Screen_Root.Next_Window ;
        } ;
    } ;
    Window_P->Map_Row = 0 ;
    Window_P->Map_Column = 0 ;
    Window_P->Map_Rows = 0 ;
    Window_P->Map_Columns = 0 ;
    Window_P->View_Row = 0 ;
    Window_P->View_Column = 0 ;
    /* Finish unmapping window */
} ;

void Map_Window( struct Window *Window_P, short int Row, short int Row_Count,
    short int Column, short int Column_Count, short int View_R,
    short int View_C, short int Level )
/* Map/remap a window to the physical screen */
/* Row, Row_Count, Column, and Column_Count define the location of the viewport
 on the physical screen.  View_R and View_C define the origin of the viewport
 in the virtual screen (the row and column of the virtual screen that will
 appear in the upper-left-hand corner of the viewport) */
{

int C, R ;
int Collision, Occluded, Partial ;
short int Old_Default ;
struct Window *Window_Pointer1 ;
short int Wid ;

    /* The next section allows the current values to remain unchanged if 0 is
     passed for row, column, row and column counts, and/or view port stuff */
    if( Row == 0 )
    {
	Row = Window_P->Map_Row ;
    } ;
    if( Column == 0 )
    {
	Column = Window_P->Map_Column ;
    } ;
    if( Row_Count == 0 )
    {
	Row_Count = Window_P->Map_Rows ;
    } ;
    if( Column_Count == 0 )
    {
	Column_Count = Window_P->Map_Columns ;
    } ;
    if( View_R == 0 )
    {
	View_R = Window_P->View_Row ;
    } ;
    if( View_C == 0 )
    {
	View_C = Window_P->View_Column ;
    } ;

    Old_Default = Screen_Root.Default ;
    /* Save default window for later comparison (unmap changes it if unmapping
       it) */
    Unmap_Window( Window_P ) ; /* Unmap window, if it is mapped */
    Wid = Window_P->Id ;
    if( Screen_Root.Default == 0 || Old_Default == Wid )
    /* Select this window */
    {
        Screen_Root.Default = Wid ;
        Screen_Root.Default_P = Window_P ;
    } ;
    if( Row > 24 || Column > Screen_Root.Screen_Width )
    {
	return ; /* Not on screen */
    } ;
    if( Row + Row_Count > 25 )
    {
	Row_Count = 25 - Row ;
    } ;
    if( Column + Column_Count > Screen_Root.Screen_Width + 1 )
    {
        Column_Count = Screen_Root.Screen_Width - Column + 1 ;
    } ;
    if( View_R > Row_Count )
    {
	View_R = Row_Count ;
    } ;
    if( View_C > Column_Count )
    {
	View_C = Column_Count ;
    } ;
    Collision = False ;

    Occluded = True ; /* Assume this window is 100% occluded */
    Partial = False ;
    Window_P->Flags = Window_P->Flags | 18 ;

    for( C = Column; C <= Column + Column_Count - 1; C++ ) /* For each column */
    {
        for( R = Row; R <= Row + Row_Count - 1; R++ )
	/* For each row in column */
	{
            if( Screen.Window[ R ][ C ] == 0 ) /* If nothing else mapped */
	    {
		Occluded = False ; /* At least partially unoccluded */
		Screen.Window[ R ][ C ] = Wid ;
	    }
	    else /* Something else mapped */
	    {
		Window_Pointer1 = Screen_Root.Next_Window ;
                while( Window_Pointer1->Id != Screen.Window[ R ][ C ] )
                {
		    Window_Pointer1 = Window_Pointer1->Next_Window ;
		    /* Locate window which is mapped to this cell */
		} ;
                if( Window_Pointer1->Level < Level )
		/* Newly mapped window occludes currently mapped */
		{
		    Occluded = False ; /* Not 100% occluded */
		    Window_Pointer1->Flags =
			Window_Pointer1->Flags & ( ~16 ) ;
			/* Other window is partially occluded */
		    Screen.Window[ R ][ C ] = Wid ;
		}
		else /* This window is occluded by something else */
		{
		    Partial = True ; /* At least partial occlusion */
		    if( Window_Pointer1->Level == Level )
		    {
			Collision = True ; /* Window collision */
		    } ;
		} ;
	    } ;
	} ;
    } ;
    if( Collision )
    {
	puts( "%Window collision" ) ;
    } ;
    if( !Occluded ) /* Window is no longer 100% occluded */
    {
        Window_P->Flags = Window_P->Flags & ( ~2 ) ;
    } ;
    Window_P->Map_Row = Row ;
    Window_P->Map_Column = Column ;
    Window_P->Map_Rows = Row_Count ;
    Window_P->Map_Columns = Column_Count ;
    Window_P->View_Row = View_R ;
    Window_P->View_Column = View_C ;
    Window_P->Low_Update = 0 ;
    Window_P->High_Update = Window_P->Size - 1 ;
    Window_P->Level = Level ;
    if( Partial || Occluded )
    {
	Window_P->Flags = Window_P->Flags & ( ~16 ) ;
	/* Not 100% unoccluded */
    } ;
} ;

void Delete_Window( struct Window *Wp ) /* Delete a window */
{

struct Window *Last_Pointer, *Window_P ;

    Unmap_Window( Wp ) ; /* Unmap it */
    Window_P = Screen_Root.Next_Window ;
    Last_Pointer = NULL ;
    while( Window_P != Wp && Window_P != NULL )
    {
        Last_Pointer = Window_P ;
        Window_P = Window_P->Next_Window ;
    } ;
    if( Last_Pointer == NULL ) /* First screen */
    {
	Screen_Root.Next_Window = Window_P->Next_Window ;
    }
    else
    {
	Last_Pointer->Next_Window = Window_P->Next_Window ;
    } ;
    free( Window_P->Data_Pointer ) ;
    free( Window_P->Attribute_Pointer ) ; /* Delete window */
    free( Window_P ) ; /* Delete window header */
} ;


static void Write_Screen( short int L, short int C )
/* Write text to physical display */
{

short int Row, Column, Char, Char1 ;
int Len, Temp, X ;
union REGS Reg ;

    Len = strlen( Screen_String ) ;
    if( Screen_Root.Driver != NULL ) /* A screen driver */
    {
        Row = L ;
        Column = C ;
        X = 0 ;
        while( X < Len ) /* For each character */
        {
            Char = Screen_String[ X ] ;
            Char1 = Attribute_String[ X ] ;
	    asm
	    {
		push ds
		push bp
		mov bx, OFFSET Screen_Root.Driver
		mov ah, BYTE PTR Row
		mov al, BYTE PTR Column
		mov dl, BYTE PTR Char
		mov dh, BYTE PTR Char1
		mov si, 0
		call [bx+si]
		pop bp
		pop ds
            } ;
            X ++ ;
        } ;
        return ;
    } ;
    if( Screen_Root.Model == 1 ) /* Unknown machine type - use ANSI.SYS */
    {
	printf( "\x27[%i;%iH%s", L, C, Screen_String ) ;
	return ;
    } ; /* if( Screen_Root.Model == 1 ) */
    if( Screen_Root.Model == 3 ) /* Monochrome PC */
    {
	L -- ;
	C -- ;
	for( Temp = 0; Temp < Len; Temp++ )
	{
	    Ibm_Pos( L, C ) ; /* Position cursor */
	    C ++ ;
	    Reg.h.ah = 0x09 ;
	    Reg.h.al = Screen_String[ Temp ] ;
	    Reg.h.bh = 0 ;
	    Reg.h.bl = Attribute_String[ Temp ] & 127 ;
	    Reg.w.cx = 1 ;
	    int86( 0x10, &Reg, &Reg ) ; /* Write character */
	} ; /* for( Temp = 0; Temp < Len; Temp++ ) */
	return ;
    } ; /* if( Screen_Root.Model == 3 ) */
    if( Screen_Root.Model == 4 ) /* CGA/EGA/VGA/SVGA PC in text mode */
    {
	L -- ;
	C -- ;
	for( Temp = 0; Temp < Len; Temp++ )
	{
	    Ibm_Pos( L, C ) ; /* Position cursor */
	    C ++ ;
	    Reg.h.ah = 0x09 ;
	    Reg.h.al = Screen_String[ Temp ] ;
	    Reg.h.bh = 0 ;
	    Reg.h.bl = Attribute_String[ Temp ] & 127 ;
	    Reg.w.cx = 1 ;
	    int86( 0x10, &Reg, &Reg ) ; /* Write character */
	} ; /* for( Temp = 0; Temp < Len; Temp++ ) */
        return ;
    } ; /* if( Screen_Root.Model == 4 ) */
    if( Screen_Root.Model == 5 ) /* CGA/EGA/VGA/SVGA PC in graphics mode */
    {
        Ibm_Pos( L - 1, C - 1 ) ; /* Position cursor */
	C = ( C - 1 ) * Text_Width ;
	L = ( L - 1 ) * Text_Height ;
	outtextxy( C, L, Screen_String ) ; /* Output text */
    } ; /* if( Screen_Root.Model == 5 ) */
} ;

void Clear_Window( struct Window *Window_P ) /* Clear a window */
{

int A ; /* Loop variable */

    memset( Window_P->Data_Pointer, 32, Window_P->Size ) ;
    memset( Window_P->Attribute_Pointer, Window_P->Attribute,
	Window_P->Size ) ;
    Window_P->Position_Row = 1 ;
    Window_P->Position_Column = 1 ;
    Window_P->Low_Update = 0 ;
    Window_P->High_Update = Window_P->Size ;
} ;

void Update_Window( struct Window *Window_P )
/* Update on screen what has changed on window */
{

int A ; /* Loop counter */
int Columns ; /* Columns in window */
int Dummy, Dummy1, Dummy2 ; /* Dummy variables used to speed up execution */
int Lr, Lc, R, C, R1, C1, Vr, Vc ; /* Row and column pointers */
int Low, High ; /* Update range */
int Offset ; /* Offset of current row - to reduce multiplies */
int Offset_Vc ; /* Offset+VC */
int Unoccluded ;
short int Screen_Width ; /* Width of screen */
int Stop ;
short int Wid ;
char *Window_A_P, *Window_D_P ; /* Used for speed */

    if( Window_P == NULL ) /* No window to update */
    {
	return ;
    } ;
    if( ( Window_P->Flags & 2 ) == 2 )  /* Window is occluded */
    {
	return ;
    } ;
    Low = Window_P->Low_Update ;
    High = Window_P->High_Update ;
    if( Low > High ) /* Nothing to update */
    {
	return ;
    } ;
    Vr = Window_P->View_Row ;
    Vc = Window_P->View_Column ;
    R = Window_P->Map_Row ; /* First row in viewport */
    Columns = Window_P->Columns ; /* Get offset to start of view row */
    Offset = ( Vr - 1 ) * Columns ;
    /* Get offset to start of first (current) view port row */
    Offset_Vc = Offset + Vc - 1 ;
    /* Offset to first (current) position in view port */
    Screen_Width = Screen_Root.Screen_Width ;

    Lr = R + Window_P->Map_Rows - 1 ; /* Last row in viewport */
    if( Window_P->Rows - Vr + 1 < Window_P->Map_Rows )
    {
	Lr = Window_P->Rows - Vr + R ;
    } ;
    C = Window_P->Map_Columns ; /* Width */
    if( Window_P->Columns - Vc + 1 < Window_P->Map_Columns )
    {
	C = Window_P->Columns - Vc + Window_P->Map_Column ;
	Screen_Width = C ;
    } ;
    /* Handle viewport which goes past edge of window */

    Screen_String[ C ] = 0 ;
    Attribute_String[ C ] = 0 ;
    Dummy = Window_P->Map_Column ;
    Wid = Window_P->Id ;
    if( ( Window_P->Flags & 16 ) == 16 && Screen_Root.Model != 1 )
    {
	Unoccluded = True ;
    }
    else
    {
	Unoccluded = False ;
    } ;
    Window_D_P = Window_P->Data_Pointer ;
    Window_A_P = Window_P->Attribute_Pointer ;
    R1 = R ;
    while( R1 <= Lr ) /* For each row in view port */
    {
        if( Unoccluded ) /* 100% unoccluded and not ANSI.SYS */
	{
            if( Offset_Vc > High )
            /* If we're already past the last updated character */
	    {
		goto Done ;
	    } ;

            if( Offset_Vc + Window_P->Columns >= Low )
            /* Past beginning of updated area */
            {
                memcpy( Screen_String, &Window_D_P[ Offset_Vc ], C ) ;
                memcpy( Attribute_String, &Window_A_P[ Offset_Vc ], C ) ;
                /* Update virtual window */
                Write_Screen( R1, Dummy ) ;
                /* Write row to physical screen */
            } ;
        }
        else /* Window is partially occluded */
        {
            C = Window_P->Map_Column ; /* Physical screen column */
            if( Offset_Vc > High )
	    {
		goto Done ;
	    } ;
            /* All done if we're already past the last updated character */

            if( Offset_Vc + Window_P->Map_Columns >= Low )
            /* If we're up to first updated character */
	    {
		while( C <= Screen_Width )
		/* Until we hit the right side of the screen */
		{
                    if( Screen.Window[ R1 ][ C ] == Wid &&
			C <= Screen_Root.Screen_Width &&
                        Offset_Vc + C > Low &&
                        Offset_Vc + C - 1 <= High )
			/* If within the updated area */
                    {
                        Stop = C ;
                        while( Screen.Window[ R1 ][ Stop ] == Wid &&
			    Stop <= Screen_Root.Screen_Width &&
			    Offset_Vc + Stop > Low &&
                            Offset_Vc + Stop - 1 <= High )
			{
			    Stop ++ ;
                            /* Find edge of changed window on screen for
                               this line */
			} ;
			Stop -- ; /* Don't go past this window */
			Lc = Stop - C + 1 ;
			Screen_String[ 0 ] = Lc ;
			/* LC=Number of characters to update */
			Dummy2 = Vc + C - 1 ; /* Screen column */
			Dummy = Offset + C - Vc ; /* Window offset */
			memcpy( &Screen_String[ 1 ], &Window_D_P[ Dummy ], Lc ) ;
			memcpy( &Attribute_String[ 1 ], &Window_A_P[ Dummy ], Lc ) ;
			if( Screen_Root.Model == 1 ) /* ANSI.SYS */
			{
			    memcpy( &Screen.Data[ R1 , Dummy2 ],
				&Window_D_P[ Dummy ], Lc ) ;
                            Write_Screen( R1, C ) ;
                            C = Stop ;
                        } ;
                        C ++ ;
                    } ;
                    Offset = Offset + Columns ;
                    /* Increment pointer into window */
                } ;
	    } ; /* if */
	} ; /* if( Unoccluded ) */
	Offset_Vc = Offset_Vc + Columns ; /* Offset into virtual screen */
	R1 ++ ;
    } ; /* while */
Done:
    Window_P->Low_Update = 32767 ;
    Window_P->High_Update = 0 ;
    /* Nothing needs updates anymore */
} ;

void Update_Screen() /* Update all windows */
{

struct Window *Window_P ;

    Window_P = Screen_Root.Next_Window ; /* Get first window */
    while( Window_P != NULL ) /* Until end of windows */
    {
        Update_Window( Window_P ) ;
        Window_P = Window_P->Next_Window ;
    } ;
} ;

void Refresh_Window( struct Window *Window_P )
{
    if( ( Window_P->Flags & 2 ) == 2 ) /* Window is occluded */
    {
	return ;
    } ;
    Window_P->Low_Update = 0 ;
    Window_P->High_Update = Window_P->Size - 1 ;
    Update_Window( Window_P ) ;
} ;

void Refresh_Screen() /* Refresh physical screen */
{

struct Window *P ;

    P = Screen_Root.Next_Window ; /* Get first window */
    while( P != NULL ) /* For each existing window */
    {
        Refresh_Window( P ) ;
        P = P->Next_Window ;
    } ;
} ;

static void Handle_Lf( int *Do_Update, int *Position_Row, int *Position_Column,
    int *Rows, int *Flags, int *Low_Update, int *High_Update, int *Size,
    int *Columns, int *Offset, int *Attribute, char *Window_A_P,
    char *Window_D_P ) /* Handling of Linefeed */
{

int B, Offset1 ;

    *Do_Update = True ; /* Update potentially needed */
    *Position_Row = *Position_Row + 1 ; /* Increment row */
    if( *Position_Row > *Rows ) /* Past end of window */
    {
	*Low_Update = 0 ; /* Refresh from beginning of window */
	if( ( *Flags & 1 ) == 0 ) /* Scroll */
	{
	    Offset1 = *Size - *Columns ;
            /* Number of bytes from origin to end of next-to-last line */
            memmove( Window_D_P, &Window_D_P[ *Columns ], Offset1 ) ;
            memmove( Window_A_P, &Window_A_P[ *Columns ], Offset1 ) ;
            /* Do scroll */
            B = *Size - *Columns ; /* Beginning of last line */
            memset( &Window_D_P[ B ], 32, *Columns ) ;
            memset( &Window_A_P[ B ], *Attribute, *Columns ) ;
            /* Zero last line */
            *High_Update = *Size - 1 ; /* End of window */
            *Position_Row = *Rows ;
            /* Current row is last row of window */
            *Offset = B ; /* New offset */
            Input_Row -- ;
            if( Input_Row < 1 )
            {
		Input_Row = 1 ;
		Input_Column = 1 ;
            } ;
            return ;
        }
        else /* No scroll */
        {
            *Position_Row = 1 ; /* To top of window */
            Offset1 = *Position_Column - 1 ; /* To top row */
            if( Offset1 > *High_Update )
	    {
		*High_Update = Offset1 ;
	    } ;
        } ;
    }
    else /* Not past end of window */
    {
	Offset1 = *Offset + *Columns ; /* Move to next line */
	if( Offset1 < *Low_Update )
	{
	    *Low_Update = Offset1 ;
	} ;
	if( Offset1 > *High_Update )
	{
	    *High_Update = Offset1 ;
	} ;
    } ;
} ;

void Window_Output( char *Output_Buffer ) /* Output to current window */
{

int A, B, C ;
int Columns ; /* Window column count */
int Dummy, Dummy1 ; /* Dummy variables to speed execution */
int Do_Update ;
struct Window *Window_P ;
int Offset ;
int Position_Row, Position_Column, Rows, Flags ;
char Ch, Ch1 ;
int Low_Update, High_Update ;
union REGS R ;
int Size ;
int Ord_Ch, Ord_Ch1 ;
char *Window_A_P, *Window_D_P ; /* For speed */
short int Attribute ;
int Wrap ;

    Window_P = Screen_Root.Default_P ; /* Find window */
    if( Window_P != NULL )
    {
	Position_Row = Window_P->Position_Row ;
	Position_Column = Window_P->Position_Column ;
	Rows = Window_P->Rows ;
	Columns = Window_P->Columns ; /* Get columns in window */
	Size = Window_P->Size ; /* Get size of window */
	Low_Update = Window_P->Low_Update ;
	High_Update = Window_P->High_Update ;
	Flags = Window_P->Flags ;
	if( ( Flags & 4 ) == 0 )
	{
	    Wrap = True ;
	}
	else
	{
	    Wrap = False ;
	} ;
	Attribute = Window_P->Attribute ;
	Window_A_P = Window_P->Attribute_Pointer ;
	Window_D_P = Window_P->Data_Pointer ;
    } ;
    Do_Update = False ;
    Offset = ( Position_Row - 1 ) * Columns + Position_Column - 1 ;
    /* Determine offset in window */
    C = 0 ;
    while( C < strlen( Output_Buffer ) )
    {
        Ch = Output_Buffer[ C ] ; /* Get next character */
        Ord_Ch = ( unsigned char ) Output_Buffer[ C ] ;
        if( Screen_Root.Printer_Mode == 1 ) /* Print mode active */
        {
            Ch1 = Ch ; /* Translate */
            R.w.ax = 0x0500 ; /* Printer output */
            if( Ord_Ch < 32 || Ord_Ch > 126 ) /* Special character */
	    {
                if( Ord_Ch == 1 )
                {
                    R.w.dx = 27 ;
                    intdos( &R, &R ) ;
                    R.w.dx = 'O' ;
                    intdos( &R, &R ) ;
                } ; /* if( Ord_Ch == 1 ) */
                R.w.dx = Ch1 ;
                intdos( &R, &R ) ;
	    } ; /* if( Ord_Ch < 32 || Ord_Ch > 126 ) */
        } ; /* if( Screen_Root.Printer_Mode == 1 ) */
        if( Window_P == NULL )
	{
	    printf( "%c", Ch ) ; /* No current window defined */
	}
        else
        {
            Ord_Ch1 = Ord_Ch ;
	    if( Print_Characters[ Ord_Ch ] ) /* If a printing character */
            {
                if( Ord_Ch == 9 ) /* Tab */
                {
                    Dummy = ( Position_Column + 8 ) & ( ~7 ) ;
                    for( Dummy1 = Position_Column; Dummy1 <= Dummy - 1; Dummy1++ )
		    {
                        if( Dummy1 <= Columns )
                        /* If not past right side of window */
                        {
                          Window_A_P[ Offset + Dummy1 - Position_Column ] =
                            Attribute ;
                          Window_D_P[ Offset + Dummy1 - Position_Column ] = 32 ;
                          /* Put character on window */
                        } ; /* if( Dummy1 <= Columns ) */
		    } ; /* for( Dummy1 = Position_Column; */
                    Offset = Offset + Dummy - Position_Column ;
                    Position_Column = Dummy ; /* Move to next tab position */
                } ; /* if( Ord_Ch == 9 ) */
                /* Move to next tab column.  Tab stops are every 8 columns */
		if( Position_Column > Columns ) /* Past right side of window */
		{
		    if( Wrap ) /* Wrap */
		    {
			if( Position_Row > Rows ) /* End of window */
			{
			    Position_Column = 0 ;
			}
                        else
			{
			    Position_Column = 1 ;
			    /* Go back to first column */
			} ;
			Handle_Lf( &Do_Update, &Position_Row, &Position_Column,
				&Rows, &Flags, &Low_Update, &High_Update, &Size,
				&Columns, &Offset, &Attribute, Window_A_P,
				Window_D_P ) ;
			/* Implied linefeed.  This also determines the new
			   offset. */
		    }
		    else
		    {
			Position_Column = Columns + 1 ;
			Offset = ( Position_Row - 1 ) * Columns ;
		    } ; /* if( Wrap ) */
		    /* Don't get more than one position beyond end of window */
                } ; /* if( Position_Column > Columns ) */
		if( Position_Column <= Columns && Ord_Ch != 9 )
                /* If not past right side of window and not a tab */
                {
                    Window_A_P[ Offset ] = Attribute ;
                    Window_D_P[ Offset ] = Ord_Ch1 ;
                    /* Put character on window */
                    Offset ++ ;
                    Position_Column ++ ;
                    /* Increment column count for printing characters */
                } ; /* Position_Column <= Columns && Ord_Ch != 9 ) */
            }
	    else /* Non-printing character */
            {
                if( Ord_Ch == 7 ) /* BELL */
                {
                    R.w.ax = 0x0200 ;
                    R.w.dx = 7 ;
                    intdos( &R, &R ) ;
                } ; /* if( Ord_Ch == 7 ) */
                if( Ord_Ch == 8 && /* Backspace */
                    Position_Column > 1 ) /* Not at start of line */
		{
                    Offset -- ;
                    Position_Column -- ; /* Back up one column */
		} ;
                if( Ord_Ch == 254 ) /* Special backspace */
		{
                    if( Position_Column > 1 ) /* Not at start of line */
		    {
                        Offset -- ;
                        Position_Column -- ; /* Back up one column */
		    }
		    else /* At start of line */
		    {
                        if( Position_Row > 1 ) /* Not at home position */
			{
                            Offset -- ;
                            Position_Row -- ;
                            Position_Column = Columns ;
			} ; /* if( Position_Row > 1 ) */
		    } ; /* if( Position_Column > 1 ) */
		} ; /* if( Ord_Ch == 254 ) */
                if( Ord_Ch == 13 ) /* Return */
                {
                    Position_Column = 1 ;
                    /* Move to first column of current line */
                    Offset = ( Position_Row - 1 ) * Columns ;
                } ; /* if( Ord_Ch == 13 ) */
            } ;
	    if( Cursor_Characters[ Ord_Ch ] )
            /* If a character which affects cursor position */
	    {
                /* Determine what region of the window has changed */
		if( Offset - 1 < Low_Update )
		{
		    Low_Update = Offset - 1 ;
		} ;
                if( Low_Update < 0 )
		{
		    Low_Update = 0 ;
		} ;
                if( Offset + 1 > High_Update )
		{
		    High_Update = Offset + 1 ;
		} ;
                if( High_Update > Size - 1 )
		{
		    High_Update = Size - 1 ;
		} ;
	    }
	    else
	    {
		if( Ord_Ch == 10 )
		{
		    Handle_Lf( &Do_Update, &Position_Row, &Position_Column,
			&Rows, &Flags, &Low_Update, &High_Update, &Size,
			&Columns, &Offset, &Attribute, Window_A_P,
			Window_D_P ) ;
		} ;
	    } ; /* if( Cursor_Characters[ Ord_Ch ] ) */
	} ; /* if( Window_P == NULL ) else */
        C ++ ;
    } ; /* while( C < strlen( Output_Buffer ) ) */
    if( Window_P != NULL ) /* If a window defined */
    {
        Window_P->Position_Row = Position_Row ;
        Window_P->Position_Column = Position_Column ;
        Window_P->Low_Update = Low_Update ;
        Window_P->High_Update = High_Update ;
    } ; /* if( Window_P != NULL ) */
    if( Do_Update && ( Flags & 8 ) == 0 )
    {
	Update_Window( Window_P ) ; /* Update if needed */
    } ; /* if( Do_Update && ( Flags & 8 ) == 0 ) */
} ; /* Window_Output */


void Init_Smu( int X, int Y, int Z, int A )
/* This procedure initializes the SMU root, with machine type of X.  If X is
 0, we try to determine machine type.  Y is the printer type.  Z is the number
 of input lines to save (default is 20).  A is 0 for overwrite mode, or 1 for
 insert mode */
{

int C, I, R ;
union REGS Reg ;

    Monochrome = (Character far *) MK_FP( 0xB000, 0 ) ;
    Screen_Root.Insert_Mode = False ;
    Screen_Root.Model = 0 ;
    Attribute_Default = 0 ;
    Screen_Root.Next_Window = NULL ; /* No windows defined yet */
    Screen_Root.Window_Count = 0 ;
    Screen_Root.Default = 0 ; /* No default */
    Screen_Root.Default_P = NULL ;
    Screen_Root.Printer_Type = 0 ; /* Type of printer */
    Screen_Root.Printer_Mode = 2 ; /* Print mode disabled */
    Screen_Root.Input_Buffer_Max = 20 ; /* Number of input lines to save */
    Screen_Root.Last_Input = NULL ;
    Screen_Root.Insert_Mode = False ; /* Overwrite mode */
    Screen_Root.Driver = NULL ; /* No video driver */
    if( Z == 0 )
    {
	Z = 20 ; /* Default is 20 */
    } ;
    if( A == 0 )
    {
	Screen_Root.Insert_Mode = False ;
    }
    else
    {
	Screen_Root.Insert_Mode = True ;
    } ;
    if( X != 0 ) /* A specific machine type specified */
    {
	Screen_Root.Model = X ; /* Set model to specified type */
    }
    else
    {
	Reg.w.ax = 0x0F00 ; /* Get video mode */
	int86( 0x10, &Reg, &Reg ) ; /* Do BIOS call */
	switch( Reg.h.al ) /* Video type */
	{
	    case 0 :
	    case 1 : /* 40 column text */
	    {
		Screen_Root.Model = 3 ; /* Text mode */
		Reg.w.ax = 0x0007 ; /* Set mode monochrome */
		int86( 0x10, &Reg, &Reg ) ;
		break ;
	    } ;
	    case 2 :
	    case 3 : /* Text mode CGA/EGA/VGA/SVGA */
	    {
		Screen_Root.Model = 4 ;
		break ;
	    } ;
	    case 4 :
	    case 5 :
	    case 6 :
	    case 0x0D :
	    case 0x0E :
	    case 0x0F :
	    case 0x10 :
	    case 0x11 :
	    case 0x12 :
	    case 0x13 : /* Graphics mode */
	    {
               	Screen_Root.Model = 5 ;
		C = 0 ;
		I = Reg.h.al ;
		initgraph( &C, &I, "" ) ;
		C = graphresult() ;
		if( C != grOk )
		{
		    puts( grapherrormsg( C ) ) ;
		    exit( EXIT_FAILURE ) ;
		} ;
		setviewport( 0, 0, getmaxx(), getmaxy(), True ) ;
		break ;
	    } ;
	    case 7 : /* Monochrome (MDA) */
	    {
		I = inportb( 0x3Ba ) & 128 ;
		C = 0 ;
		while( C < 4096 )
		{
		    R = inportb( 0x3Ba ) & 128 ;
		    if( R != I )
		    {
			C = 4096 ;
		    } ;
		    C = C + 1 ;
		} ;
		Screen_Root.Model = 3 ; /* Assume MDA */
		if( R != I ) /* Hercules card */
		{
		    if( ( inportb( 0x3B8 ) & 1 ) == 1 )
		    /* Hercules graphics */
		    {
			Screen_Root.Model = 6 ;
		    }
		    else
		    {
			Screen_Root.Model = 1 ; /* Unknown */
		    } ;
		} ;
	    } ;
	} ;
    } ;
    Attribute_Default = 0 ;
    Init_Screen() ; /* Set up depending upon model */
    Screen_Root.Next_Window = NULL ; /* No windows defined yet */
    Screen_Root.Window_Count = 0 ;
    Screen_Root.Default = 0 ; /* No default */
    Screen_Root.Default_P = NULL ;
    Screen_Root.Printer_Type = Y ; /* Type of printer */
    Screen_Root.Printer_Mode = 2 ; /* Print mode disabled */
    Screen_Root.Input_Buffer_Max = 20 ; /* Number of input lines to save */
    Screen_Root.Last_Input = NULL ;
    Screen_Root.Insert_Mode = False ; /* Overwrite mode */
    Screen_Root.Driver = NULL ; /* No video driver */
    memset( &Screen.Window, 0, 2106 ) ; /* No windows mapped */
    memset( &Screen.Data, 32, 2106 ) ; /* Space fill screen */
} ;

void Set_Screen( int Model, int Printer_Model, int Input_Buffers, int Mode )
/* Set current screen attributes */
{

struct Input_Record *Input_Pointer, *Input_Pointer1 ;
int Temp ;

    if( Mode >= 0 )
    {
	Screen_Root.Printer_Mode = Mode ;
    } ;
    if( Printer_Model > 0 )
    {
        Screen_Root.Printer_Type = Printer_Model ; /* Type of printer */
    } ;
    if( Input_Buffers > 0 ) /* Number of input lines to save */
    {
	Temp = Screen_Root.Input_Buffer_Max ;
	if( Temp < Input_Buffers ) /* Reduction in buffers */
	{
	    Input_Pointer = Screen_Root.Last_Input ;
	    while( Input_Pointer != NULL )
	    {
		Input_Pointer = Input_Pointer->Last_Input ;
		Temp -- ;
		if( Temp == 0 ) /* New end of buffers */
		{
		    if( Input_Pointer->Last_Input != NULL ) /* Others exist */
		    {
			Input_Pointer1 = Input_Pointer ;
			/* Save current record */
			Input_Pointer = Input_Pointer->Last_Input ;
			/* Go to last input record */
			Input_Pointer1->Last_Input = NULL ;
			/* Terminate list at new point */
			while( Input_Pointer != NULL ) /* Until done */
			{
			    free( Input_Pointer ) ;
			    /* Delete old input records */
			    Input_Pointer = Input_Pointer->Last_Input ;
			} ;
		    } ;
		} ;
	    } ;
	} ;
	Screen_Root.Input_Buffer_Max = Input_Buffers ;
    } ;
    if( Model > 0 ) /* Change model */
    {
        Screen_Root.Model = Model ;
        Init_Screen() ; /* Set up as requested */
    } ;
} ;

void Set_Window( struct Window *Window_P, int Row, int Column, int Flag,
	int Attributes )
/* Set window flags/current position */
{

int Temp ; /* Temporary variable for accumulating attributes */

    /* Attribute settings:
                         AND 1 = Reverse-video
                         AND 2 = High-intensity
                         AND 4 = Blink
                         AND 8 = Low-intensity
                         AND 16 = No blink
                         AND 32 = Normal video
                         AND 64 = Modify existing flags
                         AND 128-16384 = reserved */
    if( Window_P == NULL )
    {
	return ;
    } ;
    if( Attributes < 0 ) /* If attributes to change */
    {
	Attributes = -Attributes ; /* Recover bits */
	switch( Screen_Root.Model ) /* Default settings */
	{
	    case 1 :
	    {
		Temp = 0 ;
		break ;
	    } ;
	    case 2 :
	    {
		Temp = 14 ;
		break ;
	    } ;
	    case 3 :
	    case 4 :
	    {
		Temp = 7 ;
	    } ;
	} ;
	if( ( Attributes & 64 ) == 64 )
	{
	    Temp = Window_P->Attribute ; /* Get current attributes */
	} ;
	if( Screen_Root.Model == 1 )
	{
	    Temp = Attributes & 14 ; /* Unknown */
	} ;
	if( ( Attributes & 1 ) == 1 ) /* Reverse-video */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp | 1 ; /* Reverse to normal */
		    break ;
		} ;
		case 3 :
		case 4 :
		{
		    Temp = Temp & ( ~7 ) | 0x70 ; /* Normal to reverse */
		} ;
	    } ;
	} ;
	if( ( Attributes & 2 ) == 2 ) /* High-intensity */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp & ( ~2 ) ;
		    break ;
		} ;
		case 3 :
		case 4 :
		{
		    Temp = Temp | 8 ;
		} ;
	    } ;
	} ;
	if( ( Attributes & 4 ) == 4 ) /* Blink */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp & ( ~4 ) ;
		    break ;
		} ;
		case 3 :
		{
		    Temp = Temp | 128 ;
		    break ;
		} ;
		case 4 :
		{
		    Temp = Temp | 64 ;
		} ;
	    } ;
	} ;
	if( ( Attributes & 8 ) == 8 )/* Low-intensity */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp | 2 ;
		    break ;
		} ;
		case 3 :
		case 4 :
		{
		    Temp = Temp & ( ~8 ) ;
		} ;
	    } ;
	} ;
	if( ( Attributes & 16 ) == 16 ) /* No blink */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp | 4 ;
		    break ;
		} ;
		case 3 :
		{
		    Temp = Temp & ( ~128 ) ;
		    break ;
		} ;
		case 4 :
		{
		    Temp = Temp & ( ~64 ) ;
		} ;
	    } ;
	} ;
	if( ( Attributes & 32 ) == 32 ) /* Normal video */
	{
	    switch( Screen_Root.Model )
	    {
		case 2 :
		{
		    Temp = Temp & ( ~1 ) ; /* Reverse to normal */
		    break ;
		} ;
		case 3 :
		case 4 :
		{
		    Temp = Temp & ( ~0x70 ) | 7 ;
		} ;
	    } ;
	} ;
	Window_P->Attribute = Temp ; /* Now set attributes */
    } ;
    /* Set attributes if flag (negative) */
    if( Row < 0 || Column < 0 || Flag != 0 )
    {
	Window_P->Flags = Flag ;
    } ;
    /* Flags are always set if row or column are negative */
    if( Row > 0 )
    {
	Window_P->Position_Row = Row ;
    } ;
    if( Column > 0 )
    {
	Window_P->Position_Column = Column ;
    } ;
} ;

void Select_Window( struct Window *Window_P )
{
    Screen_Root.Default = Window_P->Id ;
    Screen_Root.Default_P = Window_P ;
} ;

void Erase_Eow( struct Window *Window_P ) /* Erase to end of window */
{

int A, B, Offset ;

    Offset = ( Window_P->Position_Row - 1 ) * Window_P->Columns +
        Window_P->Position_Column - 1 ; /* Get current offset */
    if( Window_P->Low_Update > Offset )
    {
	Window_P->Low_Update = Offset ;
    } ;
    Window_P->High_Update = Window_P->Size - 1 ;
    B = Window_P->Size - Offset ; /* Find size of end of window */
    memset( &Window_P->Data_Pointer[ Offset ], 32, B ) ;
    memset( &Window_P->Attribute_Pointer[ Offset ], Window_P->Attribute, B ) ;
    /* Spaces and null attributes */
    if( ( Window_P->Flags & 8 ) == 0 ) /* If auto-update mode */
    {
        Update_Window( Screen_Root.Default_P ) ;
    } ;
} ;

void Erase_Eol( struct Window *Window_P )
{

int A, B, Offset ;

    Offset = ( Window_P->Position_Row - 1 ) * Window_P->Columns +
        Window_P->Position_Column - 1 ; /* Get current offset */
    B = Window_P->Columns - Window_P->Position_Column ; /* Find size of end of line */
    if( Window_P->Low_Update > Offset )
    {
	Window_P->Low_Update = Offset ;
    } ;
    if( Window_P->High_Update < Offset + B - 1 )
    {
	Window_P->High_Update = Offset + B - 1 ;
    } ;
    memset( &Window_P->Data_Pointer[ Offset ], 32, B ) ;
    memset( &Window_P->Attribute_Pointer[ Offset ], Window_P->Attribute, B ) ;
    /* Spaces and null attributes */
    if( ( Window_P->Flags & 8 ) == 0 ) /* If auto-update mode */
    {
        Update_Window( Screen_Root.Default_P ) ;
    } ;
} ;

void Insert_Line( struct Window *Window_P, short int Direction )
/* Insert line at current position.  Direction of 0=lose top line,
   otherwise=lose bottom line */
{

int A, B, C, Offset ;

    Window_P->Position_Column = 1 ; /* Reset cursor to BOL */
    Offset = ( Window_P->Position_Row - 1 ) * Window_P->Columns ;
    /* Get beginning of current line's offset */
    if( Direction == 0 ) /* Lose top line */
    {
	if( Offset == 0 )
	/* Insert line at line 1 does nothing if losing top line */
	{
	    return ;
	} ;
        Window_P->Low_Update = 0 ;
        if( Window_P->High_Update < Offset + Window_P->Columns - 1 )
	{
	    Window_P->High_Update = Offset + Window_P->Columns - 1 ;
	} ;
	Offset = Offset - Window_P->Columns ; /* Blank previous line */
        B = Offset - 1 ; /* Number of bytes to move */
        memmove( Window_P->Data_Pointer,
	    &Window_P->Data_Pointer[ Window_P->Columns ], B ) ;
        memmove( Window_P->Attribute_Pointer,
	    &Window_P->Attribute_Pointer[ Window_P->Columns ], B ) ;
        /* Scroll up */
    }
    else /* Lose bottom line */
    {
        if( Window_P->Low_Update >= Offset )
	{
	    Window_P->Low_Update = Offset ;
	} ;
        Window_P->High_Update = Window_P->Size - 1 ;
        B = Window_P->Size - Offset - Window_P->Columns ;
        /* Number of bytes to move */
        memmove( &Window_P->Data_Pointer[ Offset + Window_P->Columns ],
	    &Window_P->Data_Pointer[ Offset ], B ) ;
        memmove( &Window_P->Attribute_Pointer[ Offset + Window_P->Columns ],
	    &Window_P->Attribute_Pointer[ Offset ], B ) ;
        /* Scroll down */
    } ;
    memset( &Window_P->Data_Pointer[ Offset ], 32, Window_P->Columns ) ;
    memset( &Window_P->Attribute_Pointer[ Offset ],
        Window_P->Attribute, Window_P->Columns ) ;
    /* Spaces and null attributes for newly created line */
    if( ( Window_P->Flags & 8 ) == 0 ) /* If auto-update mode */
    {
        Update_Window( Screen_Root.Default_P ) ;
    } ;
} ;

void Delete_Line( struct Window *Window_P, short int Direction )
/* Delete line at current position.  Direction: 0=gain top line, otherwise=gain
 bottom line */
{

int A, B, C, Offset ;

    Window_P->Position_Column = 1 ; /* Reset cursor to BOL */
    A = Window_P->Columns ; /* Number of columns in row */
    Offset = ( Window_P->Position_Row - 1 ) * A ;
    /* Get beginning of current line's offset */

    if( Direction == 0 ) /* Gain top line */
    {
        Window_P->Low_Update = 0 ;
        if( Window_P->High_Update < Offset )
	{
	    Window_P->High_Update = Offset ;
	} ;

        memmove( &Window_P->Data_Pointer[ A ], Window_P->Data_Pointer,
	    Offset ) ;
        memmove( &Window_P->Attribute_Pointer[ A ],
	    Window_P->Attribute_Pointer, Offset ) ;
        /* Scroll down */

        memset( Window_P->Data_Pointer, 32, A ) ;
        memset( Window_P->Attribute_Pointer, Window_P->Attribute, A ) ;
        /* Erase first line */
    }
    else /* Gain bottom line */
    {
        if( Window_P->Low_Update > Offset + A - 1 )
	{
            Window_P->Low_Update = Offset + A - 1 ;
	    /* Adjust low update pointer */
	} ;
        Window_P->High_Update = Window_P->Size - 1 ;
	/* Adjust high update pointer */
        B = Window_P->Size - Offset - A ; /* Amount of data to scroll */

        memmove( &Window_P->Data_Pointer[ Offset ],
	    &Window_P->Data_Pointer[ Offset + A ], B ) ; /* Move data */
        memmove( &Window_P->Attribute_Pointer[ Offset ],
	    &Window_P->Attribute_Pointer[ Offset + A ], B ) ;
	    /* Move attributes */

        /* Scroll up */

        B = Window_P->Size - A ; /* Beginning of last line */
        memset( &Window_P->Data_Pointer[ B ], 32, A ) ;
        memset( &Window_P->Attribute_Pointer[ B ], Window_P->Attribute, A ) ;
        /* Clear last line */
    } ;
    if( ( Window_P->Flags & 8 ) == 0 ) /* If auto-update mode */
    {
        Update_Window( Screen_Root.Default_P ) ;
    } ;
} ;

static void Insert_Input( char *X )
/* Insert string as most recent input record */
{

int Count ;
struct Input_Record *Input_Pointer, *Input_Pointer1 ;

    if( Screen_Root.Input_Buffer_Max == 0 )
    {
	return ; /* No input saves */
    } ;
    Count = 1 ;
    Input_Pointer = malloc( sizeof( *Input_Pointer ) ) ;
    strcpy( Input_Pointer->Input_Line, X ) ;
    Input_Pointer->Last_Input = Screen_Root.Last_Input ;
    Input_Pointer->Next_Input = NULL ;
    /* Set up new record */
    Screen_Root.Last_Input = Input_Pointer ;
    /* Link it as the most recent last input */
    if( Input_Pointer->Last_Input != NULL )
    {
        Input_Pointer->Last_Input->Next_Input = Input_Pointer ;
    } ;
    /* Set back pointer for previous last input, if any */
    while( Input_Pointer->Last_Input != NULL ) /* Until end of list */
    {
        Input_Pointer = Input_Pointer->Last_Input ; /* backup one record */
        Count = Count + 1 ; /* Count them up */
    } ;
    while( Count > Screen_Root.Input_Buffer_Max ) /* Get rid of extra input */
    {
	Input_Pointer1 = Input_Pointer ; /* Save new oldest record */
	Input_Pointer = Input_Pointer->Next_Input ; /* Get next record */
	Input_Pointer->Last_Input = NULL ;
	Count = Count - 1 ; /* One less input record */
	free( Input_Pointer1 ) ; /* Bye bye */
    } ;
} ;


short int Get_Window_Attribute( int R, int C )
/* Return attribute of character in current window at row R, column C.  Use
 default positions if R, C, or both are zero (0) */
{

struct Window *Window_P ;
int Offset ;

    Window_P = Screen_Root.Default_P ; /* Get current window */
    if( R == 0 )
    {
	R = Window_P->Position_Row ;
    } ;
    if( C == 0 )
    {
	C = Window_P->Position_Column ;
    } ;
    if( R > Window_P->Rows ) /* Out of range */
    {
	return 0 ;
    } ;
    if( C > Window_P->Columns ) /* Ditto */
    {
	return 0 ;
    } ;
    if( R < 1 || C < 1 ) /* Ditto ditto */
    {
	return 0 ;
    } ;
    Offset = ( R - 1 ) * Window_P->Columns + C - 1 ;
    return Window_P->Attribute_Pointer[ Offset ] ;
} ;

void Set_Cursor( int Vr, int Vc ) /* Place visible cursor on screen */
{

int Offset ;
struct Window *Window_P ;
union REGS Reg ;

    Window_P = Screen_Root.Default_P ; /* Get default window */
    if( ( ( Screen_Root.Flags | Window_P->Flags ) & 32 ) == 32 )
    {
	return ;
    } ;
    /* Nope if cursor disabled */
    if( Vr == 0 )
    {
	Vr = Window_P->Position_Row ;
    } ;
    if( Vc == 0 )
    {
	Vc = Window_P->Position_Column ;
    } ;
    if( Screen_Root.Video_Mode != 7 )
    {
	Set_Curs( Vr, Vc - 1 ) ;
    }
    else
    {
	switch( Screen_Root.Model )
	{
	    case 1 : /* Unknown */
	    {
		printf( "\x27[%i;%iH", Vr, Vc - 1 ) ;
	    } ;
	    case 3 :
	    case 4 :
	    case 5 : /* IBM PC */
	    {
		Ibm_Pos( Vr - 1, Vc - 2 ) ;
	    } ;
	} ;
    } ;
} ;

void Clear_Cursor( int Vr, int Vc ) /* Blank cursor */
{

struct Window *Window_P ;
union REGS Reg ;
int Z ;

    Window_P = Screen_Root.Default_P ; /* Get default window */
    if( ( ( Screen_Root.Flags | Window_P->Flags ) & 32 ) == 32 )
    {
	return ;
    } ;
    /* Nope if cursor disabled */
    if( Screen_Root.Video_Mode != 7 ) /* Non-text */
    {
	Reg.w.ax = 0x0800 ; /* Read character at current cursor position */
	int86( 0x10, &Reg, &Reg ) ; /* Get info */
	Z = Reg.w.ax ; /* Get character */
	Reg.w.ax = 0x0900 | ( Z & 255 ) ; /* Character to write */
	Reg.w.cx = 1 ; /* Multiplier */
	Reg.w.bx = ( Screen_Root.Background << 4 ) + Screen_Root.Foreground ;
	int86( 0x10, &Reg, &Reg ) ; /* Reverse it now */
    }
    else /* Text mode */
    {
	Reg.w.ax = 0x0200 ;
	Reg.w.bx = 0 ;
	Reg.w.dx = 80 ;
	int86( 0x10, &Reg, &Reg ) ; /* Make cursor vanish */
    } ;
} ;

char Get_Screen( int Row, int Col )
{

union REGS Reg ;

    if( Screen_Root.Model == 1 ) /* ANSI.SYS stuff */
    {
        return Screen.Data[ Row ][ Col ] ; /* Return from screen image */
    } ;
    if( Screen_Root.Model == 3 ) /* IBM MDA */
    {
        return Monochrome[ Row_Start[ Row ] + Col - 1 ][ 0 ] ;
    } ;
    if( Screen_Root.Model != 4 ) /* Not IBM CGA/EGA/VGA/SVGA text */
    {
        return 0 ;
    } ;
    Reg.w.ax = 0x0800 ;
    Reg.w.bx = 0 ;
    int86( 0x10, &Reg, &Reg ) ; /* Read character */
    return Reg.h.al ;
} ;

static char *Line_Format( char *Kb_Buffer, int Cursor, int *Cursor1 )
{

static char A[ 256 ] ;
short int B ;

    A[ 0 ] = 0 ;
    for( B = 0; B < strlen( Kb_Buffer ); B++ )
    /* Build line character by character */
    {
        if( Kb_Buffer[ B ] < ' ' && Kb_Buffer[ B ] != '\t' )
	{
	    strcat( A, "^" ) ;
	    A[ strlen( A ) + 1 ] = 0 ;
	    A[ strlen( A ) ] = Kb_Buffer[ B ] + 64 ; /* Control character */
	}
	else
	{
	    A[ strlen( A ) + 1 ] = 0 ;
	    A[ strlen( A ) ] = Kb_Buffer[ B ] ;
	} ;
        if( B == Cursor - 1 ) /* At current cursor position */
	{
            if( Kb_Buffer[ B ] < ' ' && Kb_Buffer[ B ] != '\t' )
	    {
		*Cursor1 = strlen( A ) - 1 ;
	    }
	    else
	    {
		*Cursor1 = strlen( A ) ;
	    } ;
	} ;
    } ;
    return &A ; /* Return value */
} ;

static void Erase_Line( struct Window *Window_P, char *Command )
{
    if( Window_P == NULL )
    {
	puts( "" ) ;
    }
    else
    {
	Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
	for( Dummy = 1; Dummy <= strlen( Command ); Dummy++ )
	{
	    Window_Output( " " ) ;
	} ;
	Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
	Update_Window( Screen_Root.Default_P ) ;
    } ;
    Kb_Buffer[ 0 ] = 0 ;
} ;

void Window_Input_Line( struct Window *Window_P, int Flags )
/* Get a line from the keyboard */
{

char B[ 256 ], Command[ 256 ], Command0[ 256 ] ;
char Backup ; /* Backup character to use */
char Ch1 ;
short int Cursor, Cursor1 ;
short int Ord_B, Vc, Vr ;
/* Beginning column & row, ASCII value of current characters, etc. */
int Offset, Offset1 ;
int Old_Length ;
union REGS R ;
struct Input_Record *Input_Pointer ;
short int Wid ;
char Temp[ 256 ] ;

/* Window_Input_Line */
    Input_Pointer = NULL ;
    Command[ 0 ] = 0 ;
    Cursor = 1 ;
    Cursor1 = 1 ;
    if( Window_P != NULL )
    {
	Backup = 254 ;
        Input_Column = Window_P->Position_Column ;
        Input_Row = Window_P->Position_Row ;
	Wid = Window_P->Id ;
    }
    else
    {
	Backup = 8 ;
        Input_Column = 1 ;
        Input_Row = 1 ;
    } ;

Loop:
    if( Window_P != NULL )
    {
	Offset = Window_P->Position_Column ;
	Offset1 = Window_P->Position_Row ;
	if( ( Window_P->Flags & 4 ) == 0 ) /* Wrap */
	{
	    if( Offset > Window_P->Columns ) /* Past right of window */
	    {
		Offset = 1 ;
		Offset1 = Offset1 + 1 ;
	    } ;
	} ;
	if( ( Window_P->Flags & 2 ) == 0 && /* Window isn't occluded */
            Offset1 >= Window_P->View_Row &&
	    ( Offset1 <= Window_P->View_Row + Window_P->Map_Rows - 1 ) &&
	    /* Within viewport rows */
	    Offset >= Window_P->View_Column &&
	    ( Offset <= Window_P->View_Column + Window_P->Map_Columns - 1 ) )
	    /* Within viewport columns */
	{
	    Vr = Offset1 - Window_P->View_Row + Window_P->Map_Row ;
	    Vc = Offset - Window_P->View_Column + Window_P->Map_Column + 1 ;
	    if( Screen.Window[ Vr ][ Vc - 1 ] == Wid )
	    {
		Set_Cursor( Vr, Vc ) ; /* Place cursor */
	    } ;
	} ;
    } ; /* if( Window_P != NULL ) */
    strcpy( B, InKey() ) ; /* Get next key */

    if( strcmp( B, "LEFT" ) == 0 ) /* Left-arrow */
    {
        if( Cursor > 1 ) /* Not at beginning of line */
        {
            Cursor = Cursor - 1 ;
            Cursor1 = Cursor1 - 1 ;
            sprintf( Temp, "%c", Backup ) ; /* Backup */
	    Window_Output( Temp ) ;
            if( Kb_Buffer[ Cursor - 1 ] < ' ' && Kb_Buffer[ Cursor - 1 ] != '\t' )
            /* Control */
            {
                sprintf( Temp, "%c", Backup ) ; /* Backup over caret */
		Window_Output( Temp ) ;
                Cursor1 = Cursor1 - 1 ;
	    } ;
	    if( Window_P != NULL )
	    {
		if( Kb_Buffer[ Cursor - 1 ] == '\t' ) /* Tab */
		{
                    Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
                    sprintf( Temp, "%s", Left( Command, Cursor1 - 1 ) ) ;
		    Window_Output( Temp ) ;
                    Offset = Window_P->Position_Row ;
                    Offset1 = Window_P->Position_Column ;
		    sprintf( Temp, "%s", &Command[ Cursor1 - 1 ] ) ;
		    Window_Output( Temp ) ;
                    Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
                    /* Backup */
		} ;
		Update_Window( Screen_Root.Default_P ) ;
	    } ; /* if( Window_P != NULL ) */
        } ;
        goto Loop ;
    } ;
    if( strcmp( B, "RIGHT" ) == 0 ) /* Right-arrow */
    {
        if( Cursor <= strlen( Kb_Buffer ) ) /* Before end of line */
        {
            if( Kb_Buffer[ Cursor - 1 ] < ' ' &&
		Kb_Buffer[ Cursor - 1 ] != '\t' )
		/* Control character */
            {
                sprintf( Temp, "^%c", Kb_Buffer[ Cursor - 1 ] + 64 ) ;
		Window_Output( Temp ) ;
                Cursor1 = Cursor1 + 1 ;
            }
	    else
	    {
		sprintf( Temp, "%c", Kb_Buffer[ Cursor - 1 ] ) ;
		Window_Output( Temp ) ;
	    } ;
            Cursor = Cursor + 1 ;
            Cursor1 = Cursor1 + 1 ;
	    Update_Window( Screen_Root.Default_P ) ;
        } ;
        goto Loop ;
    } ;

    Ord_B = B[ 0 ] ;

    if( strcmp( B, "UP" ) == 0 || Ord_B == 2 ) /* Up-arrow or ^B */
    {
        if( Input_Pointer == NULL )
	{
	    Input_Pointer = Screen_Root.Last_Input ;
	}
	else
        {
            if( Input_Pointer->Last_Input == NULL )
	    {
		goto Loop ;
	    } ;
            Input_Pointer = Input_Pointer->Last_Input ;
        } ;
        if( Input_Pointer != NULL ) /* Some input */
        {
            Erase_Line( Window_P, Command ) ;
	    Command[ 0 ] = 0 ;
	    Cursor = 1 ;
	    Cursor1 = 1 ;
            strcpy( Kb_Buffer, Input_Pointer->Input_Line ) ;
            Cursor = strlen( Kb_Buffer ) + 1 ;
            strcpy( Command, Line_Format( Kb_Buffer, Cursor, &Cursor1 ) ) ;
            Cursor1 = strlen( Command ) + 1 ;
            sprintf( Temp, "%s", Command ) ;
	    Window_Output( Temp ) ;
            Update_Window( Screen_Root.Default_P ) ;
	} ;
	goto Loop ;
    } ;
    if( strcmp( B, "DOWN" ) == 0 ) /* Down-arrow */
    {
        if( Input_Pointer == NULL )
	{
	    goto Loop ;
	} ;
        Input_Pointer = Input_Pointer->Next_Input ;
        Erase_Line( Window_P, Command ) ;
	Command[ 0 ] = 0 ;
	Cursor = 1 ;
	Cursor1 = 1 ;
        if( Input_Pointer != NULL ) /* Some input */
        {
            strcpy( Kb_Buffer, Input_Pointer->Input_Line ) ;
            Cursor = strlen( Kb_Buffer ) + 1 ;
            strcpy( Command, Line_Format( Kb_Buffer, Cursor, &Cursor1 ) ) ;
            Cursor1 = strlen( Command ) + 1 ;
            sprintf( Temp, "%s", Command ) ;
	    Window_Output( Temp ) ;
            Update_Window( Screen_Root.Default_P ) ;
	} ;
        goto Loop ;
    } ;
    if( strcmp( B, "F4" ) == 0 ) /* Toggle print mode */
    {
        if( Screen_Root.Printer_Mode == 0 )
	{
	    Screen_Root.Printer_Mode = 1 ;
	}
	else
	{
	    if( Screen_Root.Printer_Mode == 1 )
	    {
		Screen_Root.Printer_Mode = 0 ;
	    } ;
	} ;
	/* Toggle between on and off */
	goto Loop ;
    } ;
    if( strcmp( B, "F6" ) == 0 ) /* Print screen */
    {
        if( Screen_Root.Printer_Mode != 2 ) /* Not disabled */
	{
            for( Offset = 1; Offset <= 24; Offset++ ) /* For each row */
            {
                for( Offset1 = 1; Offset1 <= Screen_Root.Screen_Width; Offset1++ ) /* For each column */
                {
                    Ord_B = Get_Screen( Offset, Offset1 ) ;
                    /* Get physical character */
                    Ch1 = Ord_B ; /* Translate */
                    R.w.ax = 0x0500 ; /* Printer output */
                    if( Ord_B < 32 || Ord_B > 126 && /* Special character */
		        Ord_B == 1 )
		    {
			R.w.dx = 27 ;
			intdos( &R, &R ) ;
			R.w.dx = 'O' ;
			intdos( &R, &R ) ;
		    } ;
                    R.w.dx = Ch1 ;
                    intdos( &R, &R ) ;
                } ;
                fputs( "", Printer_File ) ; /* End line */
            } ;
        } ;
	goto Loop ;
    } ;
    if( strcmp( B, "ENTER" ) == 0 )
    {
        B[ 0 ] = 13 ;
	B[ 1 ] = 0 ;
        Ord_B = 13 ;
    } ;
    if( strlen( B ) > 1 ) /* Special key */
    {
        printf( "\a" ) ;
        goto Loop ;
    } ;
    if( Ord_B == 1 ) /* Control-A */
    {
        Screen_Root.Insert_Mode = !Screen_Root.Insert_Mode ; /* Toggle mode */
        goto Loop ;
    } ;
    if( Ord_B == 5 ) /* ^E - goto end of line */
    {
        if( Cursor1 <= strlen( Command ) )
	{
	    sprintf( Temp, "%s", &Command[ Cursor1 ] ) ;
	    Window_Output( Temp ) ;
	} ;
        Cursor = strlen( Kb_Buffer ) + 1 ;
        Cursor1 = strlen( Command ) + 1 ;
        Update_Window( Screen_Root.Default_P ) ;
        goto Loop ;
    } ;
    if( Ord_B == 18 ) /* Control-R */
    {
        Refresh_Window( Screen_Root.Default_P ) ;
        goto Loop ;
    } ;
    if( Ord_B == 23 ) /* Control-W */
    {
        Refresh_Screen() ;
        goto Loop ;
    } ;
    if( Ord_B == 127 || Ord_B == 8 ) /* Handle delete */
    {
        if( strlen( Kb_Buffer ) > 0 ) /* Something to delete */
        {
            Ord_B = ( unsigned char ) Kb_Buffer[ Cursor - 2 ] ;
	    /* Get character to be deleted */
            strcpy( &Kb_Buffer[ Cursor - 2 ], &Kb_Buffer[ Cursor - 1 ] ) ;
            /* Remove character */

            Old_Length = strlen( Command ) ;
            strcpy( Command, Line_Format( Kb_Buffer, Cursor, &Cursor1 ) ) ;
            if( Ord_B == 9 ) /* Tab */
            {
                /* Move to beginning of tab */
		if( Window_P != NULL )
		{
		    Set_Window( Window_P, Input_Row, Input_Column, 0, 0 ) ;
		    /* Input start */
		    sprintf( Temp, "%s", Left( Command, Cursor1 - 2 ) ) ;
		    Window_Output( Temp ) ;
		    /* Write up to tab */
		    Offset = Window_P->Position_Row ;
		    Offset1 = Window_P->Position_Column ;
		    /* Save cursor position */
		}
		else /* DOS output */
		{
		    printf( "%c", Backup ) ;
		} ; /* if( Window_P != NULL ) */
                sprintf( Temp, "%s", &Command[ Cursor1 - 2 ] ) ;
		Window_Output( Temp ) ; /* Write rest of line */
                Window_Output( "        " ) ;
		if( Window_P == NULL )
		{
		    Dummy = strlen( Command ) - Cursor1 + 8 ;
		    while( Dummy > 0 )
		    {
			printf( "%c", Backup ) ;
			Dummy -- ;
		    } ;
		}
		else
		{
		    Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
		    /* Move back to cursor position */
		} ;
            }
	    else /* Non-tab */
            {
		sprintf( Temp, "%c %c", Backup, Backup ) ;
		Window_Output( Temp ) ; /* Erase old character */
                if( Ord_B < 32 ) /* Deleting a control character */
                {
                    sprintf( Temp, "%c %c", Backup, Backup ) ;
		    Window_Output( Temp ) ;
                    Cursor1 = Cursor1 - 1 ;
                } ;
            } ;
	    if( Window_P != NULL )
	    {
		Offset = Window_P->Position_Row ;
		Offset1 = Window_P->Position_Column ;
	    } ;
            Cursor = Cursor - 1 ;
            Cursor1 = Cursor1 - 1 ;
	    sprintf( Temp, "%s", &Command[ Cursor1 - 1 ] ) ;
	    Window_Output( Temp ) ;
	    Dummy = Old_Length - strlen( Command ) + strlen( Command ) -
		Cursor1 + 1 ;

            while( Old_Length > strlen( Command ) )
            {
                Old_Length = Old_Length - 1 ;
                Window_Output( " " ) ;
            } ;
            /* Update line */
	    if( Window_P == NULL ) /* Normal DOS output */
	    {
		while( Dummy > 0 )
		{
		    printf( "%c", Backup ) ;
		    Dummy -- ;
		} ;
	    }
	    else
	    {
		Set_Window( Window_P, Offset, Offset1, 0, 0 ) ; /* Set cursor */
	    } ;
            Update_Window( Screen_Root.Default_P ) ;
        } ;
	goto Loop ;
    } ;
    if( Ord_B == 3 ) /* Control-C */
    {
	Window_Output( "^C\r\n" ) ;
        Kb_Buffer[ 0 ] = 3 ;
	Kb_Buffer[ 1 ] = 0 ;
        return ;
    } ;
    if( Ord_B == 26 ) /* Control-Z */
    {
	Window_Output( "^Z\r\n" ) ;
        Kb_Buffer[ 0 ] = 26 ;
	Kb_Buffer[ 1 ] = 0 ;
        return ;
    } ;
    if( Ord_B == 25 ) /* Control-Y */
    {
	Window_Output( "^Y\r\n" ) ;
        Kb_Buffer[ 0 ] = 25 ;
	Kb_Buffer[ 1 ] = 0 ;
        return ;
    } ;
    if( Ord_B == 21 ) /* ^U - Erase line */
    {
        Erase_Line( Window_P, Command ) ;
	Command[ 0 ] = 0 ;
	Cursor = 1 ;
	Cursor1 = 1 ;
        goto Loop ;
    } ;
    if( strlen( Command ) == 253 && Ord_B != 13 && Ord_B != 10 &&
        Ord_B != 12 && Ord_B != 11 && Ord_B != 27 )
    {
        printf( "\a" ) ; /* Bell upon full input line */
        goto Loop ;
    } ;
    if( ( Ord_B < 9 || Ord_B > 13 ) && Ord_B < 32 )
    {
        Command0[ 0 ] = '^' ;
	Command0[ 1 ] = Ord_B + 64 ;
	Command0[ 2 ] = 0 ;
    }
    else
    {
	Command0[ 0 ] = B[ 0 ] ;
	Command0[ 1 ] = 0 ;
    } ;
    if( ( Ord_B < 10 || Ord_B > 13 ) && Ord_B != 27 ) /* Not a delimiter */
    {
        if( Screen_Root.Insert_Mode ) /* In insert mode */
        {
	    strcpy( Temp, Kb_Buffer ) ;
	    Temp[ Cursor ] = 0 ;
	    strcat( Temp, B ) ;
            strcat( Temp, &Kb_Buffer[ Cursor ] ) ;
	    strcpy( Kb_Buffer, Temp ) ;
	    strcpy( Temp, Command ) ;
	    Command[ Cursor1 ] = 0 ;
	    strcat( Temp, Command0 ) ;
	    strcat( Temp, &Command[ Cursor1 ] ) ;
	    strcpy( Command, Temp ) ;
        }
	else /* In over-write mode */
        {
            Old_Length = ( unsigned char ) Kb_Buffer[ Cursor - 1 ] ;

            Kb_Buffer[ Cursor - 1 ] = B[ 0 ] ;
            if( strlen( Kb_Buffer ) < Cursor )
	    {
		Kb_Buffer[ Cursor - 1 ] = 0 ;
	    } ;
            Offset = Cursor1 ;
	    strcpy( Temp, Command ) ;
	    Temp[ Cursor1 - 1 ] = 0 ;
	    strcat( Temp, Line_Format( &Kb_Buffer[ Cursor - 1 ], Cursor,
		&Cursor1 ) ) ;
	    strcpy( Command, Temp ) ;
            Cursor1 = Offset ; /* Restore position */
        } ;
	sprintf( Temp, "%s", Command0 ) ; /* Write character */
	Window_Output( Temp ) ;
	if( Window_P != NULL )
	{
	    Offset = Window_P->Position_Row ;
	    Offset1 = Window_P->Position_Column ;
	} ;
	sprintf( Temp, "%s", &Command[ Cursor1 + strlen( Command0 ) - 1 ] ) ;
	Window_Output( Temp ) ;
	/* Update line */

        if( Old_Length == 9 ) /* Deleted a tab */
	{
	    Window_Output( "       " ) ;
	    if( Window_P == NULL )
	    {
		for( Dummy = 1; Dummy <= 8; Dummy++ )
		{
		    printf( "%c", Backup ) ;
		} ;
	    } ;
	} ;
	if( Window_P == NULL )
	{
	    Dummy = strlen( Command ) - Cursor1 - strlen( Command0 ) + 1 ;
	    while( Dummy > 0 )
	    {
		printf( "%c", Backup ) ;
		Dummy -- ;
	    } ;
	} ;
        Cursor = Cursor + 1 ; /* Position in buffer */
        Cursor1 = Cursor1 + strlen( Command0 ) ;
        Set_Window( Window_P, Offset, Offset1, 0, 0 ) ;
        Update_Window( Screen_Root.Default_P ) ;
        goto Loop ;
    } ;

    /* A delimeter if we get here */
    if( ( Window_P == NULL || ( Window_P->Flags & 128 ) == 0 ) &&
	( Flags & 128 ) == 0 )
    /* CRLF allowed */
    {
	Window_Output( "\r" ) ;
	if( Ord_B != 10 )
	{
	    Window_Output( "\n" ) ; /* Make sure delimiter causes CRLF */
	} ;
    } ;
} ;

void Window_Read_Input( int Flags ) /* Read a line */
{

char B[ 256 ] ;
short int Ord_B ;
struct Window *Window_P ;

    Kb_Buffer[ 0 ] = 0 ;
    Line_Count = 0 ; /* Reset line count on input */
    Window_P = Screen_Root.Default_P ; /* Get default window */
    Window_Input_Line( Window_P, Flags ) ;
    if( strlen( Kb_Buffer ) > 0 )
    {
	Insert_Input( Kb_Buffer ) ; /* Save this line */
    } ;
} ;

void Window_Read() /* Read a line */
{
    Window_Read_Input( 0 ) ; /* Do standard input */
} ;

void Load_Driver( char *X )
/* Load a screen driver from the file specified in X */
{

struct FILE *F ;
int hF ;
int G ;
char B[ 256 ] ;

    hF = open( X, O_RDONLY | O_BINARY ) ; /* Open driver file */
    if( hF == -1 )
    {
	return ;
    } ;
    F = fdopen( hF, "rb" ) ;
    if( F == NULL )
    {
	return ;
    } ;
    G = filelength( hF ) ; /* Get file size in bytes */
    if( G == 0 )
    {
	return ; /* File is empty */
    } ;
    Screen_Root.Driver = malloc( G ) ; /* Create room for driver */
    fread( Screen_Root.Driver, G, 1, F ) ; /* Load driver */
    fclose( F ) ;
    close( hF ) ;
} ;

char Get_Window_Char( int R, int C )
/* Return character in window at row R, column C.  Use default positions if R, C,
 or both are zero (0) */
{

struct Window *Window_P ;
int Offset ;

    Window_P = Screen_Root.Default_P ; /* Get current window */
    if( R == 0 )
    {
	R = Window_P->Position_Row ;
    } ;
    if( C == 0 )
    {
	C = Window_P->Position_Column ;
    } ;
    if( R > Window_P->Rows )
    {
	return 0 ; /* Out of range */
    } ;
    if( C > Window_P->Columns )
    {
	return 0 ; /* Ditto */
    } ;
    if( R < 1 || C < 1 )
    {
	return 0 ; /* Ditto ditto */
    } ;
    Offset = ( R - 1 ) * Window_P->Columns + C - 1 ;
    return Window_P->Data_Pointer[ Offset ] ;
} ;

void Set_Viewport( short int Row, short int Column )
/* Set viewport row and column */
{

struct Window *Window_P ;

    Window_P = Screen_Root.Default_P ;
    if( Row > 0 )
    {
	Window_P->View_Row = Row ;
    } ;
    if( Column > 0 )
    {
	Window_P->View_Column = Column ;
    } ;
    if( Row > 0 || Column > 0 )
    {
	Window_P->Low_Update = 0 ;
	Window_P->High_Update = 32767 ;
    } ;
    if( ( Window_P->Flags & 8 ) == 0 ) /* If auto-update mode */
    {
        Update_Window( Screen_Root.Default_P ) ;
    } ;
} ;
