{
        Program Name : Test
        Package Name : LB
        Purpose      : Routines to aid in testing.
        Institution  : Conroy & Conroy Co.
        Date Written : 24-May-2000
        Written By   : Alan Conroy
        Version      : 1
        Revision     : 0
        Edit         :

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains routines which aid in testing other code.
}

unit Test ;

interface

function Allocations : integer ;
function Allocation_Count : integer ;
procedure Begin_Memory_Test ;
procedure Clear_Memory_Statistics ;
function Deallocations : integer ;
procedure End_Memory_Test ;
function Reallocations : integer ;

implementation

// Delphi memory test routines...

// Data...
var _Allocations, _Deallocations, _Reallocations : integer ;
    _Allocation_Count : integer ;
    In_New_MM : boolean ;
    Old_Memory_Manager : TMemoryManager ;


// Internal utility routines...

function Test_GetMem( Size : Integer ) : Pointer ;

begin
    inc( _Allocations ) ;
    inc( _Allocation_Count ) ;
    Result := Old_Memory_Manager.GetMem( Size ) ;
end ;


function Test_FreeMem( P : Pointer ) : Integer ;

begin
    inc( _Deallocations ) ;
    dec( _Allocation_Count ) ;
    Result := Old_Memory_Manager.FreeMem( P ) ;
end ;


function Test_ReallocMem( P : Pointer ; Size : Integer ) : Pointer ;

begin
    inc( _Reallocations ) ;
    Result := P ;
    if( ( P = nil ) and ( Size = 0 ) ) then
    begin
        exit ;
    end ;
    if( P = nil ) then
    begin
        inc( _Allocations ) ;
        inc( _Allocation_Count ) ;
    end ;
    if( Size = 0 ) then
    begin
        inc( _Deallocations ) ;
        dec( _Allocation_Count ) ;
    end ;
    Result := Old_Memory_Manager.ReallocMem( P, Size ) ;
end ;


const New_Memory_Manager : TMemoryManager = (
                                               GetMem : Test_GetMem ;
                                               FreeMem : Test_FreeMem ;
                                               ReallocMem : Test_ReallocMem ;
                                            ) ;


// API...

function Allocations : integer ;

begin
    Result := _Allocations ;
end ;


function Allocation_Count : integer ;

begin
    Result := _Allocation_Count ;
end ;


procedure Begin_Memory_Test ;

begin
    if( not In_New_MM ) then
    begin
        GetMemoryManager( Old_Memory_Manager ) ;
        SetMemoryManager( New_Memory_Manager ) ;
        In_New_MM := True ;
    end ;
end ;


procedure Clear_Memory_Statistics ;

begin
    _Allocations := 0 ;
    _Deallocations := 0 ;
    _Reallocations := 0 ;
    _Allocation_Count := 0 ;
end ;


function Deallocations : integer ;

begin
    Result := _Deallocations ;
end ;


procedure End_Memory_Test ;

begin
    if( In_New_MM ) then
    begin
        SetMemoryManager( Old_Memory_Manager ) ;
        In_New_MM := False ;
    end ;
end ;


function Reallocations : integer ;

begin
    Result := _Reallocations ;
end ;



begin
    In_New_MM := False ;
    Clear_Memory_Statistics ;
end.

