{$N+}
{
        Program Name : CEF
        Package Name : CEF
        Purpose      : CEF definitions
        Institution  :
        Date Written : 27-Apr-2000
        Written By   : Alan Conroy
        Version      : 1.7

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Released to the public domain.
        
        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains all of the CEF type definitions in a format for
        Delphi.
}

unit CEF ;

interface

uses _UEHDefs, { TUEC }
     _DebugIn, // TDebug_Interface
     _Streams ; // TCOM_Stream

const Interface_Version = 20 ; // 2.0

      Severity_Information = 0 ; { Informational message }
      Severity_Warning = 1 ; { Warning }
      Severity_Error = 2 ; { Non-fatal error }
      Severity_Fatal = 3 ; { Fatal error.  The application should shut down. }

      Access_None = 0 ; { Not supported }
      Access_Read = 1 ; { Read (input) }
      Access_Input = Access_Read ;
      Access_Write = 2 ; { Write (output) }
      Access_Output = Access_Write ;
      Access_RW = Access_Read or Access_Write ; { Read or write (input or output) }
      Access_IO = Access_RW ;
      Access_Execute = 4 ; { Execution }
      Access_All = Access_RW or Access_Execute ;

      Component_Type_Unknown = 0 ;
      Component_Type_CPU = 1 ;
      Component_Type_Bus = 2 ;
      Component_Type_Memory = 3 ;
      Component_Type_Motherboard = 4 ;
      Component_Type_IO = 5 ;
      Component_Type_UI = 6 ;
      Component_Type_Cable = 7 ;
      Component_Type_Keyboard = 8 ;

      Segment_Default_Code = 0 ;
      Segment_Default_Data = 1 ;

      IO_Type_Memory = 0 ; // Memory read/write
      IO_Type_IO = 1 ; // I/O read/write
      // Note: all other IO types are specific to components

      { Child_Notice codes... }
      Child_Notice_Terminating = 0 ; // Child is about to terminate
      Child_Notice_Request_Terminate = 1 ; // Child is requesting termination.  If notice is set to 0 on return, the termination is disallowed
      Child_Notice_Receive_Data = 2 ; // Child has received asynchronus data

{ Error codes... }
const { TStreamer error codes... }
      StreamerErr_Facility = -1 ;
      StreamerErr_Success = 0 ;
      StreamerErr_No_Stream = 1 ; { No stream is open }

      { TComponent error codes... }
      ComponentErr_Facility = -1 ;
      ComponentErr_Success = 0 ;
      ComponentErr_No_Watchpoints = 1 ;
      ComponentErr_Invalid_Operation = 2 ;
      ComponentErr_Component_Not_Found = 3 ;
      ComponentErr_Address_Out_Of_Range = 4 ;

      { TMemory error codes... }
      MemoryErr_Facility = -1 ;
      MemoryErr_Success = 0 ;

      { TMaster_Assembler error codes... }
      MasterAssemblerErr_Facility = -1 ;
      MasterAssemblerErr_Success = 0 ;

      { TAssembler error codes... }
      AssemblerErr_Facility = -1 ;
      AssemblerErr_Success = 0 ;

      { TCPU error codes... }
      CPUErr_Facility = -1 ;
      CPUErr_Success = 0 ;
      CPUErr_No_Breakpoint = 1 ;

type TSymbol_Record = record
        Address : int64 ; { Address of identifier's storage within segment. }
        Segment : longint ; { Segment for identifier. }
        Size :	longint ; { Size of identifier's data, in bits. }
        Typ : 	longint ; { Data type of identifier:
                                    0 = Undefined
                                    1 = Integer
                                    2 = IEEE double-precision floating point
                                    3 = string
                                    4 = boolean
                                    5 = character
                                    other = other }
        Flags :	longint ; { Flags for identifier
                                    and 1 = Constant
                                    and 2 - 32768 = reserved
                                    other = other }
        Data :	int64 ;	{ Actual value of identifier, if < 65 bits }
        DataP :	pointer ; { Pointer to value, if > 64 bits }
    end ;
    PSymbol_Record = ^TSymbol_Record ;

    TAssembler_Status = class
                            public
                                function Filename : PChar ;
                                    virtual ; stdcall ; abstract ;

                                function Get_Aborted : boolean ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Code : int64 ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Data : int64 ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Errors : longint ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Warnings : longint ;
                                    virtual ; stdcall ; abstract ;
                                function Get_Error_Text : PChar ;
                                    virtual ; stdcall ; abstract ;

                                procedure Set_Aborted( Value : boolean ) ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Code( Value : int64 ) ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Data( Value : int64 ) ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Errors( Value : longint ) ;
                                    virtual ; stdcall ; abstract ;
                                procedure Set_Warnings( Value : longint ) ;
                                    virtual ; stdcall ; abstract ;
			        procedure Set_Error_Text( Value : PChar ) ;
                                    virtual ; stdcall ; abstract ;

                                procedure Get_Error( Index : longint ;
                                    var Text, Filename : PChar ;
                                    var Line, Severity : longint ) ;
                                    virtual ; stdcall ; abstract ;

                                procedure Log_Error( Text, Filename : PChar ;
                                    Line, Severity : longint ) ;
                                    virtual ; stdcall ; abstract ;

                                procedure Set_Line( Value : longint ) ;
                                    virtual ; stdcall ; abstract ;

                                property Aborted : boolean
                                    read Get_Aborted
                                    write Set_Aborted ;
                                property Code : int64
                                    read Get_Code
                                    write Set_Code ;
                                property Data : int64
                                    read Get_Data
                                    write Set_Data ;
                                property Errors : longint
                                    read Get_Errors
                                    write Set_Errors ;
                                property Warnings : longint
                                    read Get_Warnings
                                    write Set_Warnings ;
                                property Error_Text : PChar
                                    read Get_Error_Text
                                    write Set_Error_Text ;
                        end ; // TAssembler_Status

type TComponent = class ;
     TCPU = class ;
     TCable = class ;
     TMemory = class ;
     TMaster_Clock = class ;
     TUser_Interface = class ;
     TKeyboard = class ;

     TUI_Interface = class
         { This method is called whenever a component blocks (Blocked is true)
           or unblocks (Blocked is false).  Component is the component in
           question. }
         procedure Block( Component : TComponent ; Blocked : boolean ) ;
             virtual ; stdcall ;

         { This method is called when a CPU triggers a breakpoint.  The method
           returns True if the CPU is to continue execution of the instruction
           or False if not. CPU is the CPU component which triggered the
           breakpoint, Address is the execution address, and Physical indicates
           whether the address is physical or virtual.  Space is the index of
           the address space to assign the breakpoint to.  0 is the default
           address space. }
         function Breakpoint_Notice( Address : int64 ; Physical : boolean ;
             Space : integer ; CPU : TComponent ) : boolean ; virtual ; stdcall ;

         { This function returns the instance of the Master clock object for the
           application.  NULL is a valid return value and should be tested by
           any code calling this program (indicating that there is no master
           clock and timing issues are to be ignored). }
         function Clock : TMaster_Clock ; virtual ; stdcall ;

         { Returns a deugger object for this object. }
         function Debugger : TDebug_Interface ; virtual ; stdcall ;

         { Returns a stream for the passed (partial) filename.  If file cannot
           be opened/accessed, nil is returned. }
         function Get_File_Stream( Name : PChar ) : TCOM_Stream ;
             virtual ; stdcall ;

         // If Flag is true, the UI is hidden, otherwise it is made visible.
         procedure Hide( Flag : boolean ) ; virtual ; stdcall ;

         { Called by a blocked component (such as a CPU in wait state). }
         procedure Idle( Component : TComponent ) ; virtual ; stdcall ;

         { This logs an actual program error (exception or warning).  Severity
           is one of the following values:
               0 = Informational message
               1 = Warning
               2 = Non-fatal error
               3 = Fatal error.  The application should shut down. }
         procedure Log_Error( Component : TComponent ; Text : PChar ;
             Severity : longint ) ; virtual ; stdcall ;

         { This logs a simulated hardware error.  Severity is one of the
            following values:
                    0 = Informational message
                    1 = Warning
                    2 = Non-fatal error
                    3 = Fatal simulated hardware failure. }
         procedure Log_Simulated_Error( Component : TComponent ; Text : PChar ;
             Severity : longint ) ; virtual ; stdcall ;

         { A component calls this method in response to a Show_Status request.
           Each call provides one line of status information.  Index is 0 for
           the first line, 1 for the second, etc. }
         procedure Log_Status( Text : PChar ; Index : longint ) ; virtual ; stdcall ;

         { Logs a trace event for the specified component. }
         procedure Log_Trace( Component : TComponent ; Description : PChar ) ;
             virtual ; stdcall ;

         { Called when a component changes a signal state.  Index is the state
           index. Active is true if the new state of the indexed signal is
           active. }
         procedure Signal_Change_Notice( Component : TComponent ;
             Index : longint ; Active : boolean ) ; virtual ; stdcall ;

         { Called when a component suffers some form of exception.  Note that
           CPUs that support exception handling must report those exceptions via
           State_Change_Notice. }
         procedure Signal_Exception( Component : TComponent ;
             Description : PChar ; Index : longint ) ; virtual ; stdcall ;

         { Called when a component changes a state.  Index is the state index.
           Active is true if the new state of the indexed state is active. }
         procedure State_Change_Notice( Component : TComponent ;
             Index : longint ; Active : boolean ) ; virtual ; stdcall ;

         // Terminates the UI, which usually terminates the application
         procedure Terminate ; virtual ; stdcall ;

         { Toggle the embededed state of the specified TUser_Interface
           component. }
         procedure Toggle_Embed( Component : TComponent ) ; virtual ; stdcall ;

         { Returns the CEF specification version number to which this interface
           conforms. }
         function Version : integer ; virtual ; stdcall ;

         { Indicates that the specified component wishes to be notified of
           signal changes (or not). }
         procedure Want_Signals( Component : TComponent ; Value : boolean ) ;
             virtual ; stdcall ;

         { This method is called when a component triggers a watchpoint.
           Address is the address that triggered the watchpoint.  Access
           indicates the type of access:
               2 = Read (input)
               3 = Write (output)
           Tag is dependant upon the component and Component is the component
           which triggered the watchpoint.  Memory is true if this is a memory
           watchpoint.  Internal is true if this is a watch on component
           internals.  Port is true if this is a port I/O access. }
         procedure Watchpoint_Notice( Address : int64 ; Access, Tag : longint ;
             Component : TComponent ; Memory, Internal, Port : boolean ) ;
             virtual ; stdcall ;
     end ; // TUI_Interface


     TProfiler = class
                     public
                         { Clears profiling information for specified domain.
                           If Domain is -1, all domains are cleared. }
                         procedure Clear( Domain : integer ) ;
                             virtual ; stdcall ;

                         { Returns profiling domain name for given index.  If
                           the index is invalid, the function returns nil.  A
                           domain indicates a type of profiling data.  For
                           instance, a CPU may track the number of calls to
                           certain addresses as one domain, and the number of
                           calls to specific types of instructions in another
                           domain, etc. }
                         function Domain_Name( Index : integer ) : PChar ;
                             virtual ; stdcall ;

                         { Returns a line of information associated with the
                           specified domain.  The first line is Index 0.  If an
                           invalid index is passed, the function returns nil. }
                         function Report_Line( Domain, Index : integer ) : PChar ;
                             virtual ; stdcall ;

                         { Returns the CEF specification version number to which
                           this interface conforms. }
                         function Version : integer ; virtual ; stdcall ;
                 end ; // TProfiler


     TComponent = class
                      public
            { This function returns a UEH facility code for the component class. }
            function Facility_Code : longint ; virtual ; stdcall ;

            { Reinitalizes the component.  Note that this is not a constructor!
              It is not necessary for the application to call this method, since
              Make_Instance should have initialized the component appropriately.
              This simply allows the component to be initialized again. }
            function Initialize( UI : TUI_Interface ) : TUEC ;
                virtual ; stdcall ;

            { Terminates the component.  Note that this is not a destructor,
              although it should eventually cause the instance to be destructed.
              By convention, once this method is called, the instance is no
              longer valid. }
            function Terminate : TUEC ; virtual ; stdcall ;

            { Returns number assigned by component creator. }
            function Serial_Number : integer ; virtual ; stdcall ;

            { Adds the specified component to the list of components to notify
              along with Parent in Child_Notifications. }
            function Add_Notification( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            function Cable : TCable ; virtual ; stdcall ;

            { This method returns the Indexth child component for this component.
              It returns nil if Index is outside of the range: 0 <= n <
              Max_child_component.  Child components are components owned by
              this component (such as a motherboard would own the CPU component). }
            function Child_Component( Index : longint ) : TComponent ;
                virtual ; stdcall ;

            { Indicates a notice from a child component.  Child is the component
              and Notice is a value from Child_Notice_*. }
            procedure Child_Notification( Child : TComponent ;
                var Notice : longint ; var Params : int64 ) ; virtual ; stdcall ;

            { Clears a watchpoint at the specified address.  Access is the type
              of access:
                            1 = Read or write (input or output)
                            2 = Read (input)
                            3 = Write (output)
              If Memory is true, the watch occurs on memory accesses from the
              component.
              If Memory is false, the watch occurs on port accesses from the
              component. }
            function Clear_Watchpoint( Address : int64 ; Memory : boolean ;
                Access : longint ) : TUEC ; virtual ; stdcall ;

            { This method returns a value indicating the type of the component,
              as follows:
                            0 = Unknown (TComponent)
                            1 = CPU
                            2 = Bus
                            3 = Memory
                            4 = Motherboard
                            5 = I/O device
                            6 = User Interface }
            function Component_Type : longint ; virtual ; stdcall ;

            { For components which accept input, this connects another component
              to this component's default input port. }
            function Connect_Input( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            { For components which generate output, this connnects another
              component to this component's default output port. }
            function Connect_Output( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            function CPU : TCPU ; virtual ; stdcall ;

            { Returns debug interface for this component. }
            function Debugger : TDebug_Interface ; virtual ; stdcall ;

            { This removes the specified component from the list of components
              to be notified.  The component must have been added via
              Add_Notification. }
            function Delete_Notification( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            { This method writes Size bits of data from the specified address
              from the specified buffer.  Memory is True if this is a memory
              write versus a I/O write.  This is solely for the UI, and the
              write does not trigger watchpoints or have any other side-effects
              other than changingthe contents of the component. }
            function Deposit( Address : int64 ; Size : longint ; Buffer : pointer ;
                Memory : boolean ) : TUEC ; virtual ; stdcall ;

            { This disconnects the specified component from this component's
              default input port. }
            function Disconnect_Input( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            { This disconnects the specified component from this component's
              default output port. }
            function Disconnect_Output( Component : TComponent ) : TUEC ;
                virtual ; stdcall ;

            { This method reads Size bits of data from the specified address
              into the specified buffer.  Size is updated with the actual number
              of bits copied.  Memory is True if this is a memory read versus an
              I/O read.  This is solely for the UI - the read triggers no
              watchpoints and has no other side-effects.  It also ignores access
              modes. }
            function Examine( Address : int64 ; var Size : longint ;
                Buffer : pointer ; Memory : boolean ) : TUEC ; virtual ; stdcall ;

            { Returns the access mode of the specified address.  Return values
              are:
                            0 = Not supported (address out of range, etc.)
                            1 = Read/Write (I/O)
                            2 = Read-only (input-only)
                            3 = Write-only (write-only)
              Memory is true if address is a memory address and false if it is
              an I/O address. }
            function Get_Access_Mode( Address : int64 ;
                Memory : boolean ) : longint ; virtual ; stdcall ;

            { Provides a description of the exception conditions signaled via
              Signal_Exception. }
            function Get_Exception_Description( Index : longint ) : PChar ;
                virtual ; stdcall ;

            function Get_Parent : TComponent ; virtual ; stdcall ;

            { Returns True if profiling for this component is on. }
            function Get_Profiling : boolean ; virtual ; stdcall ;

            { Returns the component's default read latency.  If the component
              doesn't support latency, this should return 0. }
            function Get_Read_Latency : longint ; virtual ; stdcall ;

            { This returns the state of a signal with the specified name.  If
              the component recognizes the signal name, it returns True.
              Otherwise it returns False.  If the result is True, State is set
              to the current value of the signal. }
            function Get_Signal( Name : PChar ; var State : boolean ) : boolean ;
                virtual ; stdcall ;

            { Get the name of a component state.  If index is invalid, the
              function returns nil. }
            function Get_State_Name( Index : longint ) : PChar ;
                virtual ; stdcall ;

            function Get_Tag : longint ; virtual ; stdcall ;

            function Get_Trace : boolean ; virtual ; stdcall ;

            { Returns the component's default write latency.  If the component
              doesn't support latency, this should return 0. }
            function Get_Write_Latency : longint ; virtual ; stdcall ;

            { This method returns the Indexth input component for this component.
              It returns NULL if Index is outside of the range: 0 <= n <
              Max_input_component }
            function Input_Component( Index : longint ) : TComponent ;
                virtual ; stdcall ;

            function Keyboard : TKeyboard ; virtual ; stdcall ;
        
            function Memory : TMemory ; virtual ; stdcall ;

            { This method returns a string containing the name of this component. }
            function Name : PChar ; virtual ; stdcall ;

            { This method returns the Indexth output component for this
              component.  It returns NULL if Index is outside of the range:
              0 <= n < Max_output_component }
            function Output_Component( Index : longint ) : TComponent ;
                virtual ; stdcall ;

            function Profiler : TProfiler ; virtual ; stdcall ;

            { Calling this method requests the component to transmit data to its
              default output port.  IO_Type is the type of read operation (see
              IO_Type_*).  Size is the number of bits being requested (0=default
              for component).  The component should verify that the address is
              within its range before responding.  If not, it should do nothing.
              If the component will respond to the request, the method returns
              true, otherwise it returns false. }
            function Read( Address : int64 ; Size : longint ;
                IO_Type : longint ) : boolean ; virtual ; stdcall ;

            { Reset the component.  Behavior is implementation dependant. }
            procedure Reset ; virtual ; stdcall ;

            { This method causes the component to restore its contents (but NOT
              state) from the passed stream.  It is the responsibility of the
              component to restore the contents of any child components (but not
              attached components). }
            function Restore_Contents( Stream : TCOM_Stream ) : TUEC ;
                virtual ; stdcall ;

            { This method causes the component to restore its current state (but
              NOT contents) from the passed stream.  It is the responsibility of
              the component to restore the state of any child components (but
              not attached components). }
            function Restore_State( Stream : TCOM_Stream ) : TUEC ;
                virtual ; stdcall ;

            { This method causes the component to save its contents (but NOT
              state) to the passed stream.  It is the responsibility of the
              component to save the contents of any child components (but not
              attached components). }
            function Save_Contents( Stream : TCOM_Stream ) : TUEC ;
                virtual ; stdcall ;

            { This method causes the component to save its current state (but
              NOT contents) to the passed stream.  It is the responsibility of
              the component to save the state of any child components (but not
              attached components). }
            function Save_State( Stream : TCOM_Stream ) : TUEC ;
                virtual ; stdcall ;

            { Sets the access mode of the specified address range (Low to High,
              inclusive).  Typ is:
                            1 = Read/Write (I/O)
                            2 = Read-only (input-only)
                            3 = Write-only (write-only)
              Memory is true if address is a memory address and false if it is
              an I/O address. }
            function Set_Access_Mode( Low, High : int64 ; Memory : boolean ;
                Typ : longint ) : TUEC ; virtual ; stdcall ;

            procedure Set_Parent( Component : TComponent ) ; virtual ; stdcall ;

            { Turns profiling for this component on or off.  Profiling is off by
              default.  If Children is true, all child components will be
              assigned the same profiling value (this doesn't affect input or
              output components). }
            procedure Set_Profiling( _On, Children : boolean ) ;
                virtual ; stdcall ;

            { Sets the component's default read latency in nanoseconds.  If the
              component supports latency, it should block by this amount on each
              read. }
            procedure Set_Read_Latency( Value : longint ) ; virtual ; stdcall ;

            { This sets the state of a signal with the specified name.  This
              should only be called when the signal state actually changes (that
              is, State should never be the same on two consecutive calls). }
            procedure Set_Signal( Name : PChar ; State : boolean ) ;
                virtual ; stdcall ;

            { User-defined value associated with component. }
            procedure Set_Tag( Value : longint ) ; virtual ; stdcall ;

            procedure Set_Trace( Value : boolean ) ; virtual ; stdcall ;

            { An component-specific initialization string can be sent to the
              component after initialization, via this method. }
            procedure Set_Up( P : PChar ) ; virtual ; stdcall ;

            { Sets a watchpoint at the specified address.  When the component is
              about to access this address, it will generate a watchpoint notice
              to the UI interface.  Access is the type of access:
                            1 = Read or write (input or output)
                            2 = Read (input)
                            3 = Write (output)
              If Memory is true, the watch occurs on memory accesses from the
              CPU.
              If Memory is false, the watch occurs on port accesses from the CPU. }
            function Set_Watchpoint( Address : int64 ; Memory : boolean ;
                Access : longint ) : TUEC ; virtual ; stdcall ;

            { Sets the component's default write latency in nanoseconds.  If the
              component supports latency, it should block by this amount on each
              write. }
            procedure Set_Write_Latency( Value : longint ) ;  virtual ; stdcall ;

            { Calling this method causes the component to report its status to
              the UI interface, in a human-readable form.  For instance, a CPU
              component would report its registers and their contents. }
            procedure Show_Status ; virtual ; stdcall ;

            { Called when the specified component has a signal change.  This is
              only called if this component requested signal notices from the
              UI. }
            procedure Signal_Change_Notice( Component : TComponent ;
                Index : longint ; Active : boolean ) ; virtual ; stdcall ;

            { Returns the number of defined signals. }
            function Signal_Count : longint ; virtual ; stdcall ;

            { Returns the name of the supported signals.  The first index is 0.
              If there is no signal for the specified index, or index is less
              than 0, the result is nil. }
            function Signal_Name( Index : longint ) : PChar ; virtual ; stdcall ;

            { Returns True if the signal is an output from the component.
              Result is indefined if Index is invalid. }
            function Signal_Out( Index : longint ) : boolean ; virtual ; stdcall ;

            { Returns True if the signal is active-low.  Result is undefined if
              Index is invalid. }
            function Signal_Active_Low( Index : longint ) : boolean ;
                virtual ; stdcall ;

            { Returns the index of the signal with the specified name.  If the
              name is invalid, -1 is returned. }
            function Signal_Index( Name : PChar ) : integer ; virtual ; stdcall ;

            { Returns True if the component supports the specified feature.  If
              the component does not inherently support the feature (such as a
              CPU feature for a memory component), the result is undefined. }
            function Support_Feature( ID : longint ) : boolean ;
                virtual ; stdcall ; stdcall ;

            function User_Interface : TUser_Interface ;
                virtual ; stdcall ; stdcall ;

            { Returns the CEF specification version number to which this
              interface conforms. }
            function Version : longint ; virtual ; stdcall ;

            { Calling this allows a component to resume its activity after it
              sends a Block request to the Master clock. }
            procedure Wake ; virtual ; stdcall ;

            { Calling this method writes Value to the component's input port
              with the specified address.  IO_Type is the type of write
              operation (see IO_Type_*).  Size is the number of bits in Value to
              write (0=defaulr for component) and may be 0 to 32, inclusive.
              The component should verify that the address is within its range
              before responding.  If not, it should do nothing, and return no
              errors. }
            function Write( Address : int64 ; Value, Size : longint ;
                IO_Type : longint ) : TUEC ; virtual ; stdcall ;

            { This is the same as calling Write_Byte once for each byte in the
              string, incrementing Address once per byte.  It is provided to
              allow more efficient means of transmitting large amounts of data.
              Size is the number of bits in Value to write (0=default for
              component).  The component should verify that the address is
              within its range before responding.  If not, it should do nothing,
              and return no errors. }
            function Write_String( Address : int64 ; Value : PChar ;
                Size : longint ; IO_Type : longint ) : TUEC ; virtual ; stdcall ;

            property Tag : longint
                read Get_Tag
                write Set_Tag ;

            property Parent : TComponent
                read Get_Parent
                write Set_Parent ;
        end ; // TComponent


    TMaster_Clock = class
	 { This procedure registers a request for component to receive a Wake
           signal at the specified time delta (in picoseconds). }
	 procedure Block( Component : TComponent ; Time_Delta : int64 ) ;
             virtual ; stdcall ;

         { Returns debug interface for this clock. }
         function Debugger : TDebug_Interface ; virtual ; stdcall ;

	 { Initializes the clock to time index 0 and sets the UI interface. }
	 procedure Initialize( UI : TUI_Interface ) ; virtual ; stdcall ;

	 { This function returns the current simulated master clock (which is
           measured in nanoseconds) time index. }
	 function Get_Time_Index : int64 ; virtual ; stdcall ;

         { Returns value passed by creator. }
         function Serial_Number : integer ; virtual ; stdcall ;

         { Returns True if the clock supports the specified feature.  If the
           clock does not inherently support the feature (such as a CPU
           feature), the result is undefined. }
         function Support_Feature( ID : integer ) : boolean ;
             virtual ; stdcall ; stdcall ;

         { Returns the CEF specification version number to which this interface
           conforms. }
         function Version : integer ; virtual ; stdcall ;
    end ; // TMaster_Clock


     TCable = class
                public
                    { Returns true if this cable is a serial cable.  It is a
                      parallel cable otherwise. }
                    function Serial : boolean ; virtual ; stdcall ;

                    { Returns name of the protocol supported by the cable (nil
                      if generic/any). }
                    function Protocol : PChar ; virtual ; stdcall ;

                    { Transmit one data item.  Speed is the bit rate of the
                      transfer (0=automatch), Value is the data to transmit,
                      Data_Size is the number of bits in Value (32 max),
                      Stop_Bits indicates the number of stop bits. }
                    function Transmit( Speed : int64 ;
                        Value, Data_Size, Stop_Bits : longint ) : TUEC ;
                        virtual ; stdcall ;

                    { Transmits a string of data by calling Transmit once for
                      each data item in Value. }
                    function Transmit_String( Speed : int64 ; Value : PChar ;
                        Data_Size, Stop_Bits : longint ) : TUEC ;
                        virtual ; stdcall ;

                    { Called when received data from another cable component }
                    procedure Receive( Source : TComponent ; Speed : int64 ;
                        Value, Data_Size, Stop_Bits : integer ) ;
                        virtual ; stdcall ;

                    function Get_Data( var Speed : int64 ;
                        var Value, Data_Size, Stop_Bits : integer ) : boolean ;
                        virtual ; stdcall ;
             end ; // TCable


     TUser_Interface = class
                           public
                               function Get_Hidden : boolean ;
                                   virtual ; stdcall ;
                               procedure Set_Hidden( Value : boolean ) ;
                                   virtual ; stdcall ;

                               function Get_Parent_Window : THandle ;
                                   virtual ; stdcall ;
                               procedure Set_Parent_Window( Value : THandle ) ;
                                   virtual ; stdcall ;

                               function Get_Caption : PChar ;
                                   virtual ; stdcall ;
                               procedure Set_Caption( Value : PChar ) ;
                                   virtual ; stdcall ;

                               procedure Set_Size( Height, Width : integer ) ;
                                   virtual ; stdcall ;

                               function Optimal_Height : integer ;
                                   virtual ; stdcall ;

                               function Optimal_Width : integer ;
                                   virtual ; stdcall ;

                               { Returns the CEF specification version number to
                                 which this interface conforms. }
                               function Version : integer ; virtual ; stdcall ;

                               property Hidden : boolean
                                   read Get_Hidden
                                   write Set_Hidden ;
                               property Parent_Window : THandle
                                   read Get_Parent_Window
                                   write Set_Parent_Window ;
                               property Caption : PChar
                                   read Get_Caption
                                   write Set_Caption ;
                       end ;


    TKeyboard = class
                    public
                        { Retrieves the next key in a keystroke from the
                          keyboard.  Returns nil or null string if end of
                          current keystroke. }
                        function Get_Key : PChar ; virtual ; stdcall ;

                        { Returns True if the specified key is down in the
                          pending keystroke.  If Name is nil, returns True if
                          any key is down. }
                        function Get_Key_Down( Name : PChar ) : boolean ;
                            virtual ; stdcall ;

                        { Sets the state of a given key to up or down.  The key
                          is reset when the next keystroke is returned.  This
                          can be used to synchronize the keyboard component's
                          visible state with the physical keyboard. }
                        procedure Set_Key_Down( Name : PChar ; State : boolean ) ;
                            virtual ; stdcall ;

                        { Returns the name of the key associated with the
                          passed index.  Returns nil if the index is invalid. }
                        function Get_Key_Name( Index : integer ) : PChar ;
                            virtual ; stdcall ;

                        { Returns True if the specified LED is lit.  If the LED
                          name is not recognized, the function returns false. }
                        function Get_LED_State( Name : PChar ) : boolean ;
                            virtual ; stdcall ;

                        { Sets the state of a given LED to lit or not.  This
                          can be used to synchronize the keyboard component's
                          visible state with the physical keyboard. }
                        procedure Set_LED_State( Name : PChar ; State : boolean ) ;
                            virtual ; stdcall ;

                        { Returns the name of the LED associated with the
                          passed index.  Returns nil if the index is invalid. }
                        function Get_LED_Name( Index : integer ) : PChar ;
                            virtual ; stdcall ;

                        { Returns the CEF specification version number to which this interface
                          conforms. }
                        function Version : integer ; virtual ; stdcall ;
                end ;


     TMemory = class
                   public // API...
                       { Dumps an image of memory between Start and Start + Size - 1, inclusive,
                         to the passed buffer, which must be large enough to hold Size bytes.
                         Any range of addresses that are outside the memory's range are zero
                         filled. }
                       procedure Dump( Start, Size : int64 ; Buffer : pointer ) ;
                           virtual ; stdcall ;

                       { Returns the set of addresses that this component
                         responds to. }
                       procedure Get_Address_Range( var Low, High : int64 ) ;
                           virtual ; stdcall ;

                       { Loads memory between Start and Start + Size - 1, inclusive,
                         with the data in the passed buffer.  Any data outside
                         the memory's range is ignored. }
                       procedure Load( Start, Size : int64 ; Buffer : pointer ) ;
                           virtual ; stdcall ;

                       { Defines the set of addresses that this component r
                         esponds to. }
                       function Set_Address_Range( Low, High : int64 ) : TUEC ;
                           virtual ; stdcall ;

                       { Returns the CEF specification version number to which
                         this interface conforms. }
                       function Version : integer ; virtual ; stdcall ;
               end ; { TMemory }


     TMaster_Assembler = class
                             public // API...
        { Adds a reference for back-patching. }
        procedure Add_Reference( Name : PChar ; Size : longint ;
            Address : int64 ) ; virtual ; stdcall ;

        { Adds a symbol to the current scope, with the name in Name and the
          specified information. }
        function Add_Symbol( Name : PChar ; P : pSymbol_Record ) : TUEC ;
            virtual ; stdcall ;

        { Assembles code from the Input stream, directing the image data to the
          Output stream, and placing any list output to the Listing stream.
          Note that Listing can be NULL if no listing is desired.  Output can be
          NULL to do a syntax-check-only assembly.  Status is used to provide
          assembly statistics. }
        function Assemble( Input, Output, Listing : TCOM_Stream ;
            Status : TAssembler_Status ) : TUEC ; virtual ; stdcall ;

        { Apply all backpatches. }
        procedure Backpatch( Status : TAssembler_Status ;
            Stream : TCOM_Stream ) ; virtual ; stdcall ;

        { Evaluates an expression, returning the numeric value in Value. }
        function Evaluate( X : PChar ; var Value : int64 ) : TUEC ;
            virtual ; stdcall ;

        { This method is used during assmebly by a TAssembler to insert assembly
          source in-place.  For instance, translating a directive into a
          standard directive, or doing macro expansions, etc.  If the expanded
          item is a directive that results in output, Res will be that output
          and Res_Length will be the size of that data, in bytes. }
        function Expand( Source : PChar ; var Res : PChar ;
            var Res_Length : longint ; Status : TAssembler_Status ) : TUEC ;
            virtual ; stdcall ;

        { Returns facility code for class }
        function Facility_Code : longint ; virtual ; stdcall ;

        { Returns True if case sensitivity to identifier names is on. }
        function Get_Case_Sensitive : boolean ; virtual ; stdcall ;

        { Returns the symbol record for the identifier in Name.  Returns NULL,
          if Name is undeclared. }
        function Get_Symbol( Name : PChar ) : PSymbol_Record ;
            virtual ; stdcall ;

        { Returns the next token in the input stream during an assembly.
          Returns a null string if an error occurs (such as no more input).
          This is useful for assemblers which need more input for a directive
          which is continued on additional lines. }
        function Get_Token : PChar ; virtual ; stdcall ;

        { Returns entire current (remaining) input line. }
        function Grab_Line( Current : boolean ) : PChar ; virtual ; stdcall ;

        { Inserts the data from Input into the current token stream. }
        procedure In_Line( Input : TCOM_Stream ) ; virtual ; stdcall ;

        { Logs an assembly error.  Severity indicates the seriousness of the
          error:
                        0 = Informational
                        1 = Warning
                        2 = Error
                        3 = Fatal error (abort assembly) }
        procedure Log_Error( Text : PChar ; Severity : longint ) ; virtual ; stdcall ;

        // Map current token position to the passed address.
        procedure Map( Address : int64 ) ; virtual ; stdcall ;

        { Return next token without removing from the token stream.  If Same_Line
          is True, only return the next token on the current line. }
        function Peek_Token( Same_Line : boolean ) : PChar ; virtual ; stdcall ;

        { Ends the current sub-scope for identifiers. }
        procedure Pop_Scope ; virtual ; stdcall ;

        { Begins a sub-scope for identifiers within the current scope. }
        procedure Push_Scope ; virtual ; stdcall ;

        { Returns a token to the input stream during an assembly. }
        procedure Put_Token( Token : PChar ) ; virtual ; stdcall ;

        { Sets case sensitivity on identifier names. }
        procedure Set_Case_Sensitive( Value : boolean ) ; virtual ; stdcall ;

        // Remove previous mapping added with Map().
        procedure UnMap ; virtual ; stdcall ;

        { Returns the CEF specification version number to which this interface
          conforms. }
        function Version : integer ; virtual ; stdcall ;

        property Case_Sensitive : boolean
            read Get_Case_Sensitive
            write Set_Case_Sensitive ;
     end ; // TMaster_Assembler


     TAssembler = class
        { Informs the assembler the a new assembly operation is beginning.  This
          assembly will use the passed master assembler. }
        procedure Initialize( Master : TMaster_Assembler ) ; virtual ; stdcall ;

        { Informs the assembler that the current assembly operation is now
          complete. }
        procedure Terminate ; virtual ; stdcall ;

	{ This method assembles the source code in the inputs string.  If inputs
          is a null string, the assembler should fetch the data from the master
          assembler (Get_Token method).  Formatted output (if any) is returned
          in outputs, which is assumed to be in a human-readable form.  Upon
          return, machines contains the generated machine code for the
          instruction(s) passed.  Machinel is the length of the data in
          Machines.  Address is the target memory address where the Machines
          data is to go.  Segment is the current destination segment for the
          data in machines.  It has the following values:
		0 = Default code segment
		1 = Default data segment
		other = CPU-dependant or user-defined segments
          Machinel, Address, and Segment are passed the current values from the
          master assembler, but the called assembler may change them.  Address
          is virtual (physical if virtual addresssing not supported by CPU). }
        function Assemble( inputs : PChar ; var outputs, machines : PChar ;
            var MachineL : longint ; var Address : int64 ;
            var Segment : longint ; Status : TAssembler_Status ) : TUEC ;
            virtual ; stdcall ;

        { Returns the default radix (base) of numeric literals. }
        function Default_Radix : longint ; virtual ; stdcall ;

        { Returns the default size of numeric literals, in bits. }
        function Default_Size : longint ; virtual ; stdcall ;

        { Returns facility code for this class. }
        function Facility_Code : longint ; virtual ; stdcall ;

        { Return a bar-delimited list of source file extensions.  The first
          one should be the default. }
        function Source_Extensions : PChar ; virtual ; stdcall ;

        { Returns a list of valid characters for starting a symbol name. }
        function Valid_Symbol_Initial : PChar ; virtual ; stdcall ;

        { Returns a list of valid characters for a symbol name, past the first. }
        function Valid_Symbol_After : PChar ; virtual ; stdcall ;

        { Returns the CEF specification version number to which this interface
          conforms. }
        function Version : integer ; virtual ; stdcall ;
     end ; { TAssembler }


     TCPU = class
        { Returns true if the CPU is big-endian.  Returns False if small-endian. }
        function Big_Endian : boolean ; virtual ; stdcall ;

        { Clears a breakpoint at the specified address.  If Physical is true,
          the address is a physical address, otherwise it is a virtual
          address. }
        function Cancel_Breakpoint( Address : int64 ; Space : longint ;
            Physical : boolean ) : TUEC ; virtual ; stdcall ;

	{ Clears a watchpoint at the specified address.  Access is the type of
	  access:
		1 = Read or write (input or output)
		2 = Read (input)
		3 = Write (output)
          If Memory is true, the address is an internal cache address.  Otherwise
          it is a register index. }
        function Clear_Internal_Watchpoint( Address : int64 ; Memory : boolean ;
            Access : longint ) : TUEC ; virtual ; stdcall ;

        { Indicates the specified base used by the CPU instruction set. }
        function Default_Base : longint ; virtual ; stdcall ;

        { Disassembles the instructions at the specified address, for the
          specified number of bytes and directs the output to the passed
          stream.  Constant values are shown in the specified base.  If base is
          0, the CPU's default base is used. }
        function Disassemble( Address : int64 ; Base, Size : longint ;
            Stream : TCOM_Stream ) : TUEC ; virtual ; stdcall ;

        // Returns facility code
        function Facility_Code : longint ; virtual ; stdcall ;

        { This method cosntructs and returns an assembler object for this CPU.
          It is passed a TMaster_Assembler object. }
        function Get_Assembler( Master : TMaster_Assembler ) : TAssembler ;
            virtual ; stdcall ;

        { Returns the CPU clock speed (in Hz). }
        function Get_Clock_Speed : longint ; virtual ; stdcall ;

        { Returns the current CPU memory position (usually the Program Counter). }
        function Get_Current_Address( Space : longint ;
            Physical : boolean ) : int64 ; virtual ; stdcall ;

        { Indicates the lowest physical memory address accessable by the CPU. }
        function Get_Low_Memory : int64 ; virtual ; stdcall ;

        { Indicates the highest physical memory address accessable by the CPU. }
        function Get_High_Memory : int64 ; virtual ; stdcall ;

        { Indicates the lowest port address accessable by the CPU. }
        function Get_Low_Port : int64 ; virtual ; stdcall ;

        { Indicates the highest port address accessable by the CPU. }
        function Get_High_Port : int64 ; virtual ; stdcall ;

        { Indicates the lowest virtual memory address accessable by the CPU. }
        function Get_Low_Virtual_Memory( Space : longint ) : int64 ;
            virtual ; stdcall ;

        { Indicates the highest virtual memory address accessable by the CPU. }
        function Get_High_Virtual_Memory( Space : integer ) : int64 ; virtual ; stdcall ;

        { Halts CPU execution.  The effect upon the CPU is implementation
          dependant. }
        procedure Halt ; virtual ; stdcall ;

        { Returns true if the CPU is in a halted state (non-execution mode). }
        function Halted : boolean ; virtual ; stdcall ;

        { Returns a description of the memory space with the specified index.
          It returns nil for any invalid index.  0 is always valid.  Physical is
          true for physical memory space and false for virtual address spaces. }
        function Memory_Space_Description( Index : longint ;
            Physical : boolean ) : PChar ; virtual ; stdcall ;

        { Returns memory page size, in bytes.  Returns 0 if non-paging CPU. }
        function Page_Size : longint ; virtual ; stdcall ;

        { Returns a description of the register associated with the passed
          index.  }
        function Register_Description( Index : longint ) : PChar ;
            virtual ; stdcall ;

        { Returns the name corresponding to the register index passed.  A null
          string or NULL result indicates that the index is out of range.  Note
          that all registers start at offset 0. }
        function Register_Name( Index : longint ) : PChar ; virtual ; stdcall ;

        { Returns the size (in bits) of the register associated with the passed
          index.  A result of 0 indicates that the index is out of range.  Note
          that all registers start at offset 0. }
        function Register_Size( Index : longint ) : integer ;
            virtual ; stdcall ;

        // Power-on (cold) reset.
        procedure Restart ; virtual ; stdcall ;

        { Begins CPU execution from the current CPU state.  Typically one
          instruction is executed, and then the CPU requests a block from the
          master clock. }
        procedure Run ; virtual ; stdcall ;

        { The causes the CPU to execute instructions directly from the passed
          stream.  This is usually for immediate instruction execution from the
          user interface, thus no profiling is done, no breakpoints apply, and
          single-stepping does not occur. }
        procedure Run_From_Stream( Stream : TCOM_Stream ) ; virtual ; stdcall ;

        { Sets a breakpoint at the specified address.  When the CPU is about to
          begin execution at this address, it will generate a breakpoint notice
          to the UI interface.  If Physical is true, the address is a physical
          address, otherwise it is a virtual address. }
        function Set_Breakpoint( Address : int64 ; Space : integer ;
            Physical : boolean ) : TUEC ; virtual ; stdcall ;

        { Sets the CPU clock speed (in Hz). }
        procedure Set_Clock_Speed( Value : longint ) ; virtual ; stdcall ;

        { Sets the current CPU memory position. }
        procedure Set_Current_Address( Space : integer ; Physical : boolean ;
            Value : int64 ) ; virtual ; stdcall ;

	{ Sets a watchpoint at the specified address.  When the CPU is about to
          access this address, it will generate a watchpoint notice to the UI
          interface.  Access is the type of access:
		1 = Read or write (input or output)
		2 = Read (input)
		3 = Write (output)
          If Memory is true, the address applies to internal cache.  If Memory is
          false, address is the register. }
        function Set_Internal_Watchpoint( Address : int64 ; Memory : boolean ;
            Access : integer ) : TUEC ; virtual ; stdcall ;

        { Executes a single instruction based on the current CPU context.  This
          is the same as Run(), but execution doesn't continue after the CPU is
          unblocked.  If Into is true, only a single instruction is excecuted.
          If Into is false, and the instruction to be executed is a nested
          instruction (such as a subroutine call), execution continues until the
          execution returns to the instruction following the call. }
        procedure Step( Into : boolean ) ; virtual ; stdcall ;

        { Stop the CPU execution. }
        procedure Stop ; virtual ; stdcall ;

        { Returns True if the CPU supports virtual address mapping. }
        function Support_Virtual_Address : boolean ; virtual ; stdcall ;

        { Returns address of top of the indexth stack. }
        function Top_Of_Stack( Index : longint ) : int64 ; virtual ; stdcall ;

        { Returns the physical address for the passed virtual address in the
          indicated address space.  For a CPU with no virtual addresses, this
          returns the value passed. }
        function Translate( Space : longint ; Address : int64 ) : int64 ;
            virtual ; stdcall ;

        { Returns the CEF specification version number to which this interface
          conforms. }
        function Version : integer ; virtual ; stdcall ;
    end ; // TCPU


{ File layouts... }

{ Dump file layout }

type TDump_File_Header = record
        Signature : word ;
        Subsignature : byte ; { 0 = CEF }
        Version : byte ; { File format version number times 10 (1.2=12) }
        CEF_Type : byte ; { 0 = Dump }
        Stamp : longint ; { Optional packed date/time stamp }
        Start : int64 ; { First address in image }
        Filler : array[ 17..255 ] of byte ; { Filler - should be 0 }
    end ;


{ State file layout }

type TState_File_Header = record
        Signature : word ;
        Subsignature : byte ; { 0 = CEF }
        Version : byte ; { File format version number times 10 (1.2=12) }
        CEF_Type : byte ; { 1 = State }
        Stamp : longint ; { Optional packed date/time stamp }
        Filler : array[ 9..255 ] of byte ; { Filler - should be 0 }
    end ;



type TState_File_Sub_Header = record
        Facility : longint ; { Facility code of component }
        Serial_Number : longint ; { Serial number of component }
    end ;


implementation

{ Note that Borland Pascal doesn't allow pure virtual base classes, so we must
  define a generic implementation of each of class. }

{ TUI_Interface methods... }

procedure TUI_Interface.Block( Component : TComponent ; Blocked : boolean ) ;

begin
end ;


function TUI_Interface.Breakpoint_Notice( Address : int64 ; Physical : boolean ;
    Space : integer ; CPU : TComponent ) : boolean ;

begin
    Breakpoint_Notice := False ;
end ;


function TUI_Interface.Clock : TMaster_Clock ;

begin
    Clock := nil ;
end ;


procedure TUI_Interface.Log_Error( Component : TComponent ; Text : PChar ;
    Severity : longint ) ;

begin
end ;


procedure TUI_Interface.Log_Simulated_Error( Component : TComponent ;
    Text : PChar ; Severity : longint ) ;

begin
end ;


procedure TUI_Interface.Log_Status( Text : PChar ; Index : longint ) ;

begin
    writeln( Text ) ;
end ;


function TUI_Interface.Debugger : TDebug_Interface ;

begin
    Result := nil ;
end ;


function TUI_Interface.Version : integer ;

begin
    Version := Interface_Version ;
end ;


procedure TUI_Interface.Watchpoint_Notice( Address : int64 ;
    Access, Tag : longint ; Component : TComponent ;
    Memory, Internal, Port : boolean ) ;

begin
end ;


procedure TUI_Interface.Terminate ;

begin
end ;


procedure TUI_Interface.Hide( Flag : boolean ) ;

begin
end ;


procedure TUI_Interface.Idle( Component : TComponent ) ;

begin
end ;


procedure TUI_Interface.State_Change_Notice( Component : TComponent ;
    Index : longint ; Active : boolean ) ;

begin
end ;


procedure TUI_Interface.Signal_Change_Notice( Component : TComponent ;
    Index : longint ; Active : boolean ) ;

begin
end ;


procedure TUI_Interface.Signal_Exception( Component : TComponent ;
    Description : PChar ; Index : longint ) ;

begin
end ;


procedure TUI_Interface.Log_Trace( Component : TComponent ;
    Description : PChar ) ;

begin
end ;


function TUI_Interface.Get_File_Stream( Name : PChar ) : TCOM_Stream ;

begin
    Result := nil ;
end ;


procedure TUI_Interface.Toggle_Embed( Component : TComponent ) ;

begin
end ;


procedure TUI_Interface.Want_Signals( Component : TComponent ;
    Value : boolean ) ;

begin
end ;



{ TComponent methods... }

function TComponent.Facility_Code : longint ;

begin
    Facility_Code := ComponentErr_Facility ;
end ;


function TComponent.Initialize( UI : TUI_Interface ) : TUEC ;

begin
end ;


function TComponent.Terminate : TUEC ;

var N : longint ;
    P : int64 ;

begin
    if( Parent <> nil ) then
    begin
        N := Child_Notice_Terminating ;
        Child_Notification( self, N, P ) ;
    end ;
end ;


function TComponent.Serial_Number : integer ;

begin
    Result := 0 ;
end ;


function TComponent.Add_Notification( Component : TComponent ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TComponent.Cable : TCable ;

begin
    Result := nil ;
end ;


function TComponent.Child_Component( Index : longint ) : TComponent ;

begin
    Child_Component := nil ;
end ;


function TComponent.Clear_Watchpoint( Address : int64 ; Memory : boolean ;
    Access : longint ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_No_Watchpoints ;
end ;


function TComponent.Component_Type : longint ;

begin
    Result := Component_Type_Unknown ;
end ;


function TComponent.Connect_Input( Component : TComponent ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Invalid_Operation ;
end ;


function TComponent.Connect_Output( Component : TComponent ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Invalid_Operation ;
end ;


function TComponent.CPU : TCPU ;

begin
    Result := nil ;
end ;


function TComponent.Debugger : TDebug_Interface ;

begin
    Debugger := nil ; // Not supported
end ;


function TComponent.Delete_Notification( Component : TComponent ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Component_Not_Found ;
end ;


function TComponent.Deposit( Address : int64 ; Size : longint ;
    Buffer : pointer ; Memory : boolean ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Address_Out_Of_Range ;
end ;


function TComponent.Disconnect_Input( Component : TComponent ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Component_Not_Found ;
end ;


function TComponent.Disconnect_Output( Component : TComponent ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Component_Not_Found ;
end ;


function TComponent.Examine( Address : int64 ; var Size : longint ;
    Buffer : pointer ; Memory : boolean ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Address_Out_Of_Range ;
end ;


function TComponent.Get_Access_Mode( Address : int64 ; Memory : boolean ) : longint ;

begin
    Get_Access_Mode := Access_None ;
end ;


function TComponent.Get_Profiling : boolean ;

begin
    Get_Profiling := False ;
end ;


function TComponent.Get_Read_Latency : longint ;

begin
    Get_Read_Latency := 0 ;
end ;


function TComponent.Get_Write_Latency : longint ;

begin
    Get_Write_Latency := 0 ;
end ;


function TComponent.Input_Component( Index : longint ) : TComponent ;

begin
    Input_Component := nil ;
end ;


function TComponent.Keyboard : TKeyboard ;

begin
    Result := nil ;
end ;


function TComponent.Memory : TMemory ;

begin
    Result := nil ;
end ;


function TComponent.Name : PChar ;

begin
    Name := nil ;
end ;


function TComponent.Output_Component( Index : longint ) : TComponent ;

begin
    Output_Component := nil ;
end ;


function TComponent.Read( Address : int64 ; Size : longint ;
    IO_Type : longint ) : boolean ;

begin
    Read := False ;
end ;


function TComponent.Restore_Contents( Stream : TCOM_Stream ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Success ;
end ;


function TComponent.Restore_State( Stream : TCOM_Stream ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Success ;
end ;


function TComponent.Save_Contents( Stream : TCOM_Stream ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Success ;
end ;


function TComponent.Save_State( Stream : TCOM_Stream ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Success ;
end ;


function TComponent.Set_Access_Mode( Low, High : int64 ; Memory : boolean ;
    Typ : longint ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Address_Out_Of_Range ;
end ;


procedure TComponent.Set_Profiling( _On, Children : boolean ) ;

begin
end ;


procedure TComponent.Set_Read_Latency( Value : longint ) ;

begin
end ;


function TComponent.Set_Watchpoint( Address : int64 ; Memory : boolean ;
    Access : longint ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := ComponentErr_Invalid_Operation ;
end ;


procedure TComponent.Set_Write_Latency( Value : longint ) ;

begin
end ;


procedure TComponent.Show_Status ;

begin
end ;


procedure TComponent.Set_Up( P : PChar ) ;

begin
end ;


function TComponent.Support_Feature( ID : integer ) : boolean ;

begin
    Result := False ;
end ;


function TComponent.User_Interface : TUser_Interface ;

begin
    Result := nil ;
end ;


function TComponent.Version : integer ;

begin
    Version := Interface_Version ;
end ;


procedure TComponent.Wake ;

begin
end ;


function TComponent.Write( Address : int64 ; Value, Size : longint ;
    IO_Type : longint ) : TUEC ;

begin
    Write := Deposit( Address, Size, @Value, ( IO_Type = IO_Type_Memory ) ) ;
end ;


function TComponent.Write_String( Address : int64 ; Value : PChar ;
    Size : longint ; IO_Type : longint ) : TUEC ;

var Loop : integer ;

begin
    for Loop := 0 to Size - 1 do
    begin
	Write_String := Write( Address, ord( Value[ Loop ] ), 1, IO_Type ) ;
	Address := Address + 1 ;
    end ;
end ;


procedure TComponent.Child_Notification( Child : TComponent ;
    var Notice : longint ; var Params : int64 ) ;

begin
end ;


procedure TComponent.Set_Signal( Name : PChar ; State : boolean ) ;

begin
end ;


function TComponent.Get_Signal( Name : PChar ; var State : boolean ) : boolean ;

begin
    Result := False ;
end ;


function TComponent.Signal_Count : longint ;

begin
    Result := 0 ;
end ;


function TComponent.Signal_Name( Index : longint ) : PChar ;

begin
    Result := nil ;
end ;


function TComponent.Signal_Out( Index : longint ) : boolean ;

begin
    Result := False ;
end ;


function TComponent.Signal_Active_Low( Index : longint ) : boolean ;

begin
    Result := False ;
end ;


function TComponent.Signal_Index( Name : PChar ) : integer ;

begin
    Result := -1 ;
end ;


procedure TComponent.Reset ;

begin
end ;


function TComponent.Get_State_Name( Index : longint ) : PChar ;

begin
    Result := nil ;
end ;


function TComponent.Get_Exception_Description( Index : longint ) : PChar ;

begin
    Result := nil ;
end ;


procedure TComponent.Set_Tag( Value : longint ) ;

begin
end ;


function TComponent.Get_Tag : longint ;

begin
    Result := 0 ;
end ;


function TComponent.Get_Parent : TComponent ;

begin
    Result := nil ;
end ;


procedure TComponent.Set_Parent( Component : TComponent ) ;

begin
end ;


function TComponent.Profiler : TProfiler ;

begin
    Result := nil ;
end ;


function TComponent.Get_Trace : boolean ;

begin
    Result := False ;
end ;


procedure TComponent.Set_Trace( Value : boolean ) ;

begin
end ;


procedure TComponent.Signal_Change_Notice( Component : TComponent ;
    Index : longint ; Active : boolean ) ;

begin
end ;



{ TMaster_Clock methods... }

procedure TMaster_Clock.Block( Component : TComponent ; Time_Delta : int64 ) ;

begin
    Component.Wake ;
end ;


procedure TMaster_Clock.Initialize( UI : TUI_Interface ) ;

begin
end ;


function TMaster_Clock.Get_Time_Index : int64 ;

begin
    Get_Time_Index := 0 ;
end ;


function TMaster_Clock.Serial_Number : integer ;

begin
    Result := 0 ;
end ;


function TMaster_Clock.Support_Feature( ID : integer ) : boolean ;

begin
    Result := False ;
end ;


function TMaster_Clock.Debugger : TDebug_Interface ;

begin
    Result := nil ;
end ;


function TMaster_Clock.Version : integer ;

begin
    Version := Interface_Version ;
end ;



{ TMemory methods...}

function TMemory.Set_Address_Range( Low, High : int64 ) : TUEC ;

begin
    Result.Facility := -1 ;
    Result.Code := MemoryErr_Success ;
end ;


procedure TMemory.Get_Address_Range( var Low, High : int64 ) ;

begin
end ;


procedure TMemory.Dump( Start, Size : int64 ; Buffer : pointer ) ;

begin
end ;


procedure TMemory.Load( Start, Size : int64 ; Buffer : pointer ) ;

begin
end ;


function TMemory.Version : integer ;

begin
    Version := Interface_Version ;
end ;



{ TMaster_Assembler methods... }

function TMaster_Assembler.Assemble( Input, Output, Listing : TCOM_Stream ;
    Status : TAssembler_Status ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := MasterAssemblerErr_Success ;
end ;


procedure TMaster_Assembler.Add_Reference( Name : PChar ; Size : longint ;
    Address : int64 ) ;

begin
end ;


procedure TMaster_Assembler.Backpatch( Status : TAssembler_Status ;
    Stream : TCOM_Stream ) ;

begin
end ;


function TMaster_Assembler.Evaluate( X : PChar ; var Value : int64 ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := MasterAssemblerErr_Success ;
end ;


function TMaster_Assembler.Expand( Source : PChar ; var Res : PChar ;
    var Res_Length : longint ; Status : TAssembler_Status ) : TUEC ;


begin
    Result.Facility := Facility_Code ;
    Result.Code := MasterAssemblerErr_Success ;
end ;


function TMaster_Assembler.Facility_Code : longint ;

begin
    Result := MasterAssemblerErr_Facility ;
end ;


procedure TMaster_Assembler.Log_Error( Text : PChar ; Severity : longint ) ;

begin
    writeln( Text ) ;
end ;


procedure TMaster_Assembler.In_Line( Input : TCOM_Stream ) ;

begin
end ;


function TMaster_Assembler.Get_Token : PChar ;

begin
    Get_Token := nil ;
end ;


procedure TMaster_Assembler.Put_Token( Token : PChar ) ;

begin
end ;


function TMaster_Assembler.Peek_Token( Same_Line : boolean ) : PChar ;

begin
   Peek_Token := nil ;
end ;


procedure TMaster_Assembler.Push_Scope ;

begin
end ;


procedure TMaster_Assembler.Pop_Scope ;

begin
end ;


function TMaster_Assembler.Get_Symbol( Name : PChar ) : PSymbol_Record ;

begin
    Get_Symbol := nil ;
end ;


function TMaster_Assembler.Add_Symbol( Name : PChar ; P : pSymbol_Record ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := MasterAssemblerErr_Success ;
end ;


procedure TMaster_Assembler.Set_Case_Sensitive( Value : boolean ) ;

begin
end ;


function TMaster_Assembler.Get_Case_Sensitive : boolean ;

begin
    Get_Case_Sensitive := False ;
end ;


function TMaster_Assembler.Grab_Line( Current : boolean ) : PChar ;

begin
    Result := nil ;
end ;


procedure TMaster_Assembler.Map( Address : int64 ) ;

begin
end ;


procedure TMaster_Assembler.UnMap ;

begin
end ;


function TMaster_Assembler.Version : integer ;

begin
    Version := Interface_Version ;
end ;



{ TAssembler methods... }

procedure TAssembler.Initialize( Master : TMaster_Assembler ) ;

begin
end ;


procedure TAssembler.Terminate ;

begin
end ;


function TAssembler.Assemble( inputs : PChar ; var outputs, machines : PChar ;
    var MachineL : longint ; var Address : int64 ;
    var Segment : longint ; Status : TAssembler_Status ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := AssemblerErr_Success ;
end ;


function TAssembler.Default_Radix : longint ;

begin
    Default_Radix := 16 ; { Hexadecimal }
end ;


function TAssembler.Default_Size : longint ;

begin
    Default_Size := 32 ; { 32-bit words }
end ;


function TAssembler.Facility_Code : longint ;

begin
    Result := AssemblerErr_Facility ;
end ;


function TAssembler.Source_Extensions : PChar ;

begin
    Result := nil ;
end ;


function TAssembler.Valid_Symbol_Initial : PChar ;

begin
    Result := nil ;
end ;


function TAssembler.Valid_Symbol_After : PChar ;

begin
    Result := nil ;
end ;


function TAssembler.Version : integer ;

begin
    Version := Interface_Version ;
end ;



{ TCPU methods... }

function TCPU.Big_Endian : boolean ;

begin
    Result := False ; // Default to little endian
end ;


function TCPU.Facility_Code : longint ;

begin
    Result := -1 ;
end ;


function TCPU.Get_Assembler( Master : TMaster_Assembler ) : TAssembler ;

begin
    Get_Assembler := nil ;
end ;


function TCPU.Cancel_Breakpoint( Address : int64 ; Space : integer ;
    Physical : boolean ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := CPUErr_No_Breakpoint ;
end ;


function TCPU.Disassemble( Address : int64 ; Base, Size : longint ;
    Stream : TCOM_Stream ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := CPUErr_No_Breakpoint ;
end ;


function TCPU.Get_Clock_Speed : longint ;

begin
    Get_Clock_Speed := 1 ; { 1 Hz }
end ;


procedure TCPU.Halt ;

begin
end ;


function TCPU.Halted : boolean ; 

begin
    Halted := True ;
end ;


procedure TCPU.Run ;

begin
end ;


procedure TCPU.Run_From_Stream( Stream : TCOM_Stream ) ;

begin
end ;


function TCPU.Set_Breakpoint( Address : int64 ; Space : integer ;
    Physical : boolean ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := CPUErr_Success ;
end ;


procedure TCPU.Set_Clock_Speed( Value : longint ) ;

begin
end ;


procedure TCPU.Step( Into : boolean ) ;

begin
end ;


function TCPU.Translate( Space : integer ; Address : int64 ) : int64 ;

begin
    Translate := Address ;
end ;


function TCPU.Default_Base : integer ;

begin
    Result := 10 ;
end ;


function TCPU.Get_Low_Memory : int64 ;

begin
    Result := 0 ;
end ;


function TCPU.Get_High_Memory : int64 ;

begin
    Result := $7FFFFFFFFFFFFFFF ;
end ;


function TCPU.Get_Low_Virtual_Memory( Space : integer ) : int64 ;

begin
    Result := 0 ;
end ;


function TCPU.Get_High_Virtual_Memory( Space : integer ) : int64 ;

begin
    Result := 0 ;
end ;


function TCPU.Get_Low_Port : int64 ;

begin
    Result := 0 ;
end ;


function TCPU.Get_High_Port : int64 ;

begin
    Result := $7FFFFFFFFFFFFFFF ;
end ;


function TCPU.Support_Virtual_Address : boolean ;

begin
    Result := False ;
end ;


function TCPU.Register_Name( Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


function TCPU.Register_Size( Index : integer ) : integer ;

begin
    Result := 0 ;
end ;


function TCPU.Register_Description( Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


procedure TCPU.Restart ;

begin
end ;


function TCPU.Clear_Internal_Watchpoint( Address : int64 ; Memory : boolean ;
    Access : integer ) : TUEC ;

begin
    Result.Facility := Facility_Code ;
    Result.Code := CPUErr_No_Breakpoint ;
end ;


function TCPU.Set_Internal_Watchpoint( Address : int64 ; Memory : boolean ;
    Access : integer ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TCPU.Get_Current_Address( Space : integer ; Physical : boolean ) : int64 ;

begin
    Result := 0 ;
end ;


procedure TCPU.Set_Current_Address( Space : integer ; Physical : boolean ;
    Value : int64 ) ;

begin
end ;


procedure TCPU.Stop ;

begin
end ;


var Temp_Memory_Space_Description : string = 'Default' ;

function TCPU.Memory_Space_Description( Index : integer ;
    Physical : boolean ) : PChar ;

begin
    if( Index = 0 ) then
    begin
        Result := PChar( Temp_Memory_Space_Description ) ;
    end else
    begin
        Result := nil ;
    end ;
end ;


function TCPU.Page_Size : integer ;

begin
    Result := 0 ;
end ;


function TCPU.Top_Of_Stack( Index : integer ) : int64 ;

begin
    Result := 0 ;
end ;


function TCPU.Version : integer ;

begin
    Version := Interface_Version ;
end ;



// TProfiler methods...

procedure TProfiler.Clear( Domain : integer ) ;

begin
end ;


function TProfiler.Domain_Name( Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


function TProfiler.Report_Line( Domain, Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


function TProfiler.Version : integer ;

begin
    Result := Interface_Version ;
end ;



// TCable methods...

function TCable.Serial : boolean ;

begin
    Result := True ;
end ;


function TCable.Protocol : PChar ;

begin
    Result := nil ;
end ;


function TCable.Transmit( Speed : int64 ; Value, Data_Size, Stop_Bits : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


function TCable.Transmit_String( Speed : int64 ; Value : PChar ;
    Data_Size, Stop_Bits : longint ) : TUEC ;

begin
    fillchar( Result, sizeof( Result ), 0 ) ;
end ;


procedure TCable.Receive( Source : TComponent ; Speed : int64 ;
    Value, Data_Size, Stop_Bits : integer ) ;

begin
end ;


function TCable.Get_Data( var Speed : int64 ;
    var Value, Data_Size, Stop_Bits : integer ) : boolean ;

begin
    Result := False ;
end ;



// TKeyboard methods...

function TKeyboard.Get_Key : PChar ;

begin
    Result := nil ;
end ;


function TKeyboard.Get_Key_Down( Name : PChar ) : boolean ;

begin
    Result := False ;
end ;


procedure TKeyboard.Set_Key_Down( Name : PChar ; State : boolean ) ;

begin
end ;


function TKeyboard.Get_Key_Name( Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


function TKeyboard.Get_LED_State( Name : PChar ) : boolean ;

begin
    Result := False ;
end ;


procedure TKeyboard.Set_LED_State( Name : PChar ; State : boolean ) ;

begin
end ;


function TKeyboard.Get_LED_Name( Index : integer ) : PChar ;

begin
    Result := nil ;
end ;


function TKeyboard.Version : integer ;

begin
    Version := Interface_Version ;
end ;



// TUser_Interface methods...

function TUser_Interface.Get_Hidden : boolean ;

begin
    Result := True ;
end ;


procedure TUser_Interface.Set_Hidden( Value : boolean ) ;

begin
end ;


function TUser_Interface.Get_Parent_Window : THandle ;

begin
    Result := 0 ;
end ;


procedure TUser_Interface.Set_Parent_Window( Value : THandle ) ;

begin
end ;


procedure TUser_Interface.Set_Size( Height, Width : integer ) ;

begin
end ;


function TUser_Interface.Optimal_Height : integer ;

begin
    Result := 0 ;
end ;


function TUser_Interface.Optimal_Width : integer ;

begin
    Result := 0 ;
end ;


function TUser_Interface.Get_Caption : PChar ;

begin
    Result := nil ;
end ;


procedure TUser_Interface.Set_Caption( Value : PChar ) ;

begin
end ;


function TUser_Interface.Version : integer ;

begin
    Version := Interface_Version ;
end ;



end.
